
; ===========================================================================
;                        Eternity II Path Solver
; ===========================================================================

			.586
			.model	flat
			.code

USEHINTS	EQU		0					; flag: 1=use hints 1 and 2, 0=no
;BACKTRACE	EQU		16					; backtrace level (1=normal)
;BACKROW		EQU		6					; minimal fast backtrace row

; ---------------------------------------------------------------------------
;                               Constants
; ---------------------------------------------------------------------------

WDTH		EQU		16					; width of board
HGHT		EQU		16					; height of board
BLOCKS		EQU		WDTH*HGHT			; number of blocks (=256)
COLORS		EQU		23					; number of colors (0=edge)
COLBITS		EQU		8					; number of color bits
COLNUM		EQU		256					; bumber of colors in hash tables

PERIOD		EQU		10000000			; view period

H0BLOCK		EQU		138					; hint 0 block (=block 139)
H0SPIN		EQU		2					; hint 0 spin
H0POSX		EQU		7					; hint 0 position X
H0POSY		EQU		8					; hint 0 position Y
H0POS		EQU		H0POSX+H0POSY*WDTH	; hint 0 position (I,8)
H0COLD		EQU		9					; hint 0 color down
H0COLL		EQU		12					; hint 0 color left
H0COLU		EQU		8					; hint 0 color up
H0COLR		EQU		9					; hint 0 color right

H1BLOCK		EQU		180					; hint 1 block (=block 181)
H1SPIN		EQU		3					; hint 1 spin
H1POSX		EQU		2					; hint 1 position X
H1POSY		EQU		13					; hint 1 position Y
H1POS		EQU		H1POSX+H1POSY*WDTH	; hint 1 position (N,3)
H1COLD		EQU		21					; hint 1 color down
H1COLL		EQU		12					; hint 1 color left
H1COLU		EQU		20					; hint 1 color up
H1COLR		EQU		9					; hint 1 color right

H2BLOCK		EQU		254					; hint 2 block (=block 255)
H2SPIN		EQU		3					; hint 2 spin
H2POSX		EQU		13					; hint 2 position X
H2POSY		EQU		2					; hint 2 position Y
H2POS		EQU		H2POSX+H2POSY*WDTH	; hint 2 position (C,14)
H2COLD		EQU		22					; hint 2 color down
H2COLL		EQU		20					; hint 2 color left
H2COLU		EQU		22					; hint 2 color up
H2COLR		EQU		18					; hint 2 color right

IF (USEHINTS EQ 1)
HINTNUM		EQU		3					; number of hints
ELSE
HINTNUM		EQU		1					; number of hints
ENDIF

DOWN		EQU		0					; DOWN direction
LEFT		EQU		1					; LEFT direction
UP			EQU		2					; UP direction
RIGHT		EQU		3					; RIGHT direction

; ---------------------------------------------------------------------------
;                              Structures
; ---------------------------------------------------------------------------

; --------- Atom (32 bytes)

CAtom		STRUCT

Next		dd		?					; next atom in hash list
ColD		db		?					; color down
ColL		db		?					; color left
ColU		db		?					; color up
ColR		db		?					; color right
Block		db		?					; block number
Spin		db		?					; block spin (clockwise)
			db		?,?					; ... padding
			dd		?					; ... padding
			
CAtom		ENDS

; --------- Field (8 bytes)

CField		STRUCT

ColD		db		?					; color down
ColL		db		?					; color left
ColU		db		?					; color up
ColR		db		?					; color right
Atom		dd		?					; current atom

CField		ENDS

HASHSIZE	EQU		4					; size of hash entry
HASHBITS	EQU		2					; size of hash entry in bits

OFFW		EQU		(WDTH*SIZE CField)	; offset of next row
OFF1		EQU		(SIZE CField)		; offset of next field

; ---------------------------------------------------------------------------
;                              Externals
; ---------------------------------------------------------------------------

extern		_SaveBest:NEAR
extern		_DispState:NEAR

extern		_Atom:CAtom
extern		_Hash:DWORD
extern		_RHash:DWORD
extern		_DHash:DWORD
extern		_RDHash:DWORD
extern		_Board:CField
extern		_Used:BYTE
extern		_Level:DWORD
extern		_MaxLevel:DWORD
;extern		_Tests0:DWORD

; ===========================================================================
;                             Macros
; ===========================================================================

; ---------------------------------------------------------------------------
;                        Get atom from hash table
; ---------------------------------------------------------------------------
; X,Y = field coordinates (constant)
; SRC = pointer to field
; DST = destination register, on input it contains colors of previous field
; DSTH = destination register HIGH (8 bit)
; ---------------------------------------------------------------------------

GetHash		MACRO	X,Y,SRC,DST,DSTH

IF (X EQ 0) AND (Y EQ 0)
			mov		DST,[_Hash]
ELSEIF (X EQ 0)
			movzx	DST,byte ptr [SRC-OFFW].CField.ColD
			shl		DST,COLBITS+HASHBITS
 IF (Y EQ HGHT-1)
			mov		DST,[_DHash+DST]
 ELSE
			mov		DST,[_Hash+DST]
 ENDIF
ELSEIF (Y EQ 0)
			shr		DST,3*8
 IF (X EQ WDTH-1)
			mov		DST,[_RHash+DST*HASHSIZE]
 ELSE
			mov		DST,[_Hash+DST*HASHSIZE]
 ENDIF
 
ELSEIF (X EQ WDTH-1) AND (Y EQ HGHT-1)
			shr		DST,3*8
			mov		DSTH,[SRC-OFFW].CField.ColD
			mov		DST,[_RDHash+DST*HASHSIZE]
ELSEIF (X EQ WDTH-1)
			shr		DST,3*8
			mov		DSTH,[SRC-OFFW].CField.ColD
			mov		DST,[_RHash+DST*HASHSIZE]
ELSEIF (Y EQ HGHT-1)
			shr		DST,3*8
			mov		DSTH,[SRC-OFFW].CField.ColD
			mov		DST,[_DHash+DST*HASHSIZE]
ELSEIF (X EQ H0POSX+1) AND (Y EQ H0POSY)
			movzx	DST,byte ptr [SRC-OFFW].CField.ColD
			shl		DST,COLBITS+HASHBITS
			mov		DST,[_Hash+DST+H0COLR*HASHSIZE]
ELSEIF (X EQ H1POSX+1) AND (Y EQ H1POSY) AND (USEHINTS EQ 1)
			movzx	DST,byte ptr [SRC-OFFW].CField.ColD
			shl		DST,COLBITS+HASHBITS
			mov		DST,[_Hash+DST+H1COLR*HASHSIZE]
ELSEIF (X EQ H2POSX+1) AND (Y EQ H2POSY) AND (USEHINTS EQ 1)
			movzx	DST,byte ptr [SRC-OFFW].CField.ColD
			shl		DST,COLBITS+HASHBITS
			mov		DST,[_Hash+DST+H2COLR*HASHSIZE]
ELSEIF (X EQ H0POSX) AND (Y EQ H0POSY+1)
			shr		DST,3*8
			mov		DST,[_Hash+DST*HASHSIZE+H0COLD*HASHSIZE*COLNUM]
ELSEIF (X EQ H1POSX) AND (Y EQ H1POSY+1) AND (USEHINTS EQ 1)
			shr		DST,3*8
			mov		DST,[_Hash+DST*HASHSIZE+H1COLD*HASHSIZE*COLNUM]
ELSEIF (X EQ H2POSX) AND (Y EQ H2POSY+1) AND (USEHINTS EQ 1)
			shr		DST,3*8
			mov		DST,[_Hash+DST*HASHSIZE+H2COLD*HASHSIZE*COLNUM]
ELSE
			shr		DST,3*8
			mov		DSTH,[SRC-OFFW].CField.ColD
			mov		DST,[_Hash+DST*HASHSIZE]
ENDIF
			ENDM

; ---------------------------------------------------------------------------
;                 Check colors of atom before placing it
; ---------------------------------------------------------------------------
; X,Y = field coordinates (constant)
; ATOM = atom
; ERR = error address
; ---------------------------------------------------------------------------

CheckCol	MACRO	X,Y,ATOM,ERR

IF (X EQ H0POSX-1) AND (Y EQ H0POSY)
			cmp		byte ptr [ATOM].CAtom.ColR,H0COLL
			jne		ERR
ELSEIF (X EQ H1POSX-1) AND (Y EQ H1POSY) AND (USEHINTS EQ 1)
			cmp		byte ptr [ATOM].CAtom.ColR,H1COLL
			jne		ERR
ELSEIF (X EQ H2POSX-1) AND (Y EQ H2POSY) AND (USEHINTS EQ 1)
			cmp		byte ptr [ATOM].CAtom.ColR,H2COLL
			jne		ERR
ELSEIF (X EQ H0POSX) AND (Y EQ H0POSY-1)
			cmp		byte ptr [ATOM].CAtom.ColD,H0COLU
			jne		ERR
ELSEIF (X EQ H1POSX) AND (Y EQ H1POSY-1) AND (USEHINTS EQ 1)
			cmp		byte ptr [ATOM].CAtom.ColD,H1COLU
			jne		ERR
ELSEIF (X EQ H2POSX) AND (Y EQ H2POSY-1) AND (USEHINTS EQ 1)
			cmp		byte ptr [ATOM].CAtom.ColD,H2COLU
			jne		ERR
ENDIF
			ENDM

; ---------------------------------------------------------------------------
;                          Attack one position
; ---------------------------------------------------------------------------
; X,Y = field coordinates (constant)
; NXT = next start label
; PRV = previous continue label
; STRT = this start label
; CONT = this continue label
; ---------------------------------------------------------------------------
; EBP = view coutner
; ESI = pointer to field
; ECX = atom
; EDX = backtrace counter
; ---------------------------------------------------------------------------

At1			MACRO	X,Y,NXT,PRV,STRT,CONT

LOCAL	@Skip,@Find,@Max

; --------- START: Shift pointer to next field (-> ESI)
; ECX must contain colors of previous field

STRT:		

IF (X EQ 0) AND (Y EQ 0)
			mov		esi,offset _Board
ELSEIF (X EQ H0POSX+1) AND (Y EQ H0POSY)
			add		esi,2*SIZE CField
ELSEIF (X EQ H1POSX+1) AND (Y EQ H1POSY) AND (USEHINTS EQ 1)
			add		esi,2*SIZE CField
ELSEIF (X EQ H2POSX+1) AND (Y EQ H2POSY) AND (USEHINTS EQ 1)
			add		esi,2*SIZE CField
ELSE
			add		esi,SIZE CField
ENDIF
			
; --------- Get atom from hash table (-> ECX, ECX must contain colors)

			GetHash	X,Y,esi,ecx,ch
			jmp		@Skip

; --------- CONTINUE: Shift pointer to previous field (-> ESI)
; EDX must contain brutal backtrace counter

CONT:		

IF (X EQ H0POSX-1) AND (Y EQ H0POSY)
			sub		esi,2*SIZE CField
ELSEIF (X EQ H1POSX-1) AND (Y EQ H1POSY) AND (USEHINTS EQ 1)
			sub		esi,2*SIZE CField
ELSEIF (X EQ H2POSX-1) AND (Y EQ H2POSY) AND (USEHINTS EQ 1)
			sub		esi,2*SIZE CField
ELSE
			sub		esi,SIZE CField
ENDIF

; --------- CONT: Remove atom from old position			

			xor		ecx,ecx			
			xchg	ecx,[esi].CField.Atom
			movzx	eax,byte ptr [ecx].CAtom.Block
			btr		dword ptr [_Used],eax
			dec		dword ptr [_Level]
			
; --------- Backtrace counter

;			dec		edx
;			jz		AStop
						
;IF (BACKTRACE GT 1)
; IF (Y GE 0)
;			dec		edx
;			jnz		PRV
; ENDIF
;			mov		dl,2*Y+1
;ENDIF
			
; ---------	Get next atom

@Find:		mov		ecx,[ecx].CAtom.Next
@Skip:		or		ecx,ecx
			jz		PRV

; --------- Check colors of atom

			CheckCol X,Y,ecx,@Find

; --------- Try to lock block usage

			movzx	eax,byte ptr [ecx].CAtom.Block
			bts		dword ptr [_Used],eax
			jc		@Find
			
; --------- Store atom into new position (-> ECX colors)

			mov		[esi].CField.Atom,ecx
			mov		ecx,dword ptr [ecx].CAtom.ColD
			mov		dword ptr [esi].CField.ColD,ecx
			
; --------- Increase number of blocks

			xor		eax,eax
IF (X NE 15) OR (Y NE 15)
			mov		al,X + Y*WDTH + 1
ELSE
			inc		ah
ENDIF
			mov		[_Level],eax
			cmp		eax,[_MaxLevel]
			jbe		@Max
			mov		[_MaxLevel],eax
			push	ecx
			push	edx
			push	esi
			call	_SaveBest
			pop		esi
			pop		edx
			pop		ecx
			
; --------- Display current state

@Max:		dec		ebp
			jnz		NXT
			
; --------- Display current state

			push	ecx
			push	edx
			push	esi
			call	_DispState
			pop		esi
			pop		edx
			pop		ecx
			mov		ebp,PERIOD
			jmp		NXT
			
			ENDM

; ===========================================================================
;                               Solve puzzle
; ===========================================================================

public		_Attack

; --------- Push registers

_Attack:	pusha						; push all registers

; --------- Prepare counter of view period

			mov		ebp,PERIOD ;[_Tests0]		; EBP <- view period
;IF (BACKTRACE GT 1)
;			mov		edx,1
;ENDIF
;			mov		edx,1000000

; --------- Attack positions

			prvlab textequ <AStop>
			y = 0
repeat	16
			x = 0
repeat 16
			strtlab catstr <s>, %x, <_>, %y
			contlab catstr <c>, %x, <_>, %y

IF (x EQ WDTH-1)

 IF (y EQ HGHT-1)
			nxtlab textequ <AStop>
 ELSE
			nxtlab catstr <s>, %0, <_>, %y + 1
 ENDIF

ELSE

 IF (x EQ H0POSX-1) AND (y EQ H0POSY)
			nxtlab catstr <s>, %x + 2, <_>, %y
 ELSEIF (x EQ H1POSX-1) AND (y EQ H1POSY) AND (USEHINTS EQ 1)
			nxtlab catstr <s>, %x + 2, <_>, %y
 ELSEIF (x EQ H2POSX-1) AND (y EQ H2POSY) AND (USEHINTS EQ 1)
			nxtlab catstr <s>, %x + 2, <_>, %y
 ELSE
			nxtlab catstr <s>, %x + 1, <_>, %y
 ENDIF
			
ENDIF


IF (x EQ H0POSX) AND (y EQ H0POSY)
ELSEIF (x EQ H1POSX) AND (y EQ H1POSY) AND (USEHINTS EQ 1)
ELSEIF (x EQ H2POSX) AND (y EQ H2POSY) AND (USEHINTS EQ 1)
ELSE
			At1 x, y, nxtlab, prvlab, strtlab, contlab
			prvlab textequ contlab
ENDIF
			x = x + 1
endm
			y = y + 1
endm

; --------- Pop registers

AStop:		;mov		[_Tests0],ebp
			popa						; pop all registers
			ret

			END
