
//////////////////////////////////////////////////////////////////////////////
//                    Eternity II Brutal Force Solver
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// Includes

#include "Etern2BF.h"

//////////////////////////////////////////////////////////////////////////////
// Variables

struct	CAtom		Atom[ATOMNUM];			// atoms

struct	CAtom*		Hash[COLNUM2];			// middle hash table
struct	CAtom*		RHash[COLNUM2];			// right hash table
struct	CAtom*		DHash[COLNUM2];			// down hash table
struct	CAtom*		RDHash[COLNUM2];		// right+down hash table
#define HashInx(l,u) (l + (u)*COLNUM)		// index in hash table

struct	CField		Board[BLOCKS];			// puzzle board
BYTE				Used[BLOCKS/8];			// block used flag bitmap

int					Level = HINTNUM;		// current reached level
int					MaxLevel = HINTNUM;		// maximum reached level
int					Tests = 0;				// test counter
//int					Tests0 = 1000000;
DWORD				RandSeed;				// seed of random generator
//int					Resets = -1;			// reset counter

//////////////////////////////////////////////////////////////////////////////
// Initialize random generator

void RandInit()
{
	// Get system time
	SYSTEMTIME systime;
	::GetLocalTime(&systime);

	// Prepare temporary time value
	DWORD time = systime.wMilliseconds + 137*systime.wSecond + 
						2567*systime.wMinute + 358946*systime.wDay;

	// Initialize random generator
	RandSeed = (time + 486874*time);
}

//////////////////////////////////////////////////////////////////////////////
// Generate WORD random number

WORD GetRand()
{
	RandSeed = RandSeed*214013 + 2531011;
	return (WORD)(RandSeed >> 16);
}

//////////////////////////////////////////////////////////////////////////////
// Initilize atoms

void AtomInit()
{
	// Local variables
	int i,j,k;
	char cD,cL,cU,cR,tmp;
	struct CColor* color = Color;
	struct CAtom* atom = Atom;
	struct CAtom** hash;
	struct CAtom** prev;

	// Clear hash tables
	for (i = 0; i < COLNUM2; i++)
	{
		Hash[i] = NULL;
		RHash[i] = NULL;
		DHash[i] = NULL;
		RDHash[i] = NULL;
	}

	// Cycle through all blocks
	for (i = 0; i < BLOCKS; i++)
	{
		// Load colors of one block
		cD = color->ColD;
		cL = color->ColL;
		cU = color->ColU;
		cR = color->ColR;

		// Cycle through all spins
		for (j = 0; j < 4; j++)
		{
			// Initialize atom
			atom->Next = NULL;
			atom->Block = i;
			atom->Spin = j;

			// Set colors
			atom->ColD = cD;
			atom->ColL = cL;
			atom->ColU = cU;
			atom->ColR = cR;

			// Hints will not be added into hash tables
			if ((i != H0BLOCK)
#ifdef USEHINTS
				&& (i != H1BLOCK)
				&& (i != H2BLOCK)
#endif // USEHINTS
				)
			{
				// Add atom into hash tables
				hash = ((cR == 0) ? ((cD == 0) ? RDHash : RHash) : ((cD == 0) ? DHash : Hash));
				hash += HashInx(cL, cU);

				k = (GetRand() & 0xff);
				prev = hash;

				for (; k > 0; k--)
				{
					if (*prev == NULL)
						prev = hash;
					else
						prev = &((*prev)->Next);
				}

				atom->Next = *prev;
				*prev = atom;
			}

			// Rotate colors
			tmp = cD; cD = cR; cR = cU; cU = cL; cL = tmp;

			// Next atom
			atom++;
		}

		// Next block
		color++;
	}
}

//////////////////////////////////////////////////////////////////////////////
// Initialize board

void BoardInit()
{
	// Local variables
	int i;
	struct CField* board;
	struct CAtom* atom;
	Level = HINTNUM;
//	MaxLevel = HINTNUM;

	// Clear board
	for (i = 0; i < BLOCKS; i++) Board[i].Atom = NULL;

	// Clear block used flag bitmap
	for (i = 0; i < BLOCKS/8; i++) Used[i] = 0;

	// Save hint 0 into board
	board = &Board[H0POS];
	atom = &Atom[H0BLOCK*4 + H0SPIN];
	board->ColorDW = atom->ColorDW;
	board->Atom = atom;
	Used[H0BLOCK/8] |= (1 << (H0BLOCK & 7));

#ifdef USEHINTS
	// Save hint 1 into board
	board = &Board[H1POS];
	atom = &Atom[H1BLOCK*4 + H1SPIN];
	board->ColorDW = atom->ColorDW;
	board->Atom = atom;
	Used[H1BLOCK/8] |= (1 << (H1BLOCK & 7));

	// Save hint 2 into board
	board = &Board[H2POS];
	atom = &Atom[H2BLOCK*4 + H2SPIN];
	board->ColorDW = atom->ColorDW;
	board->Atom = atom;
	Used[H2BLOCK/8] |= (1 << (H2BLOCK & 7));
#endif // USEHINTS
}

//////////////////////////////////////////////////////////////////////////////
// Save board

void SaveBoard(BOOL best)
{
	// Local variables
	FILE* file;
	char filename[50];
	int i;
	struct CAtom* atom;

	// Open file
	if (best)
	{
		if (MaxLevel < 200) return;
		sprintf(filename, "Best%03d.csv", MaxLevel);
		file = fopen(filename, "w");
	}
	else
		file = fopen("Result.csv", "w");
	if (file == NULL) return;

	// Save block set
	for (i = 0; i < BLOCKS; i++)
	{
		atom = Board[i].Atom;
		if (atom == NULL)
			fprintf(file, "  0,0");
		else
			fprintf(file, "%3d,%1d", (int)(atom->Block+1), (int)atom->Spin);

		if ((i & 0xf) != 15)
			fprintf(file, ";");
		else
			fprintf(file, "\n");
	}

	// Close file
	fclose(file);
}

//////////////////////////////////////////////////////////////////////////////
// Save best board

void SaveBest()
{
	SaveBoard(TRUE);
}

//////////////////////////////////////////////////////////////////////////////
// Display current state

void DispState()
{
	Tests += 10;
	printf("\rCurrent level: %3d, Best level: %3d, Tests: %d millions... ", //, Resets: %d  ",
				Level, MaxLevel, Tests); //, Resets);

	// Save current state
	static int save = 5;
	save--;
	if (save > 0) return;
	save = 5;
	SaveBoard(FALSE);
}

//////////////////////////////////////////////////////////////////////////////
// Main function

int main(int argc, char* argv[])
{
#ifndef _DEBUG
	// Set low priority class (only on Release build)
	::SetPriorityClass(::GetCurrentProcess(), IDLE_PRIORITY_CLASS);
#endif

	// Message
	printf("Eternity II Brutal Force Solver\n");

	// Initialize random generator
	RandInit();

//	do {
//		Resets++;

		// Initilize atoms
		AtomInit();

		// Initialize board
		BoardInit();

		// Solve puzzle
		Attack();

		//// Display state
		//static int disp = 10;
		//disp--;
		//if (disp <= 0)
		//{
		//	Tests -= 10;
		//	DispState();
		//	disp = 500;
		//}

//	} while (MaxLevel < 256);

	return 0;
}
