
//////////////////////////////////////////////////////////////////////////////
//                    Eternity II Brutal Force Solver
//////////////////////////////////////////////////////////////////////////////

//#define		USEHINTS					// Flag, use hints 1 and 2

//////////////////////////////////////////////////////////////////////////////
// Includes

#pragma warning(disable : 4996)			// "xxxx" was declared deprecated

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>

//////////////////////////////////////////////////////////////////////////////
// Constants

#define		WIDTH	16					// width of area
#define		HEIGHT	16					// height of area
#define		BLOCKS	(WIDTH*HEIGHT)		// number of blocks (=256)
#define		COLORS	23					// number of colors (0=edge)
#define		COLBITS	8					// number of bits per color in hash table
#define		COLNUM	(1 << COLBITS)		// number of colors in hash tables (=256)
#define		COLNUM2	(COLORS*COLNUM)		// size of hash table (=5888)
#define		ATOMNUM	(4*BLOCKS)			// number of atoms (=1024)

#define BoardPos(x,y) (x + (y)*WIDTH)	// board position

// Hints
#define		H0BLOCK	138					// hint 0 block (=block 139)
#define		H0SPIN	2					// hint 0 spin
#define		H0POSX	7					// hint 0 position X
#define		H0POSY	8					// hint 0 position Y
#define		H0POS	BoardPos(H0POSX,H0POSY)	// hint 0 position (I,8)
#define		H0COLD	9					// hint 0 color down
#define		H0COLL	12					// hint 0 color left
#define		H0COLU	8					// hint 0 color up
#define		H0COLR	9					// hint 0 color right

#ifdef USEHINTS

#define		H1BLOCK	180					// hint 1 block (=block 181)
#define		H1SPIN	3					// hint 1 spin
#define		H1POSX	2					// hint 1 position X
#define		H1POSY	13					// hint 1 position Y
#define		H1POS	BoardPos(H1POSX,H1POSY)	// hint 1 position (N,3)
#define		H1COLD	21					// hint 1 color down
#define		H1COLL	12					// hint 1 color left
#define		H1COLU	20					// hint 1 color up
#define		H1COLR	9					// hint 1 color right

#define		H2BLOCK	254					// hint 2 block (=block 255)
#define		H2SPIN	3					// hint 2 spin
#define		H2POSX	13					// hint 2 position X
#define		H2POSY	2					// hint 2 position Y
#define		H2POS	BoardPos(H2POSX,H2POSY)	// hint 2 position (C,14)
#define		H2COLD	22					// hint 2 color down
#define		H2COLL	20					// hint 2 color left
#define		H2COLU	22					// hint 2 color up
#define		H2COLR	18					// hint 2 color right

#define		HINTNUM	3					// number of hints

#else // USEHINTS

#define		HINTNUM	1					// number of hints

#endif // USEHINTS

// Directions and turns - from down side, clockwise
enum {
	DOWN	= 0,						// DOWN direction
	LEFT	= 1,						// LEFT direction
	UP		= 2,						// UP direction
	RIGHT	= 3,						// RIGHT direction
};

//////////////////////////////////////////////////////////////////////////////
// Structures

// Color (size 4 bytes)
struct CColor {
	BYTE			ColD;				// color down
	BYTE			ColL;				// color left
	BYTE			ColU;				// color up
	BYTE			ColR;				// color right
};

// Atom (size 16 bytes)
struct CAtom {
	struct CAtom*	Next;				// next atom in hash list (NULL=none)
	union {
		struct {
			BYTE			ColD;		// color down
			BYTE			ColL;		// color left
			BYTE			ColU;		// color up
			BYTE			ColR;		// color right
		};
		DWORD			ColorDW;		// color in DWORD
	};
	BYTE			Block;				// block number
	BYTE			Spin;				// block spin (clockwise)
	BYTE			Res[2];				// ... padding
	int				Res2;				// ... padding
};

// Field (size 8 bytes)
struct CField {
	union {
		struct {
			BYTE			ColD;		// color down
			BYTE			ColL;		// color left
			BYTE			ColU;		// color up
			BYTE			ColR;		// color right
		};
		DWORD			ColorDW;		// color in DWORD
	};
	struct CAtom*	Atom;				// current atom (NULL=none)
};

//////////////////////////////////////////////////////////////////////////////
// Externals for assembler

extern "C" void SaveBest();
extern "C" void DispState();
extern "C" void Attack();

extern "C" struct CAtom Atom[];
extern "C" struct CAtom* Hash[];
extern "C" struct CAtom* RHash[];
extern "C" struct CAtom* DHash[];
extern "C" struct CAtom* RDHash[];
extern "C" struct CField Board[];
extern "C" BYTE Used[];
extern "C" int Level;
extern "C" int MaxLevel;
extern "C" int Tests0;

//////////////////////////////////////////////////////////////////////////////
// Block colors (down, left, up, right)

struct CColor Color[BLOCKS] = {
	{	0,	0,	1,	3,	},			// block 1
	{	0,	0,	1,	4,	},			// block 2
	{	0,	0,	2,	3,	},			// block 3
	{	0,	0,	3,	2,	},			// block 4
	{	0,	1,	6,	1,	},			// block 5
	{	0,	1,	7,	2,	},			// block 6
	{	0,	1,	9,	1,	},			// block 7
	{	0,	1,	9,	5,	},			// block 8
	{	0,	1,	12,	3,	},			// block 9
	{	0,	1,	14,	4,	},			// block 10
	{	0,	1,	15,	2,	},			// block 11
	{	0,	1,	19,	4,	},			// block 12
	{	0,	1,	19,	5,	},			// block 13
	{	0,	1,	21,	4,	},			// block 14
	{	0,	2,	7,	1,	},			// block 15
	{	0,	2,	8,	3,	},			// block 16
	{	0,	2,	10,	5,	},			// block 17
	{	0,	2,	13,	5,	},			// block 18
	{	0,	2,	14,	2,	},			// block 19
	{	0,	2,	15,	2,	},			// block 20
	{	0,	2,	16,	4,	},			// block 21
	{	0,	2,	17,	1,	},			// block 22
	{	0,	2,	17,	5,	},			// block 23
	{	0,	2,	18,	1,	},			// block 24
	{	0,	2,	21,	1,	},			// block 25
	{	0,	3,	6,	2,	},			// block 26
	{	0,	3,	6,	3,	},			// block 27
	{	0,	3,	7,	3,	},			// block 28
	{	0,	3,	8,	3,	},			// block 29
	{	0,	3,	14,	5,	},			// block 30
	{	0,	3,	15,	2,	},			// block 31
	{	0,	3,	18,	3,	},			// block 32
	{	0,	3,	19,	2,	},			// block 33
	{	0,	3,	19,	4,	},			// block 34
	{	0,	3,	20,	5,	},			// block 35
	{	0,	3,	22,	4,	},			// block 36
	{	0,	4,	8,	1,	},			// block 37
	{	0,	4,	11,	5,	},			// block 38
	{	0,	4,	12,	5,	},			// block 39
	{	0,	4,	13,	2,	},			// block 40
	{	0,	4,	13,	3,	},			// block 41
	{	0,	4,	15,	1,	},			// block 42
	{	0,	4,	15,	2,	},			// block 43
	{	0,	4,	15,	3,	},			// block 44
	{	0,	4,	16,	1,	},			// block 45
	{	0,	4,	18,	4,	},			// block 46
	{	0,	4,	19,	4,	},			// block 47
	{	0,	4,	20,	4,	},			// block 48
	{	0,	5,	6,	5,	},			// block 49
	{	0,	5,	7,	1,	},			// block 50
	{	0,	5,	7,	2,	},			// block 51
	{	0,	5,	9,	1,	},			// block 52
	{	0,	5,	14,	4,	},			// block 53
	{	0,	5,	16,	4,	},			// block 54
	{	0,	5,	16,	5,	},			// block 55
	{	0,	5,	19,	3,	},			// block 56
	{	0,	5,	20,	1,	},			// block 57
	{	0,	5,	20,	5,	},			// block 58
	{	0,	5,	21,	2,	},			// block 59
	{	0,	5,	22,	3,	},			// block 60
	{	6,	6,	9,	8,	},			// block 61
	{	6,	6,	10,	14,	},			// block 62
	{	6,	7,	7,	11,	},			// block 63
	{	6,	8,	6,	19,	},			// block 64
	{	6,	8,	8,	22,	},			// block 65
	{	6,	8,	10,	10,	},			// block 66
	{	6,	8, 	12,	7,	},			// block 67
	{	6,	8,	18,	9,	},			// block 68
	{	6,	8,	22,	19,	},			// block 69
	{	6,	11,	11,	14,	},			// block 70
	{	6,	11,	14,	17,	},			// block 71
	{	6,	12,	10,	8,	},			// block 72
	{	6,	12,	15,	16,	},			// block 73
	{	6,	12,	18,	15,	},			// block 74
	{	6,	12,	19,	11,	},			// block 75
	{	6,	13,	10,	15,	},			// block 76
	{	6,	13,	13,	15,	},			// block 77
	{	6,	14,	11,	20,	},			// block 78
	{	6,	14,	18,	11,	},			// block 79
	{	6,	14,	20,	21,	},			// block 80
	{	6,	15,	13,	8,	},			// block 81
	{	6,	16,	8,	8,	},			// block 82
	{	6,	16,	12,	16,	},			// block 83
	{	6,	17,	8,	13,	},			// block 84
	{	6,	17,	9,	10,	},			// block 85
	{	6,	17,	19,	17,	},			// block 86
	{	6,	17,	20,	18,	},			// block 87
	{	6,	18,	6,	21,	},			// block 88
	{	6,	18,	9,	22,	},			// block 89
	{	6,	18,	16,	20,	},			// block 90
	{	6,	19,	12,	17,	},			// block 91
	{	6,	19,	13,	15,	},			// block 92
	{	6,	19,	13,	16,	},			// block 93
	{	6,	19,	16,	21,	},			// block 94
	{	6,	19,	17,	10,	},			// block 95
	{	6,	21,	21,	11,	},			// block 96
	{	6,	22,	16,	13,	},			// block 97
	{	6,	22,	18,	19,	},			// block 98
	{	6,	22,	21,	9,	},			// block 99
	{	6,	22,	22,	21,	},			// block 100
	{	7,	7, 	17,	15,	},			// block 101
	{	7,	7, 	17,	20,	},			// block 102
	{	7,	7, 	20,	13,	},			// block 103
	{	7,	7,	22,	9,	},			// block 104
	{	7,	8, 	16,	15,	},			// block 105
	{	7,	9, 	11,	19,	},			// block 106
	{	7,	9, 	13,	19,	},			// block 107
	{	7,	9, 	16,	15,	},			// block 108
	{	7,	9,	20,	12,	},			// block 109
	{	7,	10,	15,	17,	},			// block 110
	{	7,	10,	17,	15,	},			// block 111
	{	7,	11,	18,	13,	},			// block 112
	{	7,	11,	18,	20,	},			// block 113
	{	7,	12,	10,	16,	},			// block 114
	{	7,	12,	14,	17,	},			// block 115
	{	7,	12,	17,	12,	},			// block 116
	{	7,	12,	22,	20,	},			// block 117
	{	7,	13,	11,	21,	},			// block 118
	{	7,	14,	20,	17,	},			// block 119
	{	7,	15,	19,	22,	},			// block 120
	{	7,	16,	10,	22,	},			// block 121
	{	7,	18,	9,	20,	},			// block 122
	{	7,	18,	10,	13,	},			// block 123
	{	7,	18,	18,	15,	},			// block 124
	{	7,	19,	17,	22,	},			// block 125
	{	7,	19,	21,	15,	},			// block 126
	{	7,	20,	10,	9,	},			// block 127
	{	7,	20,	13,	21,	},			// block 128
	{	7,	20,	16,	11,	},			// block 129
	{	7,	20,	18,	19,	},			// block 130
	{	7,	21,	9,	18,	},			// block 131
	{	7,	21,	17,	10,	},			// block 132
	{	7,	22,	10,	20,	},			// block 133
	{	7,	22,	12,	16,	},			// block 134
	{	7,	22,	16,	11,	},			// block 135
	{	7,	22,	20,	18,	},			// block 136
	{	8,	8,	18,	14,	},			// block 137
	{	8,	9, 	9,	11,	},			// block 138
	{	8,	9,	9,	12,	},			// block 139 HINT 0
	{	8,	9, 	9,	17,	},			// block 140
	{	8,	9, 	13,	21,	},			// block 141
	{	8,	9, 	15,	9,	},			// block 142
	{	8,	9, 	20,	17,	},			// block 143
	{	8,	9, 	21,	21,	},			// block 144
	{	8,	10,	11,	16,	},			// block 145
	{	8,	11,	8,	17,	},			// block 146
	{	8,	11,	8,	22,	},			// block 147
	{	8,	11,	11,	10,	},			// block 148
	{	8,	11,	15,	17,	},			// block 149
	{	8,	13,	9,	12,	},			// block 150
	{	8,	13,	16,	22,	},			// block 151
	{	8,	14,	12,	12,	},			// block 152
	{	8,	14,	12,	13,	},			// block 153
	{	8,	14,	22,	20,	},			// block 154
	{	8,	16,	14,	14,	},			// block 155
	{	8,	16,	14,	17,	},			// block 156
	{	8,	16,	22,	14,	},			// block 157
	{	8,	18,	14,	20,	},			// block 158
	{	8,	18,	19,	9,	},			// block 159
	{	8,	19,	21,	21,	},			// block 160
	{	8,	20,	9,	18,	},			// block 161
	{	8,	20,	10,	18,	},			// block 162
	{	8,	22,	15,	12,	},			// block 163
	{	8,	22,	16,	20,	},			// block 164
	{	9,	10,	11,	16,	},			// block 165
	{	9,	10,	16,	19,	},			// block 166
	{	9,	12,	11,	11,	},			// block 167
	{	9,	13,	12,	15,	},			// block 168
	{	9,	13,	13,	21,	},			// block 169
	{	9,	14,	16,	19,	},			// block 170
	{	9,	14,	18,	20,	},			// block 171
	{	9,	14,	21,	12,	},			// block 172
	{	9,	15,	15,	15,	},			// block 173
	{	9,	15,	17,	18,	},			// block 174
	{	9,	16,	14,	21,	},			// block 175
	{	9,	17,	14,	13,	},			// block 176
	{	9,	18,	10,	16,	},			// block 177
	{	9,	19,	17,	20,	},			// block 178
	{	9,	19,	19,	15,	},			// block 179
	{	9,	20,	14,	20,	},			// block 180
	{	9,	21,	12,	20,	},			// block 181 HINT 1
	{	9,	21,	14,	12,	},			// block 182
	{	10,	10,	13,	19,	},			// block 183
	{	10,	11,	22,	14,	},			// block 184
	{	10,	12,	13,	17,	},			// block 185
	{	10,	12,	19,	19,	},			// block 186
	{	10,	13,	21,	14,	},			// block 187
	{	10,	14,	10,	21,	},			// block 188
	{	10,	14,	11,	13,	},			// block 189
	{	10,	14,	20,	13,	},			// block 190
	{	10,	15,	11,	11,	},			// block 191
	{	10,	15,	15,	18,	},			// block 192
	{	10,	16,	12,	14,	},			// block 193
	{	10,	17,	21,	12,	},			// block 194
	{	10,	17,	21,	22,	},			// block 195
	{	10,	17,	22,	15,	},			// block 196
	{	10,	18,	12,	22,	},			// block 197
	{	10,	18,	13,	19,	},			// block 198
	{	10,	18,	18,	18,	},			// block 199
	{	10,	19,	13,	11,	},			// block 190
	{	10,	20,	21,	19,	},			// block 201
	{	10,	20,	22,	14,	},			// block 202
	{	10,	21,	17,	13,	},			// block 203
	{	10,	21,	17,	19,	},			// block 204
	{	10,	21,	20,	11,	},			// block 205
	{	10,	21,	22,	21,	},			// block 206
	{	11,	11,	22,	14,	},			// block 207
	{	11,	12,	13,	22,	},			// block 208
	{	11,	12,	19,	15,	},			// block 209
	{	11,	14,	12,	13,	},			// block 210
	{	11,	14,	20,	15,	},			// block 211
	{	11,	15,	11,	20,	},			// block 212
	{	11,	16,	19,	19,	},			// block 213
	{	11,	17,	11,	18,	},			// block 214
	{	11,	17,	16,	22,	},			// block 215
	{	11,	17,	22,	21,	},			// block 216
	{	11,	18,	13,	15,	},			// block 217
	{	11,	20,	16,	17,	},			// block 218
	{	11,	21,	12,	16,	},			// block 219
	{	11,	22,	12,	20,	},			// block 220
	{	11,	22,	21,	21,	},			// block 221
	{	11,	22,	21,	22,	},			// block 222
	{	12,	12,	17,	22,	},			// block 223
	{	12,	12,	18,	14,	},			// block 224
	{	12,	12,	20,	15,	},			// block 225
	{	12,	14,	13,	15,	},			// block 226
	{	12,	14,	17,	17,	},			// block 227
	{	12,	16,	13,	19,	},			// block 228
	{	12,	18,	14,	22,	},			// block 229
	{	12,	18,	20,	19,	},			// block 230
	{	12,	19,	17,	18,	},			// block 231
	{	12,	19,	17,	21,	},			// block 232
	{	13,	13,	13,	18,	},			// block 233
	{	13,	14,	20,	16,	},			// block 234
	{	13,	16,	14,	16,	},			// block 235
	{	13,	16,	14,	18,	},			// block 236
	{	13,	16,	17,	15,	},			// block 237
	{	13,	17,	16,	20,	},			// block 238
	{	13,	18,	15,	22,	},			// block 239
	{	13,	18,	21,	15,	},			// block 240
	{	13,	19,	14,	21,	},			// block 241
	{	13,	19,	16,	21,	},			// block 242
	{	14,	15,	15,	17,	},			// block 243
	{	14,	15,	18,	19,	},			// block 244
	{	14,	16,	22,	18,	},			// block 245
	{	14,	21,	20,	22,	},			// block 246
	{	15,	15,	21,	22,	},			// block 247
	{	15,	16,	17,	16,	},			// block 248
	{	15,	17,	16,	21,	},			// block 249
	{	15,	18,	20,	21,	},			// block 250
	{	16,	16,	22,	19,	},			// block 251
	{	16,	17,	19,	17,	},			// block 252
	{	17,	19,	20,	18,	},			// block 253
	{	18,	20,	21,	20,	},			// block 254
	{	18,	22,	20,	22,	},			// block 255 HINT 2
	{	19,	22,	21,	22,	},			// block 256
};
