
; ===========================================================================
;                        Eternity II Frequency Solver
; ===========================================================================

			.586
			.model	flat
			.code

USEHINTS	EQU		0					; flag: 1=use hints 1 and 2, 0=no

; ---------------------------------------------------------------------------
;                               Constants
; ---------------------------------------------------------------------------

WDTH		EQU		16					; width of board
HGHT		EQU		16					; height of board
BLOCKS		EQU		WDTH*HGHT			; number of blocks (=256)
ATOMNUM		EQU		(4*BLOCKS)			; number of atoms (4 spins)
ATOMBITS	EQU		10					; number of bits per all atoms

PERIOD		EQU		10000000			; view period

H0BLOCK		EQU		138					; hint 0 block (=block 139)
H0SPIN		EQU		2					; hint 0 spin
H0POSX		EQU		7					; hint 0 position X
H0POSY		EQU		8					; hint 0 position Y
H0POS		EQU		H0POSX+H0POSY*WDTH	; hint 0 position (I,8)
H0ATOM		EQU		H0BLOCK*4+H0SPIN	; hint 0 atom

H1BLOCK		EQU		180					; hint 1 block (=block 181)
H1SPIN		EQU		3					; hint 1 spin
H1POSX		EQU		2					; hint 1 position X
H1POSY		EQU		13					; hint 1 position Y
H1POS		EQU		H1POSX+H1POSY*WDTH	; hint 1 position (N,3)
H1ATOM		EQU		H1BLOCK*4+H1SPIN	; hint 1 atom

H2BLOCK		EQU		254					; hint 2 block (=block 255)
H2SPIN		EQU		3					; hint 2 spin
H2POSX		EQU		13					; hint 2 position X
H2POSY		EQU		2					; hint 2 position Y
H2POS		EQU		H2POSX+H2POSY*WDTH	; hint 2 position (C,14)
H2ATOM		EQU		H2BLOCK*4+H2SPIN	; hint 2 atom

IF (USEHINTS EQ 1)
HINTNUM		EQU		3					; number of hints
ELSE
HINTNUM		EQU		1					; number of hints
ENDIF

DIR_DOWN	EQU		0					; DOWN direction
DIR_LEFT	EQU		1					; LEFT direction
DIR_UP		EQU		2					; UP direction
DIR_RIGHT	EQU		3					; RIGHT direction

PERIOD		EQU		10000000			; view period

; ---------------------------------------------------------------------------
;                              Structures
; ---------------------------------------------------------------------------

; --------- Hash entry (4 bytes)

CHash		STRUCT

			UNION
				STRUCT
Atom		dw		?					; atom index (= block*4 + spin)
Weight		dw		?					; weight of atom
				ENDS
AtomWeight	dd		?
			ENDS
			  
CHash		ENDS

; --------- Hash list (8 bytes)

CList		STRUCT

List		dd		?					; pointer to list of atoms
Num			dd		?					; number of atoms in the list

CList		ENDS

LISTBITS	EQU		3					; number of bits of CList

; --------- Field (4 bytes)

CField		STRUCT

Hash		dd		?					; pointer into atom hash list (NULL = none)

CField		ENDS

OFFW		EQU		(WDTH*SIZE CField)	; offset of next row
OFF1		EQU		(SIZE CField)		; offset of next field

; ---------------------------------------------------------------------------
;                              Externals
; ---------------------------------------------------------------------------

extern		_SaveBest:NEAR
extern		_DispState:NEAR

extern		_Board:CField
extern		_Used:BYTE
extern		_Level:DWORD
extern		_MaxLevel:DWORD
extern		_Tests:DWORD

extern		_Hint0L:BYTE
extern		_Hint0U:BYTE
IF (USEHINTS EQ 1)
extern		_Hint1L:BYTE
extern		_Hint1U:BYTE
extern		_Hint2L:BYTE
extern		_Hint2U:BYTE
ENDIF

extern		_Hash0:CList

extern		_HashD:CList
extern		_HashDDE:CList
extern		_HashR:CList
extern		_HashRRE:CList

extern		_Hash:CList
extern		_HashRE:CList
extern		_HashDE:CList
extern		_HashREDE:CList

; ---------------------------------------------------------------------------
;                        Get hash entry from hash table
; ---------------------------------------------------------------------------
; X,Y = field coordinates (constant)
; SRC = pointer to field
; DST = destination register
; TMP = temporary register
; ---------------------------------------------------------------------------

GetHash		MACRO	X,Y,SRC,DST,TMP

IF (X EQ 0) AND (Y EQ 0)
			mov		DST,[_Hash0].CList.List
ELSEIF (X EQ 0)
			mov		DST,[SRC-OFFW].CField.Hash
			movzx	DST,[DST].CHash.Atom
			shl		DST,LISTBITS
 IF (Y EQ HGHT-1)
			mov		DST,[_HashDDE+DST].CList.List
 ELSE
			mov		DST,[_HashD+DST].CList.List
 ENDIF
ELSEIF (Y EQ 0)
			mov		DST,[SRC-OFF1].CField.Hash
			movzx	DST,[DST].CHash.Atom
			shl		DST,LISTBITS
 IF (X EQ WDTH-1)
			mov		DST,[_HashRRE+DST].CList.List
 ELSE
			mov		DST,[_HashR+DST].CList.List
 ENDIF
ELSEIF (X EQ WDTH-1) AND (Y EQ HGHT-1)
			mov		DST,[SRC-OFF1].CField.Hash
			movzx	DST,[DST].CHash.Atom
			shl		DST,LISTBITS
			mov		TMP,[SRC-OFFW].CField.Hash
			movzx	TMP,[TMP].CHash.Atom
			shl		TMP,ATOMBITS+LISTBITS
			or		DST,TMP			
			mov		DST,[_HashREDE+DST].CList.List
ELSEIF (X EQ WDTH-1)
			mov		DST,[SRC-OFF1].CField.Hash
			movzx	DST,[DST].CHash.Atom
			shl		DST,LISTBITS
			mov		TMP,[SRC-OFFW].CField.Hash
			movzx	TMP,[TMP].CHash.Atom
			shl		TMP,ATOMBITS+LISTBITS
			or		DST,TMP			
			mov		DST,[_HashRE+DST].CList.List
ELSEIF (Y EQ HGHT-1)
			mov		DST,[SRC-OFF1].CField.Hash
			movzx	DST,[DST].CHash.Atom
			shl		DST,LISTBITS
			mov		TMP,[SRC-OFFW].CField.Hash
			movzx	TMP,[TMP].CHash.Atom
			shl		TMP,ATOMBITS+LISTBITS
			or		DST,TMP			
			mov		DST,[_HashDE+DST].CList.List
ELSE
			mov		DST,[SRC-OFF1].CField.Hash
			movzx	DST,[DST].CHash.Atom
			shl		DST,LISTBITS
			mov		TMP,[SRC-OFFW].CField.Hash
			movzx	TMP,[TMP].CHash.Atom
			shl		TMP,ATOMBITS+LISTBITS
			or		DST,TMP			
			mov		DST,[_Hash+DST].CList.List
ENDIF
			ENDM

; ---------------------------------------------------------------------------
;                 Check colors of atom before placing it
; ---------------------------------------------------------------------------
; X,Y = field coordinates (constant)
; ATOM = atom
; ERR = error address
; ---------------------------------------------------------------------------

CheckCol	MACRO	X,Y,ATOM,ERR

IF (X EQ H0POSX-1) AND (Y EQ H0POSY)
			bt		dword ptr [_Hint0L],ATOM
			jnc		ERR
ELSEIF (X EQ H1POSX-1) AND (Y EQ H1POSY) AND (USEHINTS EQ 1)
			bt		dword ptr [_Hint1L],ATOM
			jnc		ERR
ELSEIF (X EQ H2POSX-1) AND (Y EQ H2POSY) AND (USEHINTS EQ 1)
			bt		dword ptr [_Hint2L],ATOM
			jnc		ERR
ELSEIF (X EQ H0POSX) AND (Y EQ H0POSY-1)
			bt		dword ptr [_Hint0U],ATOM
			jnc		ERR
ELSEIF (X EQ H1POSX) AND (Y EQ H1POSY-1) AND (USEHINTS EQ 1)
			bt		dword ptr [_Hint1U],ATOM
			jnc		ERR
ELSEIF (X EQ H2POSX) AND (Y EQ H2POSY-1) AND (USEHINTS EQ 1)
			bt		dword ptr [_Hint2U],ATOM
			jnc		ERR
ENDIF
			ENDM

; ---------------------------------------------------------------------------
;                          Attack one position
; ---------------------------------------------------------------------------
; X,Y = field coordinates (constant)
; NXT = next start label
; PRV = previous continue label
; STRT = this start label
; CONT = this continue label
; ---------------------------------------------------------------------------
; EBP = view counter
; ESI = pointer to field
; EAX = temporary
; EBX = atom
; ECX = hash
; ---------------------------------------------------------------------------

At1			MACRO	X,Y,NXT,PRV,STRT,CONT

LOCAL	@Skip,@Find,@Max

; --------- START: Shift pointer to next field (-> ESI)
; ECX must contain colors of previous field

STRT:		

IF (X EQ 0) AND (Y EQ 0)
			mov		esi,offset _Board
ELSEIF (X EQ H0POSX+1) AND (Y EQ H0POSY)
			add		esi,2*SIZE CField
ELSEIF (X EQ H1POSX+1) AND (Y EQ H1POSY) AND (USEHINTS EQ 1)
			add		esi,2*SIZE CField
ELSEIF (X EQ H2POSX+1) AND (Y EQ H2POSY) AND (USEHINTS EQ 1)
			add		esi,2*SIZE CField
ELSE
			add		esi,SIZE CField
ENDIF
			
; --------- Get hash entry from hash table (-> ECX)

			GetHash	X,Y,esi,ecx,eax
			jmp		@Skip

; --------- CONTINUE: Shift pointer to previous field (-> ESI)

CONT:		

IF (X EQ H0POSX-1) AND (Y EQ H0POSY)
			sub		esi,2*SIZE CField
ELSEIF (X EQ H1POSX-1) AND (Y EQ H1POSY) AND (USEHINTS EQ 1)
			sub		esi,2*SIZE CField
ELSEIF (X EQ H2POSX-1) AND (Y EQ H2POSY) AND (USEHINTS EQ 1)
			sub		esi,2*SIZE CField
ELSE
			sub		esi,SIZE CField
ENDIF

; --------- CONT: Remove atom from old position			

			xor		ecx,ecx			
			xchg	ecx,[esi].CField.Hash
			movzx	eax,[ecx].CHash.Atom
			shr		eax,2
			btr		dword ptr [_Used],eax
			dec		[_Level]

; ---------	Get next atom

@Find:		add		ecx,SIZE CHash
@Skip:		mov		ebx,[ecx].CHash.AtomWeight
			or		ebx,ebx
			js		PRV
			movzx	ebx,bx

; --------- Check colors of atom

			CheckCol X,Y,ebx,@Find

; --------- Try to lock block usage

			shr		ebx,2
			bts		dword ptr [_Used],ebx
			jc		@Find
			
; --------- Store atom into new position

			mov		[esi].CField.Hash,ecx
			
; --------- Increase number of blocks

			mov		eax,[_Level]
			inc		eax
			mov		[_Level],eax
			cmp		eax,[_MaxLevel]
			jbe		@Max
			mov		[_MaxLevel],eax
			push	esi
			push	ebp
			call	_SaveBest
			pop		ebp
			pop		esi
			
; --------- Display current state

@Max:		dec		ebp
			jnz		NXT
			
; --------- Display current state

			push	esi
			call	_DispState
			pop		esi
			mov		ebp,PERIOD
			jmp		NXT
			
			ENDM

; ===========================================================================
;                               Solve puzzle
; ===========================================================================

public		_Attack

; --------- Push registers

_Attack:	pusha						; push all registers

; --------- Prepare counter of view period

			mov		ebp,PERIOD			; EBP <- view period

; --------- Attack positions

			prvlab textequ <AttackStop>
			y = 0
repeat HGHT
			x = 0
repeat WDTH
			strtlab catstr <s>, %x, <_>, %y
			contlab catstr <c>, %x, <_>, %y

IF (x EQ WDTH-1)

 IF (y EQ HGHT-1)
			nxtlab textequ <AttackStop>
 ELSE
			nxtlab catstr <s>, %0, <_>, %y + 1
 ENDIF

ELSE

 IF (x EQ H0POSX-1) AND (y EQ H0POSY)
			nxtlab catstr <s>, %x + 2, <_>, %y
 ELSEIF (x EQ H1POSX-1) AND (y EQ H1POSY) AND (USEHINTS EQ 1)
			nxtlab catstr <s>, %x + 2, <_>, %y
 ELSEIF (x EQ H2POSX-1) AND (y EQ H2POSY) AND (USEHINTS EQ 1)
			nxtlab catstr <s>, %x + 2, <_>, %y
 ELSE
			nxtlab catstr <s>, %x + 1, <_>, %y
 ENDIF
			
ENDIF

IF (x EQ H0POSX) AND (y EQ H0POSY)
ELSEIF (x EQ H1POSX) AND (y EQ H1POSY) AND (USEHINTS EQ 1)
ELSEIF (x EQ H2POSX) AND (y EQ H2POSY) AND (USEHINTS EQ 1)
ELSE
			At1 x, y, nxtlab, prvlab, strtlab, contlab
			prvlab textequ contlab
ENDIF
			x = x + 1
endm
			y = y + 1
endm

; --------- Pop registers

AttackStop:	popa						; pop all registers
			ret

			END
