//////////////////////////////////////////////////////////////////////////////
//                    Eternity II Frequency Solver
//////////////////////////////////////////////////////////////////////////////

#include "Etern2Freq.h"

//////////////////////////////////////////////////////////////////////////////
// Variables

struct CField		Board[BLOCKS];			// puzzle board
BYTE				Used[BLOCKS/8];			// block used flag bitmap

int					Level = HINTNUM;		// current level
int					MaxLevel = HINTNUM;		// maximum level
int					Tests = 0;				// test counter
DWORD				RandSeed;				// seed of random generator

int					Stat[STATMAX];			// block statistic
int					StatNum[SIDENUM];		// number of neighbors

struct CHash*		Hint0 = NULL;			// pointer to hint 0 in hash buffer
BYTE				Hint0L[ATOMNUM/8];		// left to hint 0 flags
BYTE				Hint0U[ATOMNUM/8];		// up to hint 0 flags

#ifdef USEHINTS
struct CHash*		Hint1 = NULL;			// pointer to hint 0 in hash buffer
BYTE				Hint1L[ATOMNUM/8];		// left to hint 1 flags
BYTE				Hint1U[ATOMNUM/8];		// up to hint 1 flags
struct CHash*		Hint2 = NULL;			// pointer to hint 0 in hash buffer
BYTE				Hint2L[ATOMNUM/8];		// left to hint 2 flags
BYTE				Hint2U[ATOMNUM/8];		// up to hint 2 flags
#endif // USEHINTS

struct CHash		HashBuf[MAXHASH];		// buffer of hash entries

struct CList		Hash0[1];				// hash list of LEFT-UP corners

struct CList		HashD[ATOMNUM];			// hash list in DOWN direction
struct CList		HashDDE[ATOMNUM];		// hash list in DOWN direction with DOWN edge
struct CList		HashR[ATOMNUM];			// hash list in RIGHT direction
struct CList		HashRRE[ATOMNUM];		// hash list in RIGHT direction with RIGHT edge

struct CList		Hash[ATOMNUM2];			// hash list in DOWN+RIGHT direction
struct CList		HashRE[ATOMNUM2];		// hash list with RIGHT edge
struct CList		HashDE[ATOMNUM2];		// hash list with DOWN edge
struct CList		HashREDE[ATOMNUM2];		// hash list with RIGHT and DOWN edge

#define HashInx2(up,left) (up*ATOMNUM+left)	// index in double hash table

// Hash lists
struct CList* HashLists[HASHNUM] = {
	Hash0,
	HashD, HashDDE, HashR, HashRRE,
	Hash, HashRE, HashDE, HashREDE,
};

// Hash sizes
int HashSizes[HASHNUM] = {
	1,
	ATOMNUM, ATOMNUM, ATOMNUM, ATOMNUM,
	ATOMNUM2, ATOMNUM2, ATOMNUM2, ATOMNUM2,
};

//////////////////////////////////////////////////////////////////////////////
// Initialize random generator

void RandInit()
{
	// Get system time
	SYSTEMTIME systime;
	::GetLocalTime(&systime);

	// Prepare temporary time value
	DWORD time = systime.wMilliseconds + 137*systime.wSecond + 
						2567*systime.wMinute + 358946*systime.wDay;

	// Initialize random generator
	RandSeed = (time + 486874*time);
}

//////////////////////////////////////////////////////////////////////////////
// Generate WORD random number

WORD GetRand()
{
	RandSeed = RandSeed*214013 + 2531011;
	return (WORD)(RandSeed >> 16);
}

//////////////////////////////////////////////////////////////////////////////
// Generate WORD random number in range 0 to MAX (including)
// Notes: It is more precise random generator than simple modulo operation.

//WORD _fastcall GetRandom(WORD max)
//{
//	WORD res;
//	WORD mask;
//
//	// Maximum value = 0
//	if (max == 0) return 0;
//	
//	// Prepare starting value of the mask
//	if (max >= (WORD)0x100) res = 0xFFFF; else res = 0xFF;
//
//	// Calculate mask
//	do {
//		mask = res;
//		res >>= 1;
//	} while (res >= max);
//
//	// Generate random number
//	do {
//		res = GetRand() & mask;
//	} while (res > max);
//
//	return res;
//}

//////////////////////////////////////////////////////////////////////////////
// Load block statistic (return TRUE = all OK)

BOOL LoadStat()
{
	// Open file with statistic
	FILE* file = fopen("Stat.bin", "rb");
	if (file == NULL)
	{
		printf("Cannot find Stat.bin file!\n");
		return FALSE;
	}

	// Read block statistic
	int res = fread(Stat, 1, STATMAX*sizeof(int), file);
	fclose(file);
	if (res != STATMAX*sizeof(int))
	{
		printf("Invalid size of Stat.bin file!\n");
		return FALSE;
	}

	// Check color statistic
	int i, j, m, n;
	for (i = 0; i < SIDENUM; i++)
	{
		for (j = 0; j < SIDENUM; j++)
		{
			n = Stat[i*SIDENUM+j];
			m = Stat[j*SIDENUM+i];
			if ((n < 0) ||
				(n > 10000000) ||
				(n != m))
			{
				printf("Invalid content of Stat.bin file!\n");
				return FALSE;
			}
		}
	}

	return TRUE;
}

//////////////////////////////////////////////////////////////////////////////
// Initialize statistic

void InitStat()
{
	// Local variables
	int i, j, k, n, s, t;

	// Clear hint flags
	for (i = 0; i < ATOMNUM/8; i++)
	{
		Hint0L[i] = 0;
		Hint0U[i] = 0;
#ifdef USEHINTS
		Hint1L[i] = 0;
		Hint1U[i] = 0;
		Hint2L[i] = 0;
		Hint2U[i] = 0;
#endif // USEHINTS
	}

	// Normalizace block statistic
	for (i = 0; i < SIDENUM; i++)
	{
		n = 0;
		s = 0;

		// Count sum of statistic of one block
		for (j = 0; j < SIDENUM; j++)
		{
			k = Stat[i*SIDENUM+j];
			if (k > 0)
			{
				s += k;
				n++;

				// Hint flags
				if (i == H0BLOCK*4+((4+DIR_LEFT-H0SPIN)&3))
				{
					t = (j&~3) + ((4 + DIR_RIGHT - (j&3))&3);
					Hint0L[t/8] |= 1 << (t&7);
				}

				if (i == H0BLOCK*4+((4+DIR_UP-H0SPIN)&3))
				{
					t = (j&~3) + ((4 + DIR_DOWN - (j&3))&3);
					Hint0U[t/8] |= 1 << (t&7);
				}

#ifdef USEHINTS
				if (i == H1BLOCK*4+((4+DIR_LEFT-H1SPIN)&3))
				{
					t = (j&~3) + ((4 + DIR_RIGHT - (j&3))&3);
					Hint1L[t/8] |= 1 << (t&7);
				}

				if (i == H1BLOCK*4+((4+DIR_UP-H1SPIN)&3))
				{
					t = (j&~3) + ((4 + DIR_DOWN - (j&3))&3);
					Hint1U[t/8] |= 1 << (t&7);
				}

				if (i == H2BLOCK*4+((4+DIR_LEFT-H2SPIN)&3))
				{
					t = (j&~3) + ((4 + DIR_RIGHT - (j&3))&3);
					Hint2L[t/8] |= 1 << (t&7);
				}

				if (i == H2BLOCK*4+((4+DIR_UP-H2SPIN)&3))
				{
					t = (j&~3) + ((4 + DIR_DOWN - (j&3))&3);
					Hint2U[t/8] |= 1 << (t&7);
				}
#endif // USEHINTS
			}
		}
		StatNum[i] = n;

		// Normalize statistic of one block
		if (s > 0)
		{
			for (j = 0; j < SIDENUM; j++)
			{
				t = (int)((__int64)Stat[i*SIDENUM+j]*1000*n/s);
				if (t > 0xffff) t = 0xffff;
				Stat[i*SIDENUM+j] = -t;
			}
		}
	}

	// Merge block statistics (use bigger value)
	for (i = 0; i < SIDENUM; i++)
	{
		for (j = 0; j < SIDENUM; j++)
		{
			n = Stat[i*SIDENUM+j];
			if (n < 0)
			{
				n = -n;
				s = -Stat[j*SIDENUM+i];
				if (n < s) n = s;
				Stat[i*SIDENUM+j] = n;
				Stat[j*SIDENUM+i] = n;
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// Add one atom into hash list (sorted)

void AddHash(BOOL save, struct CList* list, WORD atom, WORD weight)
{
	if (save)
	{
		int i = list->Num;
		struct CHash* hash = list->List;
		for (; (i > 0) && (weight > hash[i-1].Weight); i--)
		{
			hash[i].AtomWeight = hash[i-1].AtomWeight;
		}
		hash[i].Atom = atom;
		hash[i].Weight = weight;
		if (atom == H0ATOM) Hint0 = &hash[i];
#ifdef USEHINTS
		if (atom == H1ATOM) Hint1 = &hash[i];
		if (atom == H2ATOM) Hint2 = &hash[i];
#endif // USEHINTS
	}
	list->Num++;
}

//////////////////////////////////////////////////////////////////////////////
// Add all atoms into single hash lists

void AddAll1Hash(BOOL save)
{
	// Local variables
	int side, block, dir, atomD, atomR;
	int side2, block2, dir2, atom2U, atom2L;
	int weight;

	// LEFT-UP corners (blocks 0 to 3, edge is in DOWN and LEFT direction)
	AddHash(save, Hash0, 1*4 + 1, GetRand() & 0xf);
	AddHash(save, Hash0, 3*4 + 1, GetRand() & 0xf);
	AddHash(save, Hash0, 0*4 + 1, GetRand() & 0xf);
	AddHash(save, Hash0, 2*4 + 1, GetRand() & 0xf);

	// Cycle through all sides of blocks
	for (side = 0; side < SIDENUM; side++)
	{
		// Prepare block and direction
		block = side/4;
		dir = side & 3;

		// Index of atom to be this side in DOWN direction
		atomD = block*4 + ((4 + DIR_DOWN - dir) & 3);

		// Index of atom to be this side in RIGHT direction
		atomR = block*4 + ((4 + DIR_RIGHT - dir) & 3);

		// Cycle through all sides of neighbors
		for (side2 = 0; side2 < SIDENUM; side2++)
		{
			// Check if this neighbor can be connected
			weight = Stat[side*SIDENUM+side2];
			if (weight != 0)
			{
				// Prepare block and direction of neighbor
				block2 = side2/4;
				dir2 = side2 & 3;

				// Index of neighbor atom to be this side in UP direction
				atom2U = block2*4 + ((4 + DIR_UP - dir2) & 3);

				// Index of neighbor atom to be this side in LEFT direction
				atom2L = block2*4 + ((4 + DIR_LEFT - dir2) & 3);

				// Check, if oposite side of neighbor is edge
				if (StatNum[block2*4 + (dir2 ^ 2)] != 0)
				{
					// Add atom into hash list without edge
					AddHash(save, &HashD[atomD], atom2U, weight + GetRand() & 0x3f);
					AddHash(save, &HashR[atomR], atom2L, weight + GetRand() & 0x3f);
				}
				else
				{
					// Add atom into hash list with edge
					AddHash(save, &HashDDE[atomD], atom2U, weight + GetRand() & 0x3f);
					AddHash(save, &HashRRE[atomR], atom2L, weight + GetRand() & 0x3f);
				}
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// Add all atoms into specified double hash lists

void AddAll2Hash0(BOOL save, struct CList* listD, struct CList* listR, struct CList* list)
{
	// Local variables
	int atomD, atomR, atom, inx, weight;
	struct CHash *hashU, *hashL;

	// Cycle through all atoms for DOWN direction
	for (atomD = 0; atomD < ATOMNUM; atomD++)
	{
		// Cycle through all atoms for RIGHT direction
		for (atomR = 0; atomR < ATOMNUM; atomR++)
		{
			// Destination index in hash table
			inx = HashInx2(atomD, atomR);

			// Cycle through hash list of UP atoms
			hashU = listD[atomD].List;
			while (hashU->AtomWeight >= 0)
			{
				atom = hashU->Atom;

				// Find the same atom in LEFT atom
				hashL = listR[atomR].List;
				while (hashL->AtomWeight >= 0)
				{
					if (atom == hashL->Atom)
					{
						// Add atom into hash table
						weight = hashU->Weight*hashL->Weight/1000;
						if (weight > 0xffff) weight = 0xffff;
						AddHash(save, &list[inx], atom, weight);
					}
					hashL++;
				}
				hashU++;
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// Add all atoms into double hash lists

void AddAll2Hash(BOOL save)
{
	AddAll2Hash0(save, HashD, HashR, Hash);
	AddAll2Hash0(save, HashDDE, HashR, HashDE);
	AddAll2Hash0(save, HashD, HashRRE, HashRE);
	AddAll2Hash0(save, HashDDE, HashRRE, HashREDE);
}

//////////////////////////////////////////////////////////////////////////////
// Initialize hash lists

void InitHash()
{
	// Local variables
	int i, n;
	struct CList* list;
	struct CHash* hash;

	printf("Generating hash tables...");

	// Clear hash tables
	for (i = 0; i < HASHNUM; i++)
	{
		list = HashLists[i];
		for (n = HashSizes[i]; n > 0; n--)
		{
			list->Num = 0;
			list++;
		}
	}

	// Add all atoms into single hash tables (only count number of blocks)
	AddAll1Hash(FALSE);

	// Setup singe hash lists
	hash = HashBuf;
	for (i = 0; i < HASHNUM; i++)
	{
		n = HashSizes[i];
		if (n != ATOMNUM2)
		{
			list = HashLists[i];
			for (; n > 0; n--)
			{
				list->List = hash;
				hash += list->Num + 1;
				list->Num = 0;
				if (hash <= &HashBuf[MAXHASH]) hash[-1].AtomWeight = -1;
				list++;
			}
		}
	}

	// Re-add all atoms into single hash tables
	AddAll1Hash(TRUE);

	// Shorter hash chains to speed-up solve speed
	for (i = 0; i < HASHNUM; i++)
	{
		n = HashSizes[i];
		if (n == ATOMNUM)
		{
			list = HashLists[i];
			//if ((list == HashD) || (list == HashR))
			{
				for (; n > 0; n--)
				{
					if ((list->Num > 10))// && (list->Num < 30))
					{
						list->Num = list->Num*3/4;
						if (&list->List[list->Num] == Hint0) list->Num--;
#ifdef USEHINTS
						if (&list->List[list->Num] == Hint1) list->Num--;
						if (&list->List[list->Num] == Hint2) list->Num--;
#endif // USEHINTS
						list->List[list->Num].AtomWeight = -1;
					}
					list++;
				}
			}
		}
	}

	// Add all atoms into double hash tables (only count number of blocks)
	AddAll2Hash(FALSE);

	// Setup singe hash lists
	for (i = 0; i < HASHNUM; i++)
	{
		n = HashSizes[i];
		if (n == ATOMNUM2)
		{
			list = HashLists[i];
			for (; n > 0; n--)
			{
				list->List = hash;
				hash += list->Num + 1;
				list->Num = 0;
				if (hash <= &HashBuf[MAXHASH]) hash[-1].AtomWeight = -1;
				list++;
			}
		}
	}

	// Re-add all atoms into double hash tables
	AddAll2Hash(TRUE);
	
	// print hash buffer size
	i = (int)(hash - HashBuf);
	printf("\rHash buffer size: %d\n", i);
	if (i > MAXHASH)
	{
		printf("Hash buffer overflow!!!\n");
		exit(1);
	}

	// Calculate neighbor statistic
	int stat[8];
	for (i = 0; i < 8; i++) stat[i] = 0;
	for (i = 0; i < ATOMNUM2; i++)
	{
		if (Hash[i].Num >= 7)
			stat[7]++;
		else
			stat[Hash[i].Num]++;
	}
	printf("Neighbor statistic 1:%d,2:%d,3:%d,4:%d,5:%d,6:%d,7>:%d\n",
			stat[1], stat[2], stat[3], stat[4], stat[5], stat[6], stat[7]);

	//// Shorter hash chains to speed-up solve speed
//	for (i = 0; i < HASHNUM; i++)
//	{
//		n = HashSizes[i];
//		if (n == ATOMNUM2)
//		{
//			list = HashLists[i];
//			if (list == Hash)
//			{
//				for (; n > 0; n--)
//				{
//					if (list->Num >= 2)
//					{
//						list->Num = list->Num*3/4;
//						if (&list->List[list->Num] == Hint0) list->Num--;
//#ifdef USEHINTS
//						if (&list->List[list->Num] == Hint1) list->Num--;
//						if (&list->List[list->Num] == Hint2) list->Num--;
//#endif // USEHINTS
//						if (list->Num <= 0) list->Num = 1;
//						list->List[list->Num].AtomWeight = -1;
//					}
//					list++;
//				}
//			}
//		}
//	}
}

//////////////////////////////////////////////////////////////////////////////
// Add hint into board

void AddHint(int pos, struct CHash* hash)
{
	if (hash == NULL)
	{
		printf("Invalid statistic file!\n");
		exit(1);
	}
	Board[pos].Hash = hash;
	int block = hash->Atom/4;
	Used[block/8] |= (1 << (block & 7));
}

//////////////////////////////////////////////////////////////////////////////
// Initialize board

void InitBoard()
{
	Level = HINTNUM;
	int i;

	// Clear board
	for (i = 0; i < BLOCKS; i++) Board[i].Hash = NULL;

	// Clear used flags
	for (i = 0; i < BLOCKS/8; i++) Used[i] = 0;

	// Add hints
	AddHint(H0POS, Hint0);
#ifdef USEHINTS
	AddHint(H1POS, Hint1);
	AddHint(H2POS, Hint2);
#endif // USEHINTS
}

//////////////////////////////////////////////////////////////////////////////
// Save board

void SaveBoard(BOOL best)
{
	// Local variables
	FILE* file;
	char filename[50];
	int i;
	struct CHash* hash;

	// Open file
	if (best)
	{
		if (Level < 200) return;
		sprintf(filename, "Best%03d.csv", Level);
		file = fopen(filename, "w");
	}
	else
		file = fopen("Result.csv", "w");
	if (file == NULL) return;

	// Save block set
	for (i = 0; i < BLOCKS; i++)
	{
		hash = Board[i].Hash;
		if (hash == NULL)
			fprintf(file, "  0,0");
		else
			fprintf(file, "%3d,%1d", (int)(hash->Atom/4+1), (int)hash->Atom&3);

		if ((i & 0xf) != 15)
			fprintf(file, ";");
		else
			fprintf(file, "\n");
	}

	// Close file
	fclose(file);
}

//////////////////////////////////////////////////////////////////////////////
// Save best board

void SaveBest()
{
	SaveBoard(TRUE);
}

//////////////////////////////////////////////////////////////////////////////
// Display current state

void DispState()
{
	Tests += 10;
	printf("\rCurrent level: %3d, Best level: %3d, Tests: %d millions... ",
				Level, MaxLevel, Tests);

	// Save current state
	static int save = 5;
	save--;
	if (save > 0) return;
	save = 5;
	SaveBoard(FALSE);
}

//////////////////////////////////////////////////////////////////////////////
// Main function

int main(int argc, char* argv[])
{
#ifndef _DEBUG
	// Set low priority class (only on Release build)
	::SetPriorityClass(::GetCurrentProcess(), IDLE_PRIORITY_CLASS);
#endif

	// Message
	printf("Eternity II Frequency Solver\n");

	// Load block statistic
	if (!LoadStat()) return 1;

	// Initialize random generator
	RandInit();

	// Initialize statistic
	InitStat();

	// Initialize hash lists
	InitHash();

	// Initialize board
	InitBoard();

	// Solve puzzle
	Attack();

	// Save current best state
	SaveBest();
	DispState();
	SaveBoard(FALSE);
	return 0;
}
