//////////////////////////////////////////////////////////////////////////////
//                    Eternity II Frequency Solver
//////////////////////////////////////////////////////////////////////////////

//#define		USEHINTS				// flag, use hints

//////////////////////////////////////////////////////////////////////////////
// Includes

#pragma warning(disable : 4996)			// "xxxx" was declared deprecated

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>

//////////////////////////////////////////////////////////////////////////////
// Constants

#define		WIDTH	16				// width of area
#define		HEIGHT	16				// height of area
#define		BLOCKS	(WIDTH*HEIGHT)	// number of blocks (=256)
#define		ATOMNUM	(4*BLOCKS)		// number of atoms (4 spins)
#define		SIDENUM	(4*BLOCKS)		// number of sides of all blocks
#define		ATOMNUM2 (ATOMNUM*ATOMNUM) // size of 2-dimension table
#define		STATMAX	(SIDENUM*SIDENUM) // size of block statistic
#define		MAXHASH	10000000		// size of hash index buffer (6166300)
#define		HASHNUM	9				// number of hash lists

#define BoardPos(x,y) (x + (y)*WIDTH)	// board position
#define	AtomInx(block,spin) (block*4+spin) // index of atom

// Hints
#define		H0BLOCK	138					// hint 0 block (=block 139)
#define		H0SPIN	2					// hint 0 spin
#define		H0POSX	7					// hint 0 position X
#define		H0POSY	8					// hint 0 position Y
#define		H0POS	BoardPos(H0POSX,H0POSY)	// hint 0 position (I,8)
#define		H0ATOM	AtomInx(H0BLOCK,H0SPIN) // hint 0 atom

#ifdef USEHINTS

#define		H1BLOCK	180					// hint 1 block (=block 181)
#define		H1SPIN	3					// hint 1 spin
#define		H1POSX	2					// hint 1 position X
#define		H1POSY	13					// hint 1 position Y
#define		H1POS	BoardPos(H1POSX,H1POSY)	// hint 1 position (N,3)
#define		H1ATOM	AtomInx(H1BLOCK,H1SPIN) // hint 1 atom

#define		H2BLOCK	254					// hint 2 block (=block 255)
#define		H2SPIN	3					// hint 2 spin
#define		H2POSX	13					// hint 2 position X
#define		H2POSY	2					// hint 2 position Y
#define		H2POS	BoardPos(H2POSX,H2POSY)	// hint 2 position (C,14)
#define		H2ATOM	AtomInx(H2BLOCK,H2SPIN) // hint 2 atom

#define		HINTNUM	3					// number of hints

#else // USEHINTS

#define		HINTNUM	1					// number of hints

#endif // USEHINTS

// Directions and turns - from down side, clockwise
enum {
	DIR_DOWN	= 0,					// DOWN direction
	DIR_LEFT	= 1,					// LEFT direction
	DIR_UP		= 2,					// UP direction
	DIR_RIGHT	= 3,					// RIGHT direction
};

//////////////////////////////////////////////////////////////////////////////
// Structures

// Hash entry (size 4 bytes)
struct CHash {
	union {
		struct {
			WORD	Atom;				// atom index (= block*4 + spin)
			WORD	Weight;				// weight of atom
		};
		int			AtomWeight;
	};
};

// Hash list (size 8 bytes)
struct CList {
	struct CHash*	List;				// pointer to list of atoms
	int				Num;				// number of atoms in the list
};

// Field (size 4 bytes)
struct CField {
	struct CHash*	Hash;				// pointer into atom hash list (NULL=none)
};

//////////////////////////////////////////////////////////////////////////////
// Externals for assembler

extern "C" void SaveBest();
extern "C" void DispState();
extern "C" void Attack();

extern "C" struct CField Board[];
extern "C" BYTE Used[];
extern "C" int Level;
extern "C" int MaxLevel;
extern "C" int Tests;

extern "C" BYTE Hint0L[];
extern "C" BYTE	Hint0U[];

#ifdef USEHINTS
extern "C" BYTE Hint1L[];
extern "C" BYTE	Hint1U[];
extern "C" BYTE Hint2L[];
extern "C" BYTE	Hint2U[];
#endif // USEHINTS

extern "C" struct CList Hash0[];

extern "C" struct CList HashD[];
extern "C" struct CList HashDDE[];
extern "C" struct CList HashR[];
extern "C" struct CList HashRRE[];

extern "C" struct CList Hash[];
extern "C" struct CList HashRE[];
extern "C" struct CList HashDE[];
extern "C" struct CList HashREDE[];
