//////////////////////////////////////////////////////////////////////////////
//  Eternity II Monte Carlo puzzle solver, v1.0, (c) 2007 Miroslav Nemecek
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// Includes

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include "EternityIIMC.h"

//////////////////////////////////////////////////////////////////////////////
// Variables

// Random generator
DWORD	RandSeed;						// seed of random generator

// Puzzle board
struct	CField	Board[BLOCKS];			// puzzle board

// List of used blocks and fields
struct	CField*	Used[BLOCKS];			// list of used blocks (NULL=none)
int		UsedNum = 0;					// number of used fields and blocks
int		MaxUsed = 1;					// maximal used blocks
int		Attempts = 0;					// attempt counter

// Atom list
struct	CAtom	Atom[ATOMNUM];			// atoms
struct	CAtom*	AtomList = NULL;		// list of atoms
struct	CAtom*	AtomLast = (CAtom*)(void*)&AtomList; // last atom in the list

// Push state
struct	CField	SaveBoard[BLOCKS];		// saved puzzle board
struct	CField*	SaveUsed[BLOCKS];		// saved list of used blocks
int		SaveBreakOK = -1;
int		SaveBreakRe;
#define	SAVECOUNT 5000
int		SaveCount = SAVECOUNT;			// counter to reload best board

// Attack configuration
int		BreakCount = 10;				// asm break counter

int		BreakOK = 50;					// break counter on success
int		BreakRe = 50;					// reinicialize break counter
int		BreakBlk = 1;					// number of breaked blocks
int		BreakOKInc = 1;					// last increment of break counter on success
int		BreakReInc = 1;					// last increment of reinicialize break counter
//int		BreakBlkInc = 1;				// last increment of number of breaked blocks
int		BreakOKOld = 50;				// old break counter on success
int		BreakReOld = 50;				// old reinicialize break counter
//int		BreakBlkOld = 1;				// old number of breaked blocks
int		BreakLevel = 1;					// temporary reached level
int		BreakOld = 1;					// old reached level
int		BreakTest = 100;				// counter to optimise configuration
int		BreakTest1 = 10;				// counter to optimise one parameter
int		BreakMode = 0;					// optimised parameter
#define	BREAKLOCKCOUNT 50
int		BreakLockCount = BREAKLOCKCOUNT; // counter for lock test

//////////////////////////////////////////////////////////////////////////////
// Initialize random generator

void RandInit()
{
	// Get system time
	SYSTEMTIME systime;
	::GetLocalTime(&systime);

	// Prepare temporary time value
	DWORD time = systime.wMilliseconds + 137*systime.wSecond + 
						2567*systime.wMinute + 358946*systime.wDay;

	// Initialize random generator
	RandSeed = (time + 486874*time);
}

//////////////////////////////////////////////////////////////////////////////
// Generate BYTE random number

BYTE GetRand()
{
	RandSeed = RandSeed*214013 + 2531011;
	return (BYTE)(RandSeed >> 16);
}

//////////////////////////////////////////////////////////////////////////////
// Initialize board

void BoardInit()
{
	// Local variables
	CField* field = Board;

	// Cycle through all fields of the board
	for (int y = 0; y < HEIGHT; y++)
	{
		for (int x = 0; x < WIDTH; x++)
		{
			// Clear atom in the field
			field->Atom = NULL;

			// Initialize color
			field->Color[DIR_UP] = 0;
			field->Color[DIR_RIGHT] = 0;
			field->Color[DIR_DOWN] = 0;
			field->Color[DIR_LEFT] = 0;

			// Initialize color mask
			field->ColMask[DIR_UP] = 0;
			field->ColMask[DIR_RIGHT] = 0;
			field->ColMask[DIR_DOWN] = 0;
			field->ColMask[DIR_LEFT] = 0;

			// Initialize edge flags
			field->Edge = 0;
			if (y == 0)
			{
				field->Edge |= EDGE_UP;
				field->ColMask[DIR_UP] = 0xff;
			}

			if (x == (WIDTH-1))
			{
				field->Edge |= EDGE_RIGHT;
				field->ColMask[DIR_RIGHT] = 0xff;
			}

			if (y == (HEIGHT-1))
			{
				field->Edge |= EDGE_DOWN;
				field->ColMask[DIR_DOWN] = 0xff;
			}

			if (x == 0)
			{
				field->Edge |= EDGE_LEFT;
				field->ColMask[DIR_LEFT] = 0xff;
			}

			field++;
		}
	}

	// Clear list of used fields of blocks
	for (int i = 0; i < BLOCKS; i++) Used[i] = NULL;

	UsedNum = 0;
//	MaxUsed = 1;
}

//////////////////////////////////////////////////////////////////////////////
// Generate atoms

void GenAtom()
{
	// Local variables
	int i,j,k;
	BYTE tmp,colU,colR,colD,colL;
	const CBlock* block = Block;
	CAtom* atom = Atom;

	// Initialize atom lists
	AtomList = NULL;
	AtomLast = (CAtom*)(void*)&AtomList;

	// Cycle through all blocks
	for (i = 0; i < BLOCKS; i++)
	{
		// Load colors of one block
		colU = block->Color[DIR_UP];
		colR = block->Color[DIR_RIGHT];
		colD = block->Color[DIR_DOWN];
		colL = block->Color[DIR_LEFT];

		// Cycle through all spins
		for (j = 0; j < 4; j++)
		{
			k = -1;

			// Hints
			if ((i == H0BLOCK) && (j == H0SPIN)) k = H0POS;
#ifdef USEHINTS
			if ((i == H1BLOCK) && (j == H1SPIN)) k = H1POS;
			if ((i == H2BLOCK) && (j == H2SPIN)) k = H2POS;
#endif // USEHINTS

			// Store hint atom into board
			if (k >= 0)
			{
				Board[k-WIDTH].Color[DIR_DOWN] = colU;
				Board[k-WIDTH].ColMask[DIR_DOWN] = 0xff;

				Board[k+1].Color[DIR_LEFT] = colR;
				Board[k+1].ColMask[DIR_LEFT] = 0xff;

				Board[k+WIDTH].Color[DIR_UP] = colD;
				Board[k+WIDTH].ColMask[DIR_UP] = 0xff;

				Board[k-1].Color[DIR_RIGHT] = colL;
				Board[k-1].ColMask[DIR_RIGHT] = 0xff;

				Board[k].Edge |= EDGE_HINT;

				Board[k].Atom = atom;
				Used[i] = &Board[k];
				UsedNum++;
			}

			// Set block and spin
			atom->Block = i;
			atom->Spin = j;
			atom->Next = NULL;

			// Store colors
			atom->Color[DIR_UP] = colU;
			atom->Color[DIR_RIGHT] = colR;
			atom->Color[DIR_DOWN] = colD;
			atom->Color[DIR_LEFT] = colL;

			// Store atom in the list (except hint atoms)
#ifdef USEHINTS
			if ((i != H0BLOCK) && (i != H1BLOCK) && (i != H2BLOCK))
#else // USEHINTS
			if (i != H0BLOCK)
#endif // USEHINTS
			{
				AtomLast->Next = atom;
				AtomLast = atom;
			}

			// Rotate colors
			tmp = colL; colL = colD; colD = colR; colR = colU; colU = tmp;

			atom++;
		}
		block++;
	}
}

//////////////////////////////////////////////////////////////////////////////
// Save board

void SaveBoardFile(BOOL best)
{
	int i;

	// Open file
	FILE* file;
	if (best)
	{
		if (UsedNum < 200) return;
		char filename[50];
		sprintf(filename, "Best%d.csv", UsedNum);
		file = fopen(filename, "w");
	}
	else
	{
		file = fopen("Result.csv", "w");
	}

	if (file == NULL) return;

	// Save board
	i = 0;
	for (int y = 0; y < HEIGHT; y++)
	{
		for (int x = 0; x < WIDTH; x++)
		{
			if (Board[i].Atom == NULL)
				fprintf(file, "  0,0");
			else
			{
				CAtom* atom = Board[i].Atom;
				fprintf(file, "%3d,%1d", (int)(atom->Block + 1), (int)atom->Spin);
			}

			if (x < WIDTH-1) fprintf(file, ";");
			i++;
		}
		fprintf(file, "\n");
	}

	// Close file
	fclose(file);
}

//////////////////////////////////////////////////////////////////////////////
// Push best board

void PushState()
{
	for (int i = 0; i < BLOCKS; i++)
	{
		SaveBoard[i].Atom = Board[i].Atom;

		SaveBoard[i].Color[0] = Board[i].Color[0];
		SaveBoard[i].Color[1] = Board[i].Color[1];
		SaveBoard[i].Color[2] = Board[i].Color[2];
		SaveBoard[i].Color[3] = Board[i].Color[3];

		SaveBoard[i].ColMask[0] = Board[i].ColMask[0];
		SaveBoard[i].ColMask[1] = Board[i].ColMask[1];
		SaveBoard[i].ColMask[2] = Board[i].ColMask[2];
		SaveBoard[i].ColMask[3] = Board[i].ColMask[3];

		SaveBoard[i].Edge = Board[i].Edge;

		SaveUsed[i] = Used[i];
	}
	SaveBreakOK = BreakOK;
	SaveBreakRe = BreakRe;
}

//////////////////////////////////////////////////////////////////////////////
// Save best board

void SaveBest()
{
	SaveBoardFile(TRUE);
	PushState();
}

//////////////////////////////////////////////////////////////////////////////
// Pop best board

void PopState()
{
	if (SaveBreakOK < 0) return;

	for (int i = 0; i < BLOCKS; i++)
	{
		Board[i].Atom = SaveBoard[i].Atom;

		Board[i].Color[0] = SaveBoard[i].Color[0];
		Board[i].Color[1] = SaveBoard[i].Color[1];
		Board[i].Color[2] = SaveBoard[i].Color[2];
		Board[i].Color[3] = SaveBoard[i].Color[3];

		Board[i].ColMask[0] = SaveBoard[i].ColMask[0];
		Board[i].ColMask[1] = SaveBoard[i].ColMask[1];
		Board[i].ColMask[2] = SaveBoard[i].ColMask[2];
		Board[i].ColMask[3] = SaveBoard[i].ColMask[3];

		Board[i].Edge = SaveBoard[i].Edge;

		Used[i] = SaveUsed[i];
	}
	BreakOK = SaveBreakOK;
	BreakRe = SaveBreakRe;
	UsedNum = MaxUsed;
}

//////////////////////////////////////////////////////////////////////////////
// Display current state

void DispState()
{
	Attempts += 1;

	static int resflag = 0;
	static int undoflag = 0;

#define BREAKTEST 200
#define BREAKTEST1 10

//	BreakBlk = 1;

	// Attack optimisation
	BreakLevel += UsedNum;
	BreakTest--;
	if (BreakTest <= 0)
	{
		BreakTest = BREAKTEST;
		BreakLevel = (BreakLevel + BreakTest/2)/BreakTest;

		switch (BreakMode)
		{
		case 0:
			if (BreakLevel < BreakOld)
			{
				BreakOK = BreakOKOld;
				BreakOKInc = -BreakOKInc;
			}
			else
				if (BreakLevel > BreakOld)
					BreakOKOld = BreakOK;

			if (BreakTest1 > 1)
			{
				//if ((BreakLevel != BreakOld) || (BreakTest1 == BREAKTEST1))
				//	BreakOKInc = (GetRand() & 2)-1;
				BreakOK += BreakOKInc;
				if (BreakOK > 90) 
				{
					BreakOK = 90;
					BreakOKInc = -1;
				}
				if (BreakOK < 20)
				{
					BreakOK = 20;
					BreakOKInc = 1;
				}
			}
			else
			{
				BreakOKOld = BreakOK;
			}
			break;

		case 1:
			if (BreakLevel < BreakOld)
			{
				BreakRe = BreakReOld;
				BreakReInc = -BreakReInc;
			}
			else
				if (BreakLevel > BreakOld)
					BreakReOld = BreakRe;

			if (BreakTest1 > 1)
			{
				//if ((BreakLevel != BreakOld) || (BreakTest1 == BREAKTEST1))
				//	BreakReInc = (GetRand() & 2)-1;
				BreakRe += BreakReInc;
				if (BreakRe > 100)
				{
					BreakRe = 100;
					BreakReInc = -1;
				}
				if (BreakRe < 1)
				{
					BreakRe = 1;
					BreakReInc = 1;
				}
			}
			else
			{
				BreakReOld = BreakRe;
			}
			break;

		//case 2:
		//	if (BreakLevel < BreakOld)
		//		BreakBlk = BreakBlkOld;
		//	else
		//		BreakBlkOld = BreakBlk;
		//	if (BreakTest1 > 1)
		//	{
		//		if (GetRand() & 1)
		//		{
		//			BreakBlk++;
		//			if (BreakBlk > 50) BreakBlk = 50;
		//		}
		//		else
		//		{
		//			BreakBlk--;
		//			if (BreakBlk < 1) BreakBlk = 1;
		//		}
		//	}
		//	break;
		}

		BreakTest1--;
		if (BreakTest1 <= 0)
		{
			BreakTest1 = BREAKTEST1;
			BreakMode = GetRand() & 1;
//			do BreakMode = GetRand() & 3; while (BreakMode == 3);
		}

		if ((BreakOld == BreakLevel) || 
			(BreakOld == BreakLevel-1) ||
			(BreakOld == BreakLevel+1))
		{
			BreakLockCount--;
			if (BreakLockCount <= 0)
			{
				BreakLockCount = BREAKLOCKCOUNT;
				BreakBlk = 50;
				resflag = 20;
				BreakOKOld = BreakOK;
				BreakReOld = BreakRe;
			}
		}
		else
			BreakLockCount = BREAKLOCKCOUNT;

		BreakOld = BreakLevel;
		BreakLevel = 1;
	}

	if (UsedNum < MaxUsed)
	{
		SaveCount--;
		if (SaveCount <= 0)
		{
			SaveCount = SAVECOUNT;
			PopState();
			BreakBlk = 30;
			undoflag = 20;
		}
	}
	else
		SaveCount = SAVECOUNT;

	// Counter
	static int DisplayCounter = 10;
	DisplayCounter--;
	if (DisplayCounter > 0) return;
	DisplayCounter = 10;

	if (undoflag > 0)
	{
		printf("\rLevel %d, max. %d, shifts %d K, config OK=%d,RE=%d UNDO  ",
							UsedNum, MaxUsed, Attempts, BreakOK, BreakRe);
		undoflag--;
	}
	else
	{
		if (resflag > 0)
		{
			printf("\rLevel %d, max. %d, shifts %d K, config OK=%d,RE=%d RESET  ",
								UsedNum, MaxUsed, Attempts, BreakOK, BreakRe);
			resflag--;
		}
		else
		{
			printf("\rLevel %d, max. %d, shifts %d K, config OK=%d,RE=%d        ",
								UsedNum, MaxUsed, Attempts, BreakOK, BreakRe);
		}
	}

	SaveBoardFile(FALSE);
}

//////////////////////////////////////////////////////////////////////////////
// Load old board

void LoadBoardFile()
{
	// Local variables
	int i,j,n;
	char ch;
	CAtom* atom;
	CField* board;

#define BUFSIZE 4000
	char buf[BUFSIZE];

	// Open and read file
	FILE* file;
	char filename[50];
	for (i = 255; i >= 200; i--)
	{
		sprintf(filename, "Best%d.csv", i);
		file = fopen(filename, "rb");
		if (file != NULL)
		{
			n = fread(buf, 1, BUFSIZE, file);
			fclose(file);
			if (n > 500) break;
		}
	}
	if (i < 200) return;
	buf[n] = 0;
	n = 0;

	// Load board
	for (int y = 0; y < HEIGHT; y++)
	{
		for (int x = 0; x < WIDTH; x++)
		{
			// skip invalid characters
			do {
				ch = buf[n++];
				if (ch == 0)
				{
					n--;
					break;
				}
				ch -= '0';
			} while ((ch < 0) || (ch > 9));

			// load block number
			i = 0;
			do {
				i = i*10 + ch;
				ch = buf[n++];
				if (ch == 0)
				{
					n--;
					break;
				}
				ch -= '0';
			} while ((ch >= 0) && (ch <= 9));
			i--;
			if ((i < -1) || (i >= BLOCKS)) i = 0;

			// load spin
			j = 0;
			ch = buf[n++];
			if (ch == 0)
				n--;
			else
				j = ch - '0';
			if ((j < 0) || (j > 3)) j = 0;

			// prepare pointers
			if (i < 0) continue;
			i = i*4 + j;
			atom = &Atom[i];
			board = &Board[y*WIDTH+x];

			// Check if field is empty and atom unused
			if ((board->Edge & EDGE_HINT) == 0)
			{
				if ((board->Atom == NULL) &&
					(Used[atom->Block] == NULL))
				{
					board->Atom = atom;
					Used[atom->Block] = board;

					if ((board->Edge & EDGE_UP) == 0)
					{
						board[-WIDTH].Color[DIR_DOWN] = atom->Color[DIR_UP];
						board[-WIDTH].ColMask[DIR_DOWN] = 0xff;
					}

					if ((board->Edge & EDGE_RIGHT) == 0)
					{
						board[1].Color[DIR_LEFT] = atom->Color[DIR_RIGHT];
						board[1].ColMask[DIR_LEFT] = 0xff;
					}

					if ((board->Edge & EDGE_DOWN) == 0)
					{
						board[WIDTH].Color[DIR_UP] = atom->Color[DIR_DOWN];
						board[WIDTH].ColMask[DIR_UP] = 0xff;
					}

					if ((board->Edge & EDGE_LEFT) == 0)
					{
						board[-1].Color[DIR_RIGHT] = atom->Color[DIR_LEFT];
						board[-1].ColMask[DIR_RIGHT] = 0xff;
					}

					UsedNum++;
				}
				else
				{
					printf("Read error from %s, bad block at [%d,%d]!\n", filename, y+1, x+1);
				}
			}
		}
	}
	MaxUsed = UsedNum;
	PushState();
}

//////////////////////////////////////////////////////////////////////////////
// Main function

int main(int argc, char* argv[])
{
	// Initialize random generator
	RandInit();

	// Initialize board
	BoardInit();

	// Generate atoms
	GenAtom();

	// Load old board
	LoadBoardFile();

	// Solve puzzle
	Attack();

	// Save result
	SaveBest();

	return 0;
}
