//////////////////////////////////////////////////////////////////////////////
//  Eternity II Monte Carlo puzzle solver, v1.0, (c) 2007 Miroslav Nemecek
//////////////////////////////////////////////////////////////////////////////

#pragma warning(disable : 4996)

#define		USEHINTS				// flag, use hints

//////////////////////////////////////////////////////////////////////////////
// Constants

#define		WIDTH	16				// width of area
#define		HEIGHT	16				// height of area
#define		BLOCKS	(WIDTH*HEIGHT)	// number of blocks (=256)
#define		COLORS	23				// number of colors (with border)
#define		EDGE	0				// edge color
#define		ATOMNUM	1024			// number of atoms

// Hints
#define		H0BLOCK	138				// hint 0 block (=block 139)
#define		H0SPIN	2				// hint 0 spin
#define		H0POSX	7				// hint 0 position X
#define		H0POSY	8				// hint 0 position Y
#define		H0POS	(H0POSY*WIDTH+H0POSX) // hint 0 position (I,8)

#ifdef USEHINTS
#define		HINTNUM	3				// number of hints

#define		H1BLOCK	180				// hint 1 block (=block 181)
#define		H1SPIN	3				// hint 1 spin
#define		H1POSX	2				// hint 1 position X
#define		H1POSY	13				// hint 1 position Y
#define		H1POS	(H1POSY*WIDTH+H1POSX) // hint 1 position (N,3)

#define		H2BLOCK	254				// hint 2 block (=block 255)
#define		H2SPIN	3				// hint 2 spin
#define		H2POSX	13				// hint 2 position X
#define		H2POSY	2				// hint 2 position Y
#define		H2POS	(H2POSY*WIDTH+H2POSX) // hint 2 position (C,14)
#else // USEHINTS
#define		HINTNUM	1				// number of hints
#endif // USEHINTS

// Direction and turns - from top side to the right (i.e. in clock direction)
enum {
	DIR_UP		= 0,				// UP direction
	DIR_RIGHT	= 1,				// RIGHT direction
	DIR_DOWN	= 2,				// DOWN direction
	DIR_LEFT	= 3,				// LEFT direction
};

//////////////////////////////////////////////////////////////////////////////
// Structures

#pragma pack (push, 4)

// Block (4 bytes)
struct CBlock {
	BYTE	Color[4];				// colors in all directions
};

// Atom (12 B) (on changes update CAtom in Attack.asm)
struct CAtom {
	CAtom*	Next;					// next atom in the list (NULL=none)
									// (it MUST be first entry, hardcoded!)
	BYTE	Block;					// block number
	BYTE	Spin;					// block spin (quarter-turns clockwise)
	WORD	Res;					// ...padding
	BYTE	Color[4];				// colors in all directions
};

// Field (16 bytes) (on changes update CField in Attack.asm)
struct CField {
	CAtom*	Atom;					// atom in this field (NULL=none)
	BYTE	Color[4];				// required colors of this field
	BYTE	ColMask[4];				// color mask of this field
	BYTE	Edge;					// edge flags (see below)
	BYTE	Res[3];					// ...padding
};

#pragma pack (pop)

// Edge flags
enum {
	EDGE_UP		= 1,				// edge in UP direction
	EDGE_RIGHT	= 2,				// edge in RIGHT direction
	EDGE_DOWN	= 4,				// edge in DOWN direction
	EDGE_LEFT	= 8,				// edge in LEFT direction
	EDGE_HINT	= 16,				// field contains hint
};

//////////////////////////////////////////////////////////////////////////////
// Externals for assembler

extern "C" void SaveBest();
extern "C" void DispState();
extern "C" void Attack();

extern "C" DWORD RandSeed;
extern "C" struct CField Board[];
extern "C" struct CField* Used[];
extern "C" int UsedNum;
extern "C" int MaxUsed;
extern "C" struct CAtom Atom[];
extern "C" struct CAtom* AtomList;
extern "C" struct CAtom* AtomLast;
extern "C" int BreakCount;
extern "C" int BreakOK;
extern "C" int BreakRe;
extern "C" int BreakBlk;

//////////////////////////////////////////////////////////////////////////////
// Block color definition

const struct CBlock Block[BLOCKS] = {
	{	1,	2,	0,	0	},			// block 1
	{	1,	3,	0,	0	},			// block 2
	{	4,	2,	0,	0	},			// block 3
	{	2,	4,	0,	0	},			// block 4
	{	5,	1,	0,	1	},			// block 5
	{	6,	4,	0,	1	},			// block 6
	{	7,	1,	0,	1	},			// block 7
	{	7,	8,	0,	1	},			// block 8
	{	9,	2,	0,	1	},			// block 9
	{	10,	3,	0,	1	},			// block 10
	{	11,	4,	0,	1	},			// block 11
	{	12,	3,	0,	1	},			// block 12
	{	12,	8,	0,	1	},			// block 13
	{	13,	3,	0,	1	},			// block 14
	{	6,	1,	0,	4	},			// block 15
	{	14,	2,	0,	4	},			// block 16
	{	15,	8,	0,	4	},			// block 17
	{	16,	8,	0,	4	},			// block 18
	{	10,	4,	0,	4	},			// block 19
	{	11,	4,	0,	4	},			// block 20
	{	17,	3,	0,	4	},			// block 21
	{	18,	1,	0,	4	},			// block 22
	{	18,	8,	0,	4	},			// block 23
	{	19,	1,	0,	4	},			// block 24
	{	13,	1,	0,	4	},			// block 25
	{	5,	4,	0,	2	},			// block 26
	{	5,	2,	0,	2	},			// block 27
	{	6,	2,	0,	2 	},			// block 28
	{	14,	2,	0,	2	},			// block 29
	{	10,	8,	0,	2	},			// block 30
	{	11,	4,	0,	2	},			// block 31
	{	19,	2,	0,	2	},			// block 32
	{	12,	4,	0,	2	},			// block 33
	{	12,	3,	0,	2	},			// block 34
	{	20,	8,	0,	2	},			// block 35
	{	21,	3,	0,	2	},			// block 36
	{	14,	1,	0,	3	},			// block 37
	{	22,	8,	0,	3	},			// block 38
	{	9,	8,	0,	3	},			// block 39
	{	16,	4,	0,	3	},			// block 40
	{	16,	2,	0,	3	},			// block 41
	{	11,	1,	0,	3	},			// block 42
	{	11,	4,	0,	3	},			// block 43
	{	11,	2,	0,	3	},			// block 44
	{	17,	1,	0,	3	},			// block 45
	{	19,	3,	0,	3	},			// block 46
	{	12,	3,	0,	3	},			// block 47
	{	20,	3,	0,	3	},			// block 48
	{	5,	8,	0,	8	},			// block 49
	{	6,	1,	0,	8	},			// block 50
	{	6,	4,	0,	8	},			// block 51
	{	7,	1,	0,	8	},			// block 52
	{	10,	3,	0,	8	},			// block 53
	{	17,	3,	0,	8	},			// block 54
	{	17,	8,	0,	8	},			// block 55
	{	12,	2,	0,	8	},			// block 56
	{	20,	1,	0,	8	},			// block 57
	{	20,	8,	0,	8	},			// block 58
	{	13,	4,	0,	8	},			// block 59
	{	21,	2,	0,	8	},			// block 60
	{	7,	14,	5,	5	},			// block 61
	{	15,	10,	5,	5	},			// block 62
	{	6,	22,	5,	6	},			// block 63
	{	5,	12,	5,	14	},			// block 64
	{	14,	21,	5,	14 	},			// block 65
	{	15,	15,	5,	14 	},			// block 66
	{	9,	6,	5,	14 	},			// block 67
	{	19,	7,	5,	14	},			// block 68
	{	21,	12,	5,	14 	},			// block 69
	{	22,	10,	5,	22 	},			// block 70
	{	10,	18,	5,	22 	},			// block 71
	{	15,	14,	5,	9  	},			// block 72
	{	11,	17,	5,	9  	},			// block 73
	{	19,	11,	5,	9  	},			// block 74
	{	12,	22,	5,	9  	},			// block 75
	{	15,	11,	5,	16 	},			// block 76
	{	16,	11,	5,	16 	},			// block 77
	{	22,	20,	5,	10 	},			// block 78
	{	19,	22,	5,	10 	},			// block 79
	{	20,	13,	5,	10 	},			// block 80
	{	16,	14,	5,	11 	},			// block 81
	{	14,	14,	5,	17 	},			// block 82
	{	9,	17,	5,	17	},			// block 83
	{	14,	16,	5,	18 	},			// block 84
	{	7,	15,	5,	18	},			// block 85
	{	12,	18,	5,	18 	},			// block 86
	{	20,	19,	5,	18 	},			// block 87
	{	5,	13,	5,	19	},			// block 88
	{	7,	21,	5,	19	},			// block 89
	{	17,	20,	5,	19 	},			// block 90
	{	9,	18,	5,	12	},			// block 91
	{	16,	11,	5,	12 	},			// block 92
	{	16,	17,	5,	12 	},			// block 93
	{	17,	13,	5,	12 	},			// block 94
	{	18,	15,	5,	12 	},			// block 95
	{	13,	22,	5,	13 	},			// block 96
	{	17,	16,	5,	21 	},			// block 97
	{	19,	12,	5,	21 	},			// block 98
	{	13,	7,	5,	21	},			// block 99
	{	21,	13,	5,	21 	},			// block 100
	{	18,	11,	6,	6  	},			// block 101
	{	18,	20,	6,	6  	},			// block 102
	{	20,	16,	6,	6  	},			// block 103
	{	21,	7,	6,	6	},			// block 104
	{	17,	11,	6,	14 	},			// block 105
	{	22,	12,	6,	7  	},			// block 106
	{	16,	12,	6,	7  	},			// block 107
	{	17,	11,	6,	7  	},			// block 108
	{	20,	9,	6,	7	},			// block 109
	{	11,	18,	6,	15 	},			// block 110
	{	18,	11,	6,	15 	},			// block 111
	{	19,	16,	6,	22 	},			// block 112
	{	19,	20,	6,	22 	},			// block 113
	{	15,	17,	6,	9  	},			// block 114
	{	10,	18,	6,	9  	},			// block 115
	{	18,	9,	6,	9	},			// block 116
	{	21,	20,	6,	9  	},			// block 117
	{	22,	13,	6,	16 	},			// block 118
	{	20,	18,	6,	10 	},			// block 119
	{	12,	21,	6,	11 	},			// block 120
	{	15,	21,	6,	17 	},			// block 121
	{	7,	20,	6,	19	},			// block 122
	{	15,	16,	6,	19 	},			// block 123
	{	19,	11,	6,	19 	},			// block 124
	{	18,	21,	6,	12 	},			// block 125
	{	13,	11,	6,	12 	},			// block 126
	{	15,	7,	6,	20	},			// block 127
	{	16,	13,	6,	20 	},			// block 128
	{	17,	22,	6,	20 	},			// block 129
	{	19,	12,	6,	20 	},			// block 130
	{	7,	19,	6,	13	},			// block 131
	{	18,	15,	6,	13 	},			// block 132
	{	15,	20,	6,	21 	},			// block 133
	{	9,	17,	6,	21	},			// block 134
	{	17,	22,	6,	21 	},			// block 135
	{	20,	19,	6,	21 	},			// block 136
	{	19,	10,	14,	14	},			// block 137
	{	7,	22,	14,	7  	},			// block 138
	{	7,	9,	14,	7	},			// block 139
	{	7,	18,	14,	7  	},			// block 140
	{	16,	13,	14,	7 	},			// block 141
	{	11,	7,	14,	7  	},			// block 142
	{	20,	18,	14,	7 	},			// block 143
	{	13,	13,	14,	7 	},			// block 144
	{	22,	17,	14,	15	},			// block 145
	{	14,	18,	14,	22	},			// block 146
	{	14,	21,	14,	22	},			// block 147
	{	22,	15,	14,	22	},			// block 148
	{	11,	18,	14,	22	},			// block 149
	{	7,	9,	14,	16	},			// block 150
	{	17,	21,	14,	16	},			// block 151
	{	9,	9,	14,	10	},			// block 152
	{	9,	16,	14,	10 	},			// block 153
	{	21,	20,	14,	10	},			// block 154
	{	10,	10,	14,	17	},			// block 155
	{	10,	18,	14,	17	},			// block 156
	{	21,	10,	14,	17	},			// block 157
	{	10,	20,	14,	19	},			// block 158
	{	12,	7,	14,	19 	},			// block 159
	{	13,	13,	14,	12	},			// block 160
	{	7,	19,	14,	20 	},			// block 161
	{	15,	19,	14,	20	},			// block 162
	{	11,	9,	14,	21 	},			// block 163
	{	17,	20,	14,	21	},			// block 164
	{	22,	17,	7,	15 	},			// block 165
	{	17,	12,	7,	15 	},			// block 166
	{	22,	22,	7,	9  	},			// block 167
	{	9,	11,	7,	16	},			// block 168
	{	16,	13,	7,	16 	},			// block 169
	{	17,	12,	7,	10 	},			// block 170
	{	19,	20,	7,	10 	},			// block 171
	{	13,	9,	7,	10	},			// block 172
	{	11,	11,	7,	11 	},			// block 173
	{	18,	19,	7,	11 	},			// block 174
	{	10,	13,	7,	17 	},			// block 175
	{	10,	16,	7,	18 	},			// block 176
	{	15,	17,	7,	19 	},			// block 177
	{	18,	20,	7,	12 	},			// block 178
	{	12,	11,	7,	12 	},			// block 179
	{	10,	20,	7,	20 	},			// block 180
	{	9,	20,	7,	13	},			// block 181
	{	10,	9,	7,	13	},			// block 182
	{	16,	12,	15,	15	},			// block 183
	{	21,	10,	15,	22	},			// block 184
	{	16,	18,	15,	9 	},			// block 185
	{	12,	12,	15,	9 	},			// block 186
	{	13,	10,	15,	16	},			// block 187
	{	15,	13,	15,	10	},			// block 188
	{	22,	16,	15,	10	},			// block 189
	{	20,	16,	15,	10	},			// block 190
	{	22,	22,	15,	11	},			// block 191
	{	11,	19,	15,	11	},			// block 192
	{	9,	10,	15,	17 	},			// block 193
	{	13,	9,	15,	18 	},			// block 194
	{	13,	21,	15,	18	},			// block 195
	{	21,	11,	15,	18	},			// block 196
	{	9,	21,	15,	19 	},			// block 197
	{	16,	12,	15,	19	},			// block 198
	{	19,	19,	15,	19	},			// block 199
	{	16,	22,	15,	12	},			// block 190
	{	13,	12,	15,	20	},			// block 201
	{	21,	10,	15,	20	},			// block 202
	{	18,	16,	15,	13	},			// block 203
	{	18,	12,	15,	13	},			// block 204
	{	20,	22,	15,	13	},			// block 205
	{	21,	13,	15,	13	},			// block 206
	{	21,	10,	22,	22	},			// block 207
	{	16,	21,	22,	9 	},			// block 208
	{	12,	11,	22,	9 	},			// block 209
	{	9,	16,	22,	10 	},			// block 210
	{	20,	11,	22,	10	},			// block 211
	{	22,	20,	22,	11	},			// block 212
	{	12,	12,	22,	17	},			// block 213
	{	22,	19,	22,	18	},			// block 214
	{	17,	21,	22,	18	},			// block 215
	{	21,	13,	22,	18	},			// block 216
	{	16,	11,	22,	19	},			// block 217
	{	17,	18,	22,	20	},			// block 218
	{	9,	17,	22,	13 	},			// block 219
	{	9,	20,	22,	21 	},			// block 220
	{	13,	13,	22,	21	},			// block 221
	{	13,	21,	22,	21	},			// block 222
	{	18,	21,	9,	9  	},			// block 223
	{	19,	10,	9,	9  	},			// block 224
	{	20,	11,	9,	9  	},			// block 225
	{	16,	11,	9,	10 	},			// block 226
	{	18,	18,	9,	10 	},			// block 227
	{	16,	12,	9,	17 	},			// block 228
	{	10,	21,	9,	19 	},			// block 229
	{	20,	12,	9,	19 	},			// block 230
	{	18,	19,	9,	12 	},			// block 231
	{	18,	13,	9,	12 	},			// block 232
	{	16,	19,	16,	16	},			// block 233
	{	20,	17,	16,	10	},			// block 234
	{	10,	17,	16,	17	},			// block 235
	{	10,	19,	16,	17	},			// block 236
	{	18,	11,	16,	17	},			// block 237
	{	17,	20,	16,	18	},			// block 238
	{	11,	21,	16,	19	},			// block 239
	{	13,	11,	16,	19	},			// block 240
	{	10,	13,	16,	12	},			// block 241
	{	17,	13,	16,	12	},			// block 242
	{	11,	18,	10,	11	},			// block 243
	{	19,	12,	10,	11	},			// block 244
	{	21,	19,	10,	17	},			// block 245
	{	20,	21,	10,	13	},			// block 246
	{	13,	21,	11,	11	},			// block 247
	{	18,	17,	11,	17	},			// block 248
	{	17,	13,	11,	18	},			// block 249
	{	20,	13,	11,	19	},			// block 250
	{	21,	12,	17,	17	},			// block 251
	{	12,	18,	17,	18	},			// block 252
	{	20,	19,	18,	12	},			// block 253
	{	13,	20,	19,	20	},			// block 254
	{	20,	21,	19,	21	},			// block 255
	{	13,	21,	12,	21	},			// block 256
};
