
; ===========================================================================
;        Eternity II puzzle solver, v1.0, (c) 2007 Miroslav Nemecek
; ===========================================================================

			.586
			.model	flat
			.code

; ---------------------------------------------------------------------------
;                               Constants
; ---------------------------------------------------------------------------

COLBITS		EQU		5					; number of bits per color

DIR_UP		EQU		0					; UP direction
DIR_RIGHT	EQU		1					; RIGHT direction
DIR_DOWN	EQU		2					; DOWN direction
DIR_LEFT	EQU		3					; LEFT direction

BRD_WIDTH	EQU		16					; width of board
BRD_HEIGHT	EQU		16					; height of board
BRD_SIZE	EQU		(BRD_WIDTH*BRD_HEIGHT) ; size of board

PERIOD		EQU		1000000				; view period

; ---------------------------------------------------------------------------
;                              Structures
; ---------------------------------------------------------------------------

; --------- Atom (8 bytes) (on change update CAtom in EternityII.h)

CAtom		STRUCT

Block		db		?					; block number
Spin		db		?					; block spin (quarter-turns clockwise)
Weight		dw		?					; atom weight
Color		db		4 dup(?)			; colors in all directions

CAtom		ENDS
		
; --------- Hash list (8 bytes) (on change update CHash in EternityII.h)

CHash		STRUCT

List		dd		?					; pointer to list of atoms
Num			dd		?					; number of atoms in the list

CHash		ENDS

; --------- Field (20 bytes) (on change update CField in EternityII.h)

CField		STRUCT

Atom		dd		?					; saved pointer to atom list (NULL=none)
Color		db		4 dup(?)			; colors in all directions
Next		dd		?					; next field
Prev		dd		?					; previous field
Jump		dd		?					; jump address (or hash list index)

CField		ENDS

OFFW		EQU		(BRD_WIDTH*SIZE CField) ; offset of next row
OFF1		EQU		(SIZE CField)		; offset of next field

; ---------------------------------------------------------------------------
;                              Externals
; ---------------------------------------------------------------------------

extern		_DispState:NEAR
extern		_SaveBest:NEAR
extern		_Board:CField
extern		_StartField:DWORD
extern		_DoReset:DWORD
extern		_Used:byte
extern		_CurrentPos:DWORD
extern		_Hxxxx:CHash, _HExxx:CHash, _HCxxx:CHash
extern		_HxExx:CHash, _HEExx:CHash, _HCExx:CHash
extern		_HxCxx:CHash, _HECxx:CHash, _HCCxx:CHash
extern		_HxxEx:CHash, _HExEx:CHash, _HCxEx:CHash
extern		_HxEEx:CHash, _HEEEx:CHash, _HCEEx:CHash
extern		_HxCEx:CHash, _HECEx:CHash, _HCCEx:CHash
extern		_HxxCx:CHash, _HExCx:CHash, _HCxCx:CHash
extern		_HxECx:CHash, _HEECx:CHash, _HCECx:CHash
extern		_HxCCx:CHash, _HECCx:CHash, _HCCCx:CHash
extern		_HxxxE:CHash, _HExxE:CHash, _HCxxE:CHash
extern		_HxExE:CHash, _HEExE:CHash, _HCExE:CHash
extern		_HxCxE:CHash, _HECxE:CHash, _HCCxE:CHash
extern		_HxxEE:CHash, _HExEE:CHash, _HCxEE:CHash
extern		_HxEEE:CHash, _HEEEE:CHash, _HCEEE:CHash
extern		_HxCEE:CHash, _HECEE:CHash, _HCCEE:CHash
extern		_HxxCE:CHash, _HExCE:CHash, _HCxCE:CHash
extern		_HxECE:CHash, _HEECE:CHash, _HCECE:CHash
extern		_HxCCE:CHash, _HECCE:CHash, _HCCCE:CHash
extern		_HxxxC:CHash, _HExxC:CHash, _HCxxC:CHash
extern		_HxExC:CHash, _HEExC:CHash, _HCExC:CHash
extern		_HxCxC:CHash, _HECxC:CHash, _HCCxC:CHash
extern		_HxxEC:CHash, _HExEC:CHash, _HCxEC:CHash
extern		_HxEEC:CHash, _HEEEC:CHash, _HCEEC:CHash
extern		_HxCEC:CHash, _HECEC:CHash, _HCCEC:CHash
extern		_HxxCC:CHash, _HExCC:CHash, _HCxCC:CHash
extern		_HxECC:CHash, _HEECC:CHash, _HCECC:CHash
extern		_HxCCC:CHash, _HECCC:CHash, _HCCCC:CHash

; ---------------------------------------------------------------------------
;                                    Macros
; ---------------------------------------------------------------------------

; --------- Macro: Start atom without colors (hash=hash list)

DO_COL0		MACRO	hash

			mov		ebx,hash[0].List	; EBX <- list of atoms
			jmp		AttackNext2			; get next entry

			ENDM

; --------- Macro: Start atom with 1 color (hash=hash list,
;			off1=color offset, c1=color index)

DO_COL1		MACRO	hash,off1,c1

			movzx	ecx,[esi+off1].CField.Color[c1] ; ECX <- get color
			mov		ebx,hash[ecx*SIZE CHash].List ; EBX <- list of atoms
			jmp		AttackNext2			; get next entry

			ENDM

; --------- Macro: Start atom with 2 colors (hash=hash list,
;			off1,off2=color offsets, c1,c2=color indexes)

DO_COL2		MACRO	hash,off1,c1,off2,c2

			movzx	ecx,[esi+off2].CField.Color[c2] ; ECX <- get color 2
			shl		ecx,COLBITS			; rotate color left
			or		cl,[esi+off1].CField.Color[c1] ; ECX <- add color 1
			mov		ebx,hash[ecx*SIZE CHash].List ; EBX <- list of atoms
			jmp		AttackNext2			; get next entry

			ENDM

; --------- Macro: Start atom with 3 colors (hash=hash list,
;			off1,off2,off3=color offsets, c1,c2,c3=color indexes)

DO_COL3		MACRO	hash,off1,c1,off2,c2,off3,c3

			movzx	ecx,[esi+off3].CField.Color[c3] ; ECX <- get color 3
			shl		ecx,COLBITS			; rotate color left
			or		cl,[esi+off2].CField.Color[c2] ; ECX <- add color 2
			shl		ecx,COLBITS			; rotate color left
			or		cl,[esi+off1].CField.Color[c1] ; ECX <- add color 1
			mov		ebx,hash[ecx*SIZE CHash].List ; EBX <- list of atoms
			jmp		AttackNext2			; get next entry

			ENDM

; --------- Macro: Start atom with 4 colors (hash=hash list,
;			off1,off2,off3,off4=color offsets, c1,c2,c3,4=color indexes)

DO_COL4		MACRO	hash,off1,c1,off2,c2,off3,c3,off4,c4

			movzx	ecx,[esi+off4].CField.Color[c4] ; ECX <- get color 4
			shl		ecx,COLBITS			; rotate color left
			or		cl,[esi+off3].CField.Color[c3] ; ECX <- add color 3
			shl		ecx,COLBITS			; rotate color left
			or		cl,[esi+off2].CField.Color[c2] ; ECX <- add color 2
			shl		ecx,COLBITS			; rotate color left
			or		cl,[esi+off1].CField.Color[c1] ; ECX <- add color 1
			mov		ebx,hash[ecx*SIZE CHash].List ; EBX <- list of atoms
			jmp		AttackNext2			; get next entry

			ENDM

; ===========================================================================
;                               Solve puzzle
; ===========================================================================

public		_Attack

; --------- Push registers

_Attack:	pusha						; push all registers

; --------- Initialize attack engine

			mov		esi,offset _Board	; ESI <- board
			mov		ecx,BRD_SIZE		; ECX <- board size
AttackInit:	mov		eax,[esi].CField.Jump ; EAX <- hash list index
			mov		eax,[JumpTab+eax*4]	; EAX <- service address
			mov		[esi].CField.Jump,eax ; set service address
			add		esi,SIZE CField		; ESI <- next field
			loop	AttackInit			; initialize next field
			
; --------- Start first field

			mov		ebp,PERIOD			; EBP <- view period
			mov		edi,offset _Used	; EDI <- block used flag
			mov		esi,[_StartField]	; ESI <- start field
			jmp		[esi].CField.Jump	; jump to first field

DoHxxxx:	DO_COL0	_Hxxxx
DoHExxx:	DO_COL0	_HExxx
DoHCxxx:	DO_COL1	_HCxxx,-OFFW,DIR_DOWN
DoHxExx:	DO_COL0	_HxExx
DoHEExx:	DO_COL0	_HEExx
DoHCExx:	DO_COL1	_HCExx,-OFFW,DIR_DOWN
DoHxCxx:	DO_COL1	_HxCxx,+OFF1,DIR_LEFT
DoHECxx:	DO_COL1	_HECxx,+OFF1,DIR_LEFT
DoHCCxx:	DO_COL2	_HCCxx,-OFFW,DIR_DOWN,+OFF1,DIR_LEFT
DoHxxEx:	DO_COL0	_HxxEx
DoHExEx:	DO_COL0	_HExEx
DoHCxEx:	DO_COL1	_HCxEx,-OFFW,DIR_DOWN
DoHxEEx:	DO_COL0	_HxEEx
DoHEEEx:	DO_COL0	_HEEEx
DoHCEEx:	DO_COL1	_HCEEx,-OFFW,DIR_DOWN
DoHxCEx:	DO_COL1	_HxCEx,+OFF1,DIR_LEFT
DoHECEx:	DO_COL1	_HECEx,+OFF1,DIR_LEFT
DoHCCEx:	DO_COL2	_HCCEx,-OFFW,DIR_DOWN,+OFF1,DIR_LEFT
DoHxxCx:	DO_COL1	_HxxCx,+OFFW,DIR_UP
DoHExCx:	DO_COL1	_HExCx,+OFFW,DIR_UP
DoHCxCx:	DO_COL2	_HCxCx,-OFFW,DIR_DOWN,+OFFW,DIR_UP
DoHxECx:	DO_COL1	_HxECx,+OFFW,DIR_UP
DoHEECx:	DO_COL1	_HEECx,+OFFW,DIR_UP
DoHCECx:	DO_COL2	_HCECx,-OFFW,DIR_DOWN,+OFFW,DIR_UP
DoHxCCx:	DO_COL2	_HxCCx,+OFF1,DIR_LEFT,+OFFW,DIR_UP
DoHECCx:	DO_COL2	_HECCx,+OFF1,DIR_LEFT,+OFFW,DIR_UP
DoHCCCx:	DO_COL3	_HCCCx,-OFFW,DIR_DOWN,+OFF1,DIR_LEFT,+OFFW,DIR_UP

DoHxxxE:	DO_COL0	_HxxxE
DoHExxE:	DO_COL0	_HExxE
DoHCxxE:	DO_COL1	_HCxxE,-OFFW,DIR_DOWN
DoHxExE:	DO_COL0	_HxExE
DoHEExE:	DO_COL0	_HEExE
DoHCExE:	DO_COL1	_HCExE,-OFFW,DIR_DOWN
DoHxCxE:	DO_COL1	_HxCxE,+OFF1,DIR_LEFT
DoHECxE:	DO_COL1	_HECxE,+OFF1,DIR_LEFT
DoHCCxE:	DO_COL2	_HCCxE,-OFFW,DIR_DOWN,+OFF1,DIR_LEFT
DoHxxEE:	DO_COL0	_HxxEE
DoHExEE:	DO_COL0	_HExEE
DoHCxEE:	DO_COL1	_HCxEE,-OFFW,DIR_DOWN
DoHxEEE:	DO_COL0	_HxEEE
DoHEEEE:	DO_COL0	_HEEEE
DoHCEEE:	DO_COL1	_HCEEE,-OFFW,DIR_DOWN
DoHxCEE:	DO_COL1	_HxCEE,+OFF1,DIR_LEFT
DoHECEE:	DO_COL1	_HECEE,+OFF1,DIR_LEFT
DoHCCEE:	DO_COL2	_HCCEE,-OFFW,DIR_DOWN,+OFF1,DIR_LEFT
DoHxxCE:	DO_COL1	_HxxCE,+OFFW,DIR_UP
DoHExCE:	DO_COL1	_HExCE,+OFFW,DIR_UP
DoHCxCE:	DO_COL2	_HCxCE,-OFFW,DIR_DOWN,+OFFW,DIR_UP
DoHxECE:	DO_COL1	_HxECE,+OFFW,DIR_UP
DoHEECE:	DO_COL1	_HEECE,+OFFW,DIR_UP
DoHCECE:	DO_COL2	_HCECE,-OFFW,DIR_DOWN,+OFFW,DIR_UP
DoHxCCE:	DO_COL2	_HxCCE,+OFF1,DIR_LEFT,+OFFW,DIR_UP
DoHECCE:	DO_COL2	_HECCE,+OFF1,DIR_LEFT,+OFFW,DIR_UP
DoHCCCE:	DO_COL3	_HCCCE,-OFFW,DIR_DOWN,+OFF1,DIR_LEFT,+OFFW,DIR_UP

DoHxxxC:	DO_COL1	_HxxxC,-OFF1,DIR_RIGHT
DoHExxC:	DO_COL1	_HExxC,-OFF1,DIR_RIGHT
DoHCxxC:	DO_COL2	_HCxxC,-OFFW,DIR_DOWN,-OFF1,DIR_RIGHT
DoHxExC:	DO_COL1	_HxExC,-OFF1,DIR_RIGHT
DoHEExC:	DO_COL1	_HEExC,-OFF1,DIR_RIGHT
DoHCExC:	DO_COL2	_HCExC,-OFFW,DIR_DOWN,-OFF1,DIR_RIGHT
DoHxCxC:	DO_COL2	_HxCxC,+OFF1,DIR_LEFT,-OFF1,DIR_RIGHT
DoHECxC:	DO_COL2	_HECxC,+OFF1,DIR_LEFT,-OFF1,DIR_RIGHT
DoHCCxC:	DO_COL3	_HCCxC,-OFFW,DIR_DOWN,+OFF1,DIR_LEFT,-OFF1,DIR_RIGHT
DoHxxEC:	DO_COL1	_HxxEC,-OFF1,DIR_RIGHT
DoHExEC:	DO_COL1	_HExEC,-OFF1,DIR_RIGHT
DoHCxEC:	DO_COL2	_HCxEC,-OFFW,DIR_DOWN,-OFF1,DIR_RIGHT
DoHxEEC:	DO_COL1	_HxEEC,-OFF1,DIR_RIGHT
DoHEEEC:	DO_COL1	_HEEEC,-OFF1,DIR_RIGHT
DoHCEEC:	DO_COL2	_HCEEC,-OFFW,DIR_DOWN,-OFF1,DIR_RIGHT
DoHxCEC:	DO_COL2	_HxCEC,+OFF1,DIR_LEFT,-OFF1,DIR_RIGHT
DoHECEC:	DO_COL2	_HECEC,+OFF1,DIR_LEFT,-OFF1,DIR_RIGHT
DoHCCEC:	DO_COL3	_HCCEC,-OFFW,DIR_DOWN,+OFF1,DIR_LEFT,-OFF1,DIR_RIGHT
DoHxxCC:	DO_COL2	_HxxCC,+OFFW,DIR_UP,-OFF1,DIR_RIGHT
DoHExCC:	DO_COL2	_HExCC,+OFFW,DIR_UP,-OFF1,DIR_RIGHT
DoHCxCC:	DO_COL3	_HCxCC,-OFFW,DIR_DOWN,+OFFW,DIR_UP,-OFF1,DIR_RIGHT
DoHxECC:	DO_COL2	_HxECC,+OFFW,DIR_UP,-OFF1,DIR_RIGHT
DoHEECC:	DO_COL2	_HEECC,+OFFW,DIR_UP,-OFF1,DIR_RIGHT
DoHCECC:	DO_COL3	_HCECC,-OFFW,DIR_DOWN,+OFFW,DIR_UP,-OFF1,DIR_RIGHT
DoHxCCC:	DO_COL3	_HxCCC,+OFF1,DIR_LEFT,+OFFW,DIR_UP,-OFF1,DIR_RIGHT
DoHECCC:	DO_COL3	_HECCC,+OFF1,DIR_LEFT,+OFFW,DIR_UP,-OFF1,DIR_RIGHT
DoHCCCC:	DO_COL4	_HCCCC,-OFFW,DIR_DOWN,+OFF1,DIR_LEFT,+OFFW,DIR_UP,-OFF1,DIR_RIGHT

; --------- Backtrace to previous field

AttackBack:	mov		esi,[esi].CField.Prev ; ESI <- pointer to previous field
			mov		ebx,[esi].CField.Atom ; EBX <- return pointer to atom list
			mov		ecx,[ebx]			; ECX <- previous atom
			movzx	eax,[ecx].CAtom.Block ; EDX <- block number
			btr		dword ptr [edi],eax	; reset block used flag

; --------- Get next entry from the list

AttackNext:	add		ebx,4				; EBX <- shift pointer to next atom
AttackNext2:mov		ecx,[ebx]			; ECX <- next atom
			jecxz	AttackBack			; no such atom in the list

; --------- Check if block is already used (and set used flag)

			movzx	eax,[ecx].CAtom.Block ; EDX <- block number
			bts		dword ptr [edi],eax	; check+set block used flag
			jc		AttackNext			; block is already used
			
; --------- Store information about this atom

			cmp		[esi].CField.Atom,0	; new maximum?
			mov		eax,dword ptr [ecx].CAtom.Color ; EAX <- colors
			mov		[esi].CField.Atom,ebx ; store pointer to atom list
			mov		dword ptr [esi].CField.Color,eax ; store color to the field
			je		AttackMax			; save new reached maximum

; --------- View period

AttackNext3:dec		ebp					; counting view period
			jz		AttackView			; do view
			
; --------- Go to next field

AttackNext4:mov		esi,[esi].CField.Next ; ESI <- pointer to next field
			jmp		[esi].CField.Jump	; jump to next field

; --------- Save new maximum level

AttackMax:	mov		[_CurrentPos],esi	; save current position
			pusha						; push all registers
			call	_SaveBest			; save maximum
			popa						; pop all registers
			jmp		short AttackNext3	; continue

; --------- View

AttackView:	mov		[_CurrentPos],esi	; save current position
			pusha						; push all registers
			call	_DispState			; display status
			popa						; pop all registers
			cmp		dword ptr [_DoReset],0 ; reset required?
			jne		AttackStop			; do reset
			mov		ebp,PERIOD			; EBP <- view period
			jmp		short AttackNext4
			
; --------- Pop registers

AttackStop:	popa						; pop all registers
			ret

; --------- Table of jump addresses

JumpTab		dd		DoHxxxx, DoHExxx, DoHCxxx
			dd		DoHxExx, DoHEExx, DoHCExx
			dd		DoHxCxx, DoHECxx, DoHCCxx
			dd		DoHxxEx, DoHExEx, DoHCxEx
			dd		DoHxEEx, DoHEEEx, DoHCEEx
			dd		DoHxCEx, DoHECEx, DoHCCEx
			dd		DoHxxCx, DoHExCx, DoHCxCx
			dd		DoHxECx, DoHEECx, DoHCECx
			dd		DoHxCCx, DoHECCx, DoHCCCx

			dd		DoHxxxE, DoHExxE, DoHCxxE
			dd		DoHxExE, DoHEExE, DoHCExE
			dd		DoHxCxE, DoHECxE, DoHCCxE
			dd		DoHxxEE, DoHExEE, DoHCxEE
			dd		DoHxEEE, DoHEEEE, DoHCEEE
			dd		DoHxCEE, DoHECEE, DoHCCEE
			dd		DoHxxCE, DoHExCE, DoHCxCE
			dd		DoHxECE, DoHEECE, DoHCECE
			dd		DoHxCCE, DoHECCE, DoHCCCE
			
			dd		DoHxxxC, DoHExxC, DoHCxxC
			dd		DoHxExC, DoHEExC, DoHCExC
			dd		DoHxCxC, DoHECxC, DoHCCxC
			dd		DoHxxEC, DoHExEC, DoHCxEC
			dd		DoHxEEC, DoHEEEC, DoHCEEC
			dd		DoHxCEC, DoHECEC, DoHCCEC
			dd		DoHxxCC, DoHExCC, DoHCxCC
			dd		DoHxECC, DoHEECC, DoHCECC
			dd		DoHxCCC, DoHECCC, DoHCCCC

			dd		AttackStop

			END
