//////////////////////////////////////////////////////////////////////////////
//       Eternity II puzzle solver, v1.0, (c) 2007 Miroslav Nemecek
//////////////////////////////////////////////////////////////////////////////
// Required memory: 12 MB

//////////////////////////////////////////////////////////////////////////////
// Includes

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include "EternityII.h"

//////////////////////////////////////////////////////////////////////////////
// Variables

// Configuration
int		RandMode = RAND_RANDOM;			// random mode
int		PathMode = PATH_ROWSDOWN;		// path
BOOL	AutoReset = FALSE;					// do auto reset
int		ResetLevel = 256;				// level to do auto reset
int		DisplayUpdate = 1;				// counter to update display (millions)

// Statistic
int		Attempts = 0;					// attempt counter
int		MaxLevel = 1;					// maximum reached level
int		MaxMaxLevel = 1;
BOOL	DoReset = FALSE;				// request to reset
struct	CField* CurrentPos = NULL;		// current position in board
int		DisplayCounter = 1;
#define	RESETMAX	50
int		ResetN[RESETMAX];						// reset counters

int		ResetLev[RESETMAX] = {
	165, 180, 190, 195, 198, 200, 202, 204, 206, 210,
	212, 214, 216, 218, 220, 221, 222, 223, 224, 225,
	226, 227, 228, 229, 230, 231, 232, 233, 234, 235,
	236, 237, 238, 239, 240, 241, 242, 243, 244, 245,
	246, 247, 248, 249, 250, 251, 252, 253, 254, 255
};

//int		ResetLev[RESETMAX] = {
//	182, 190, 196, 198, 200, 202, 204, 206, 208, 210,
//	212, 214, 216, 218, 220, 222, 224, 226, 228, 230,
//	231, 232, 233, 234, 235, 236, 237, 238, 239, 240,
//	241, 242, 243, 244, 245, 246, 247, 248, 249, 250,
//	251, 252, 253, 254, 255, 255, 255, 255, 255, 255
//};

// Random generator
DWORD	RandSeedInit = 0;				// starting random seed
DWORD	RandSeed = 0;					// seed of random generator

// Color statistic
int		ColorNum[COLORS];				// number of colors (24, 48, 50, 64)

// Atom list
struct	CAtom	Atom[ATOMNUM];			// atoms (size 8 KB)

// Puzzle board
struct	CField	Board[BLOCKS];			// puzzle board (size 5 KB)
struct	CField*	StartField;				// start field

// Block used flag bitmap
BYTE	Used[BLOCKS/8];					// block used flag bitmap

// Hint atoms list
struct	CAtom*	HintList[HINTNUM*2];	// list of hint atoms

// Hash buffer (size 
struct	CAtom*	HashBuff[MAXHASH];		// hash buffer (atom lists, size 3.8 MB)

// Hash tables (x=unset, E=edge, C=color, sides Up-Right-Down-Left, size 7.5 MB)
struct CHash Hxxxx[COLNUM0]; struct CHash HExxx[COLNUM0]; struct CHash HCxxx[COLNUM1];
struct CHash HxExx[COLNUM0]; struct CHash HEExx[COLNUM0]; struct CHash HCExx[COLNUM1];
struct CHash HxCxx[COLNUM1]; struct CHash HECxx[COLNUM1]; struct CHash HCCxx[COLNUM2];
struct CHash HxxEx[COLNUM0]; struct CHash HExEx[COLNUM0]; struct CHash HCxEx[COLNUM1];
struct CHash HxEEx[COLNUM0]; struct CHash HEEEx[COLNUM0]; struct CHash HCEEx[COLNUM1];
struct CHash HxCEx[COLNUM1]; struct CHash HECEx[COLNUM1]; struct CHash HCCEx[COLNUM2];
struct CHash HxxCx[COLNUM1]; struct CHash HExCx[COLNUM1]; struct CHash HCxCx[COLNUM2];
struct CHash HxECx[COLNUM1]; struct CHash HEECx[COLNUM1]; struct CHash HCECx[COLNUM2];
struct CHash HxCCx[COLNUM2]; struct CHash HECCx[COLNUM2]; struct CHash HCCCx[COLNUM3];
struct CHash HxxxE[COLNUM0]; struct CHash HExxE[COLNUM0]; struct CHash HCxxE[COLNUM1];
struct CHash HxExE[COLNUM0]; struct CHash HEExE[COLNUM0]; struct CHash HCExE[COLNUM1];
struct CHash HxCxE[COLNUM1]; struct CHash HECxE[COLNUM1]; struct CHash HCCxE[COLNUM2];
struct CHash HxxEE[COLNUM0]; struct CHash HExEE[COLNUM0]; struct CHash HCxEE[COLNUM1];
struct CHash HxEEE[COLNUM0]; struct CHash HEEEE[COLNUM0]; struct CHash HCEEE[COLNUM1];
struct CHash HxCEE[COLNUM1]; struct CHash HECEE[COLNUM1]; struct CHash HCCEE[COLNUM2];
struct CHash HxxCE[COLNUM1]; struct CHash HExCE[COLNUM1]; struct CHash HCxCE[COLNUM2];
struct CHash HxECE[COLNUM1]; struct CHash HEECE[COLNUM1]; struct CHash HCECE[COLNUM2];
struct CHash HxCCE[COLNUM2]; struct CHash HECCE[COLNUM2]; struct CHash HCCCE[COLNUM3];
struct CHash HxxxC[COLNUM1]; struct CHash HExxC[COLNUM1]; struct CHash HCxxC[COLNUM2];
struct CHash HxExC[COLNUM1]; struct CHash HEExC[COLNUM1]; struct CHash HCExC[COLNUM2];
struct CHash HxCxC[COLNUM2]; struct CHash HECxC[COLNUM2]; struct CHash HCCxC[COLNUM3];
struct CHash HxxEC[COLNUM1]; struct CHash HExEC[COLNUM1]; struct CHash HCxEC[COLNUM2];
struct CHash HxEEC[COLNUM1]; struct CHash HEEEC[COLNUM1]; struct CHash HCEEC[COLNUM2];
struct CHash HxCEC[COLNUM2]; struct CHash HECEC[COLNUM2]; struct CHash HCCEC[COLNUM3];
struct CHash HxxCC[COLNUM2]; struct CHash HExCC[COLNUM2]; struct CHash HCxCC[COLNUM3];
struct CHash HxECC[COLNUM2]; struct CHash HEECC[COLNUM2]; struct CHash HCECC[COLNUM3];
struct CHash HxCCC[COLNUM3]; struct CHash HECCC[COLNUM3]; struct CHash HCCCC[COLNUM4];

struct CHash* HashLists[HASHNUM] = {
	Hxxxx, HExxx, HCxxx, HxExx, HEExx, HCExx, HxCxx, HECxx, HCCxx,
	HxxEx, HExEx, HCxEx, HxEEx, HEEEx, HCEEx, HxCEx, HECEx, HCCEx,
	HxxCx, HExCx, HCxCx, HxECx, HEECx, HCECx, HxCCx, HECCx, HCCCx,
	HxxxE, HExxE, HCxxE, HxExE, HEExE, HCExE, HxCxE, HECxE, HCCxE,
	HxxEE, HExEE, HCxEE, HxEEE, HEEEE, HCEEE, HxCEE, HECEE, HCCEE,
	HxxCE, HExCE, HCxCE, HxECE, HEECE, HCECE, HxCCE, HECCE, HCCCE,
	HxxxC, HExxC, HCxxC, HxExC, HEExC, HCExC, HxCxC, HECxC, HCCxC,
	HxxEC, HExEC, HCxEC, HxEEC, HEEEC, HCEEC, HxCEC, HECEC, HCCEC,
	HxxCC, HExCC, HCxCC, HxECC, HEECC, HCECC, HxCCC, HECCC, HCCCC,
};

int HashSizes[HASHNUM] = {			// total sum of sizes = 960496
	COLNUM0, COLNUM0, COLNUM1, COLNUM0, COLNUM0, COLNUM1, COLNUM1, COLNUM1, COLNUM2,
	COLNUM0, COLNUM0, COLNUM1, COLNUM0, COLNUM0, COLNUM1, COLNUM1, COLNUM1, COLNUM2,
	COLNUM1, COLNUM1, COLNUM2, COLNUM1, COLNUM1, COLNUM2, COLNUM2, COLNUM2, COLNUM3,
	COLNUM0, COLNUM0, COLNUM1, COLNUM0, COLNUM0, COLNUM1, COLNUM1, COLNUM1, COLNUM2,
	COLNUM0, COLNUM0, COLNUM1, COLNUM0, COLNUM0, COLNUM1, COLNUM1, COLNUM1, COLNUM2,
	COLNUM1, COLNUM1, COLNUM2, COLNUM1, COLNUM1, COLNUM2, COLNUM2, COLNUM2, COLNUM3,
	COLNUM1, COLNUM1, COLNUM2, COLNUM1, COLNUM1, COLNUM2, COLNUM2, COLNUM2, COLNUM3,
	COLNUM1, COLNUM1, COLNUM2, COLNUM1, COLNUM1, COLNUM2, COLNUM2, COLNUM2, COLNUM3,
	COLNUM2, COLNUM2, COLNUM3, COLNUM2, COLNUM2, COLNUM3, COLNUM3, COLNUM3, COLNUM4,
};

int HashMask[HASHNUM] = {			// total sum of sizes = 960496
	COLMSK0, COLMSK0, COLMSK1, COLMSK0, COLMSK0, COLMSK1, COLMSK1, COLMSK1, COLMSK2,
	COLMSK0, COLMSK0, COLMSK1, COLMSK0, COLMSK0, COLMSK1, COLMSK1, COLMSK1, COLMSK2,
	COLMSK1, COLMSK1, COLMSK2, COLMSK1, COLMSK1, COLMSK2, COLMSK2, COLMSK2, COLMSK3,
	COLMSK0, COLMSK0, COLMSK1, COLMSK0, COLMSK0, COLMSK1, COLMSK1, COLMSK1, COLMSK2,
	COLMSK0, COLMSK0, COLMSK1, COLMSK0, COLMSK0, COLMSK1, COLMSK1, COLMSK1, COLMSK2,
	COLMSK1, COLMSK1, COLMSK2, COLMSK1, COLMSK1, COLMSK2, COLMSK2, COLMSK2, COLMSK3,
	COLMSK1, COLMSK1, COLMSK2, COLMSK1, COLMSK1, COLMSK2, COLMSK2, COLMSK2, COLMSK3,
	COLMSK1, COLMSK1, COLMSK2, COLMSK1, COLMSK1, COLMSK2, COLMSK2, COLMSK2, COLMSK3,
	COLMSK2, COLMSK2, COLMSK3, COLMSK2, COLMSK2, COLMSK3, COLMSK3, COLMSK3, COLMSK4,
};

// Color sorting list
BYTE	ColWeight[COLORS] = {
	0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22};

//////////////////////////////////////////////////////////////////////////////
// Initialize random generator

void RandInit()
{
	// Get system time
	SYSTEMTIME systime;
	::GetLocalTime(&systime);

	// Prepare temporary time value
	DWORD time = systime.wMilliseconds + 137*systime.wSecond + 
						2567*systime.wMinute + 358946*systime.wDay;

	// Initialize random generator
	RandSeed = (time + 486874*time);
}

//////////////////////////////////////////////////////////////////////////////
// Generate random number

int GetRand()
{
	RandSeed = RandSeed*214013 + 2531011;
	return RandSeed;
}

//////////////////////////////////////////////////////////////////////////////
// Generate WORD random number in range 0 to MAX (including)
// Notes: It is more precise random generator than simple modulo operation.

WORD _fastcall GetRandom(WORD max)
{
	WORD res;
	WORD mask;

	// Maximum value = 0
	if (max == 0) return 0;
	
	// Prepare starting value of the mask
	if (max >= (WORD)0x100) res = 0xFFFF; else res = 0xFF;

	// Calculate mask
	do {
		mask = res;
		res >>= 1;
	} while (res >= max);

	// Generate random number
	do {
		RandSeed = RandSeed*214013 + 2531011;
		res = (WORD)((RandSeed >> 16) & mask);
	} while (res > max);

	return res;
}

//////////////////////////////////////////////////////////////////////////////
// Generate atoms

void GenAtom()
{
	// Local variables
	int i,j,k;
	BYTE tmp,col0,col1,col2,col3;
	const CBlock* block = Block;
	CAtom* atom = Atom;

	// Clear color statistic
	for (i = 0; i < COLORS; i++) ColorNum[i] = 0;

	// Cycle through all blocks
	for (i = 0; i < BLOCKS; i++)
	{
		// Load colors of one block
		col0 = block->Color[0];
		col1 = block->Color[1];
		col2 = block->Color[2];
		col3 = block->Color[3];

		// Add to color statistic
		ColorNum[col0]++;
		ColorNum[col1]++;
		ColorNum[col2]++;
		ColorNum[col3]++;

		// Cycle through all spins
		for (j = 0; j < 4; j++)
		{
			k = -1;

			// Hints
			if ((i == H0BLOCK) && (j == H0SPIN))
			{
				k = H0POS;
				HintList[0*2] = atom;
				HintList[0*2+1] = NULL;
				Board[k].Atom = &HintList[0*2];
			}
#ifdef USEHINTS
			if ((i == H1BLOCK) && (j == H1SPIN))
			{
				k = H1POS;
				HintList[1*2] = atom;
				HintList[1*2+1] = NULL;
				Board[k].Atom = &HintList[1*2];
			}
			if ((i == H2BLOCK) && (j == H2SPIN))
			{
				k = H2POS;
				HintList[2*2] = atom;
				HintList[2*2+1] = NULL;
				Board[k].Atom = &HintList[2*2];
			}
#endif // USEHINTS

			if (k >= 0)
			{
				Board[k].Color[0] = col0;
				Board[k].Color[1] = col1;
				Board[k].Color[2] = col2;
				Board[k].Color[3] = col3;
			}

			// Store colors
			atom->Color[0] = col0;
			atom->Color[1] = col1;
			atom->Color[2] = col2;
			atom->Color[3] = col3;

			// Store block and spin
			atom->Block = i;
			atom->Spin = j;

			// Rotate colors
			tmp = col3; col3 = col2; col2 = col1; col1 = col0; col0 = tmp;

			atom++;
		}
		block++;
	}

	// Randomize color weights
	for (i = 100; i > 0; i--)
	{
		j = GetRandom(COLORS-2);
		tmp = ColWeight[j];
		ColWeight[j] = ColWeight[j+1];
		ColWeight[j+1] = tmp;
	}

	// Generate atom weights
	for (i = 0; i < BLOCKS; i++)
	{
		// Load colors of one block
		col0 = Block[i].Color[0];
		col1 = Block[i].Color[1];
		col2 = Block[i].Color[2];
		col3 = Block[i].Color[3];

		for (j = 0; j < 4; j++)
		{
			// Add weight of colors (values 100, 50, 48, 37 + ColWeight)
			j = 500;
			j += 2400 / ColorNum[col0] + ColWeight[col0];
			j -= 2400 / ColorNum[col1] + ColWeight[col1];
			j += 2400 / ColorNum[col2] + ColWeight[col2];
			j -= 2400 / ColorNum[col3] + ColWeight[col3];

			// Set weight of atom
			Atom[i*4+j].Weight = (WORD)j;

			// Rotate colors
			tmp = col3; col3 = col2; col2 = col1; col1 = col0; col0 = tmp;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// Add atom into one hash list

void AddHash(CHash* hash, CAtom* atom)
{
	if (RandMode == RAND_UNSORT)
	{
		hash->List[hash->Num] = atom;
	}
	else
	{
		int pos, i;
		
		// Weight sorted atoms (heavy atoms goes first)
		if (RandMode == RAND_SORT)
		{
			WORD weight = atom->Weight;

			for (pos = 0; pos < hash->Num; pos++)
			{
				if (weight > hash->List[pos]->Weight) break;
			}
		}
		else
		{
			pos = GetRandom(hash->Num);
		}

		for (i = hash->Num; i > pos; i--)
		{
			hash->List[i] = hash->List[i-1];
		}
		hash->List[i] = atom;
	}
	hash->List[hash->Num+1] = NULL;
	hash->Num++;
}

//////////////////////////////////////////////////////////////////////////////
// increase hash table index (return TRUE = break)

BOOL NextHashTable(int& t0, int& t1, int& t2, int& t3)
{
	// change 0 from x to C
	if (t0 == COL_X)
	{
		t0 = COL_C;
		return FALSE;
	}

	// change 0 from C to x, continue with next bit
	if (t0 == COL_C) t0 = COL_X;

	// change 1 from x to C
	if (t1 == COL_X)
	{
		t1 = COL_C;
		return FALSE;
	}

	// change 1 from C to x, continue with next bit
	if (t1 == COL_C) t1 = COL_X;

	// change 2 from x to C
	if (t2 == COL_X)
	{
		t2 = COL_C;
		return FALSE;
	}

	// change 2 from C to x, continue with next bit
	if (t2 == COL_C) t2 = COL_X;

	// change 3 from x to C
	if (t3 == COL_X)
	{
		t3 = COL_C;
		return FALSE;
	}
	return TRUE;
}

//////////////////////////////////////////////////////////////////////////////
// Add all atoms into all hash lists

void AddAllHash(BOOL save)
{
	// Local vairables
	int i,t0,t1,t2,t3,inx;
	BYTE b,col0,col1,col2,col3;
	CAtom* atom = Atom;
	CHash* hash;

	// Cycle through all atoms
	for (i = 0; i < ATOMNUM; i++)
	{
		// Skip hint blocks
		b = atom->Block;
#ifdef USEHINTS
		if ((b != H0BLOCK) && (b != H1BLOCK) && (b != H2BLOCK))
#else // USEHINTS
		if (b != H0BLOCK)
#endif // USEHINTS
		{
			// Load colors of one atom
			col0 = atom->Color[0];
			col1 = atom->Color[1];
			col2 = atom->Color[2];
			col3 = atom->Color[3];

			// Prepare hash color types
			t0 = (col0 == EDGE) ? COL_E : COL_X;
			t1 = (col1 == EDGE) ? COL_E : COL_X;
			t2 = (col2 == EDGE) ? COL_E : COL_X;
			t3 = (col3 == EDGE) ? COL_E : COL_X;

			// store all hash list variants
			for (;;)
			{
				// prepare color index
				inx = 0;
				if (t3 == COL_C) inx = col3;
				if (t2 == COL_C) inx = (inx << COLBITS) + col2;
				if (t1 == COL_C) inx = (inx << COLBITS) + col1;
				if (t0 == COL_C) inx = (inx << COLBITS) + col0;

				// add atom into hash list
				hash = &HashLists[((t3*COLT+t2)*COLT+t1)*COLT+t0][inx];
				if (save)
					AddHash(hash, atom);
				else
					hash->Num++;

				// rotate to next hash table
				if (NextHashTable(t0, t1, t2, t3)) break;
			}
		}
		atom++;
	}
}

//////////////////////////////////////////////////////////////////////////////
// Initialize hash lists

void InitHash()
{
	// Local variables
	int i,j;
	CHash* hash;
	CAtom** list;

	// Clear hash lists
	for (i = 0; i < HASHNUM; i++)
	{
		j = HashSizes[i];
		hash = HashLists[i];
		for (; j > 0; j--)
		{
			hash->List = HashBuff;
			hash->Num = 0;
			hash++;
		}
	}

	// Add all atoms into all hash lists
	AddAllHash(FALSE);

	// Initialize hash pointers
	list = HashBuff;
	for (i = 0; i < HASHNUM; i++)
	{
		j = HashSizes[i];
		hash = HashLists[i];
		for (; j > 0; j--)
		{
			hash->List = list;
			list += hash->Num+1;
			hash->Num = 0;
			hash++;
		}
	}

	// Re-Add all atoms into all hash lists
	AddAllHash(TRUE);
}

//////////////////////////////////////////////////////////////////////////////
// Initialize board

void InitBoard()
{
	// Local variables
	int i,t0,t1,t2,t3,x,y;
#ifdef _DEBUG
	int n = 0;
#endif // _DEBUG
	CField* board;
	CField* next;
	CField* prev;
	const BYTE* strat;

	// Reset maximum level
	MaxLevel = 1;

	// Clear board
	for (i = 0; i < BLOCKS; i++)
	{
#ifdef USEHINTS
		if ((i == H0POS) || (i == H1POS) || (i == H2POS))
#else // USEHINTS
		if (i == H0POS)
#endif // USEHINTS
		{
			Board[i].Jump = HINTFLG;
		}
		else
		{
			Board[i].Atom = NULL;
			Board[i].Jump = -1;
		}
	}

	// prepare pointers to board
	i = StartFields[PathMode];
	board = &Board[i];
	StartField = board;
	prev = board;		
	strat = &Strategy[PathMode*BLOCKS];

	// Trace board
	for (; board->Jump != HINTFLG;)
	{
#ifdef _DEBUG
		// Check if next field is not used
		if ((strat[i] == i) ||
			((Board[strat[i]].Jump >= 0) && 
			(Board[strat[i]].Jump != HINTFLG)))
		{
			printf("Path error, jump from 0x%x to already used field!\n",i);
			for (x = StartFields[PathMode]; x != i;)
			{
				printf("%02X ",x);
				x = strat[x];
			}
			exit(1);
		}
		n++;
#endif // _DEBUG

		// Prepare X and Y coordinates of current field
		x = i % WIDTH;
		y = i / HEIGHT;

		// Get next field
		i = strat[i];

		// Pointer to next field
		next = &Board[i];

		// Initialize this field
		board->Prev = prev;
		board->Next = next;

		// Prepare color type in UP direction
		if (y == 0) t0 = COL_E; else
			if (board[-WIDTH].Jump >= 0) t0 = COL_C; else t0 = COL_X;

		// Prepare color type in RIGHT direction
		if (x == WIDTH-1) t1 = COL_E; else
			if (board[+1].Jump >= 0) t1 = COL_C; else t1 = COL_X;

		// Prepare color type in DOWN direction
		if (y == HEIGHT-1) t2 = COL_E; else
			if (board[+WIDTH].Jump >= 0) t2 = COL_C; else t2 = COL_X;

		// Prepare color type in LEFT direction
		if (x == 0) t3 = COL_E; else
			if (board[-1].Jump >= 0) t3 = COL_C; else t3 = COL_X;

		// Set color type
		board->Jump = ((t3*COLT+t2)*COLT+t1)*COLT+t0;

		// Shift field pointers
		prev = board;
		board = next;
	}

#ifdef _DEBUG
	// Check if all fields are used
	if (n < BLOCKS-HINTNUM)
	{
		printf("Path error, only %d fields are traced! Lost fields: \n",n+HINTNUM);
		for (i = 0; i < BLOCKS; i++)
		{
			if (Board[i].Jump < 0)
			{
				printf("%02X ",i);
			}
		}
		exit(1);
	}
#endif // _DEBUG
}

//////////////////////////////////////////////////////////////////////////////
// Save board

void SaveBoardFile(BOOL best)
{
	int i;

	// Open file
	FILE* file;
	if (best)
	{
		int n = 0;
		for (i = 0; i < BLOCKS; i++)
		{
			if (Board[i].Atom != NULL) n++;
		}
		if (n <= MaxLevel) return;
		MaxLevel = n;
		if (MaxLevel > MaxMaxLevel)
			MaxMaxLevel = MaxLevel;
		else
			return;
		if (MaxLevel < 200) return;

		char filename[50];
		sprintf(filename, "Best%03d.csv", n);
		file = fopen(filename, "w");
	}
	else
	{
		file = fopen("Result.csv", "w");
	}

	if (file == NULL) return;

	// Clear board
	CField* board = CurrentPos;
	if (board != NULL)
	{
		for (;;)
		{
			board = board->Next;
			if ((board == NULL) || 
				(board == &Board[XXXX]) ||
				(board->Atom == NULL)) break;
			board->Atom = NULL;
		}
	}

	// Save board
	i = 0;
	for (int y = 0; y < HEIGHT; y++)
	{
		for (int x = 0; x < WIDTH; x++)
		{
			if (Board[i].Atom == NULL)
				fprintf(file, "  0,0");
			else
			{
				CAtom* atom = Board[i].Atom[0];
				fprintf(file, "%3d,%1d", (int)(atom->Block + 1), (int)atom->Spin);
			}

			if (x < WIDTH-1) fprintf(file, ";");
			i++;
		}
		fprintf(file, "\n");
	}

	// Close file
	fclose(file);
}

//////////////////////////////////////////////////////////////////////////////
// Save best

void SaveBest()
{
//	int m = MaxMaxLevel;
	SaveBoardFile(TRUE);
	if (MaxLevel >= 200)
	{
		SaveBoardFile(FALSE);
	}
}

//////////////////////////////////////////////////////////////////////////////
// Auto reset

void DoAutoReset()
{
	if ((MaxLevel < ResetLevel) && AutoReset)
	{
		int i;
#define BASE  2
		for (i = 0; i < RESETMAX; i++)
		{
			if ((Attempts <= (1<<(i+BASE))) && (MaxLevel >= ResetLev[i])) return;
		}
		if (Attempts > (1<<(i-1+BASE))) return;

		for (i = 0; i < RESETMAX; i++)
		{
			if (Attempts <= (1<<(i+BASE)))
			{
				ResetN[i]++;
				if (ResetN[i] > (ResetN[i+1]*2)) ResetLev[i]--;
				if ((i > 0) && (ResetN[i-1] < ResetN[i]*2)) ResetLev[i-1]++;
				break;
			}
		}

		DoReset = TRUE;
	}
}

//////////////////////////////////////////////////////////////////////////////
// Display status

void DispState()
{
	Attempts += 1;

//	DoAutoReset();

	// Counter
	//DisplayCounter--;
	//if (DisplayCounter > 0) return;
	//DisplayCounter = DisplayUpdate;

	// Local variables
//	int i,n;

	// Count level
	//n = 0;
	//for (i = 0; i < BLOCKS; i++)
	//{
	//	if (Board[i].Atom != NULL) n++;
	//}
	printf("\rLevel %3d, Max %3d, Resets %4d, Tests %4d millions ", MaxLevel,MaxMaxLevel,Attempts,(int)(Attempts*20));

	//static int c = 2;
	//c--;
	//if (c <= 0)
	//{
	//	c = 2;
	//	SaveBoardFile(FALSE);
	//}
}

//////////////////////////////////////////////////////////////////////////////
// Get configuration number

int GetIniInt(char* name, int def)
{
	return ::GetPrivateProfileInt("Global", name, def, ".\\Etern2Rnd.ini");
}

//////////////////////////////////////////////////////////////////////////////
// Main function

int main(int argc, char* argv[])
{
	// Load configuration
	//RandMode = GetIniInt("RandomMode", RandMode);
	//if ((RandMode < 0) || (RandMode >= RAND_NUM)) RandMode = RAND_RANDOM;
	//RandSeed = GetIniInt("RandomSeed", RandSeed);
	PathMode = GetIniInt("PathMode", PathMode);
	if ((PathMode < 0) || (PathMode >= PATH_NUM)) PathMode = PATH_ROWSDOWN;
	//AutoReset = GetIniInt("AutoReset", AutoReset);
	//ResetLevel = GetIniInt("ResetLevel", ResetLevel);
//	DisplayUpdate = GetIniInt("DisplayUpdate", DisplayUpdate);
//	if (DisplayUpdate < 1) DisplayUpdate = 1;

	// Set low priority class
	::SetPriorityClass(::GetCurrentProcess(), BELOW_NORMAL_PRIORITY_CLASS);

	// Initialize random generator
	if (RandSeed == 0) RandInit();

	// Message
	//printf("Starting solving Eternity II puzzle.\n"
//			"RandomMode %d, RandomSeed %d, PathMode %d, AutoReset %d, ResetLevel %d.\n",
//			RandMode, RandSeed, PathMode, AutoReset, ResetLevel
			//);
#ifdef _DEBUG
	printf("DEBUG VERSION\n");
#endif // _DEBUG

	// Generate atoms
	GenAtom();

	// Starting random seed
	RandSeedInit = RandSeed;

	// Initialize hash lists
	InitHash();

	// Initialize board
	InitBoard();

//		Attempts = 0;
	DisplayCounter = DisplayUpdate;

	do
	{
		int i;
		// Clear block used flag bitmap
		for (i = 0; i < BLOCKS/8; i++) Used[i] = 0;

		// Clear board
		for (i = 0; i < BLOCKS; i++)
		{
#ifdef USEHINTS
			if ((i == H0POS) || (i == H1POS) || (i == H2POS))
#else // USEHINTS
			if (i == H0POS)
#endif // USEHINTS
			{}
			else
				Board[i].Atom = NULL;
		}

		// prepare pointers to board
		i = StartFields[PathMode];
		struct CField* board = &Board[i];
		StartField = board;

		// Solve puzzle
//		DoReset = FALSE;
		Attack();

		static int disp = 1;
		disp--;
		if (disp <= 0)
		{
			disp = 1;
			DispState();
			MaxLevel = 1;
		}

		// Randomize hash list
		for (i = 500; i > 0; i--)
		{
			int ii = GetRandom(HASHNUM-1);
			CHash* hash = HashLists[ii];
			int j;
			do { j = GetRand() & HashMask[ii]; } while (j >= HashSizes[ii]);
			hash += j;
			if (hash->Num > 2)
			{
				int j1 = GetRandom(hash->Num-1);
				int j2 = GetRandom(hash->Num-1);
				CAtom* tmp;
				tmp = hash->List[j1];
				hash->List[j1] = hash->List[j2];
				hash->List[j2] = tmp;
			}
		}

	} while (TRUE); //DoReset);

	// Save board
	SaveBest();

	return 0;
}
