
; ===========================================================================
;                        Eternity II Score Solver
; ===========================================================================

			.586
			.model	flat
			.code

USEHINTS	EQU		0					; flag: 1=use hints 1 and 2, 0=no

; ---------------------------------------------------------------------------
;                               Constants
; ---------------------------------------------------------------------------

WDTH		EQU		16					; width of board
HGHT		EQU		16					; height of board
BLOCKS		EQU		(WDTH*HGHT)			; number of blocks (=256)
COLORS		EQU		23					; number of colors (0=edge)
EDGECOL		EQU		0					; edge color
COLBITS		EQU		8					; number of color bits
COLNUM		EQU		256					; number of colors in hash tables
COLNUM0		EQU		1					; size of hash list with no color
COLNUM2		EQU		((COLORS+1)*COLNUM)	; size of hash list with 2 colors (=6144)

SCOREROW	EQU		12					; first row with low score

PERIOD		EQU		10000000			; view period

H0BLOCK		EQU		138					; hint 0 block (=block 139)
H0SPIN		EQU		2					; hint 0 spin
H0POSX		EQU		7					; hint 0 position X
H0POSY		EQU		8					; hint 0 position Y
H0POS		EQU		(H0POSX+H0POSY*WDTH) ; hint 0 position (I,8)
H0COLD		EQU		9					; hint 0 color down
H0COLL		EQU		12					; hint 0 color left
H0COLU		EQU		8					; hint 0 color up
H0COLR		EQU		9					; hint 0 color right

H1BLOCK		EQU		180					; hint 1 block (=block 181)
H1SPIN		EQU		3					; hint 1 spin
H1POSX		EQU		2					; hint 1 position X
H1POSY		EQU		13					; hint 1 position Y
H1POS		EQU		(H1POSX+H1POSY*WDTH) ; hint 1 position (N,3)
H1COLD		EQU		21					; hint 1 color down
H1COLL		EQU		12					; hint 1 color left
H1COLU		EQU		20					; hint 1 color up
H1COLR		EQU		9					; hint 1 color right

H2BLOCK		EQU		254					; hint 2 block (=block 255)
H2SPIN		EQU		3					; hint 2 spin
H2POSX		EQU		13					; hint 2 position X
H2POSY		EQU		2					; hint 2 position Y
H2POS		EQU		(H2POSX+H2POSY*WDTH) ; hint 2 position (C,14)
H2COLD		EQU		22					; hint 2 color down
H2COLL		EQU		20					; hint 2 color left
H2COLU		EQU		22					; hint 2 color up
H2COLR		EQU		18					; hint 2 color right

IF (USEHINTS EQ 1)
HINTNUM		EQU		3					; number of hints
ELSE
HINTNUM		EQU		1					; number of hints
ENDIF

DOWN		EQU		0					; DOWN direction
LEFT		EQU		1					; LEFT direction
UP			EQU		2					; UP direction
RIGHT		EQU		3					; RIGHT direction

; ---------------------------------------------------------------------------
;                              Structures
; ---------------------------------------------------------------------------

; --------- Atom (8 bytes)

CAtom		STRUCT

Block		db		?					; block number
Spin		db		?					; block spin (clockwise)
			dw		?					; ... padding
				UNION
					STRUCT
ColD		db		?					; color down
ColL		db		?					; color left
ColU		db		?					; color up
ColR		db		?					; color right
					ENDS
ColorDW		dd		?					; color in DWORD
				ENDS			
CAtom		ENDS

; --------- Hash entry (8 bytes)

CHash		STRUCT

Atom		dd		?					; pointer to atom (NULL=end)
Score		db		?					; score of this entry
AtomType	db		?					; atom type
Skip		db		?					; number of entries to skip to next atom type
			db		?					; ...padding

CHash		ENDS

; --------- Hash list (8 bytes)

CList		STRUCT

List		dd		?					; list of hash entries
Num			dd		?					; number of atoms in the list

CList		ENDS

; --------- Field (16 bytes)

CField		STRUCT

Hash		dd		?					; pointer into atom hash list
				UNION
					STRUCT
ColD		db		?					; color down
ColL		db		?					; color left
ColU		db		?					; color up
ColR		db		?					; color right
					ENDS
ColorDW		dd		?					; color in DWORD (ALLCOLD = no atom)
				ENDS			
Score		db		?					; score of this field
			db		?,?,?,?,?,?,?		; ...padding
CField		ENDS

HASHSIZE	EQU		SIZE CHash			; size of hash entry
HASHBITS	EQU		3					; size of hash entry in bits

OFFW		EQU		(WDTH*SIZE CField)	; offset of next row
OFF1		EQU		(SIZE CField)		; offset of next field
COLRBIT		EQU		(3*8)				; bit position of RIGHT color

; ---------------------------------------------------------------------------
;                              Externals
; ---------------------------------------------------------------------------

extern		_SaveBest:NEAR
extern		_DispState:NEAR

extern		_Board:CField
extern		_BestBoard:CField
extern		_Used:BYTE
extern		_Score:DWORD
extern		_Score0:DWORD
extern		_MaxScore:DWORD
extern		_RandSeed:DWORD
extern		_DoReset:DWORD;

extern		_Atom:CAtom

extern		_HashH0:CList
IF (USEHINTS EQ 1)
extern		_HashH1:CList
extern		_HashH2:CList
ENDIF

extern		_Hash:CList
extern		_HashR:CList
extern		_HashD:CList
extern		_HashRD:CList

extern		_Hash1:CList
extern		_Hash1R:CList
extern		_Hash1D:CList
extern		_Hash1RD:CList

; ===========================================================================
;                             Macros
; ===========================================================================

; ---------------------------------------------------------------------------
;                       Get atom list from hash table
; ---------------------------------------------------------------------------
; X,Y = field coordinates (constant)
; SRC = pointer to field
; DST = destination register, on input it contains colors of previous field
; DSTH = destination register HIGH (8 bit)
; ---------------------------------------------------------------------------

GetHash		MACRO	X,Y,SRC,DST,DSTH

IF (X EQ 0) AND (Y EQ 0)
			mov		DST,[_Hash].CList.List
ELSEIF (X EQ 0)
			movzx	DST,[SRC-OFFW].CField.ColD
			shl		DST,COLBITS+HASHBITS
 IF (Y EQ HGHT-1)
			mov		DST,[_Hash1D+DST].CList.List
 ELSE
  IF (Y LT SCOREROW)
			mov		DST,[_Hash+DST].CList.List
  ELSE
			mov		DST,[_Hash1+DST].CList.List
  ENDIF
 ENDIF
ELSEIF (Y EQ 0)
			shr		DST,COLRBIT
 IF (X EQ WDTH-1)
			mov		DST,[_HashR+DST*HASHSIZE].CList.List
 ELSE
			mov		DST,[_Hash+DST*HASHSIZE].CList.List
 ENDIF
ELSEIF (X EQ WDTH-1) AND (Y EQ HGHT-1)
			shr		DST,COLRBIT
			mov		DSTH,[SRC-OFFW].CField.ColD
			mov		DST,[_Hash1RD+DST*HASHSIZE].CList.List
ELSEIF (X EQ WDTH-1)
			shr		DST,COLRBIT
			mov		DSTH,[SRC-OFFW].CField.ColD
 IF (Y LT SCOREROW)
			mov		DST,[_HashR+DST*HASHSIZE].CList.List
 ELSE
			mov		DST,[_Hash1R+DST*HASHSIZE].CList.List
 ENDIF
ELSEIF (Y EQ HGHT-1)
			shr		DST,COLRBIT
			mov		DSTH,[SRC-OFFW].CField.ColD
			mov		DST,[_Hash1D+DST*HASHSIZE].CList.List
ELSEIF (X EQ H0POSX+1) AND (Y EQ H0POSY)
			movzx	DST,[SRC-OFFW].CField.ColD
			shl		DST,COLBITS+HASHBITS
 IF (Y LT SCOREROW)
			mov		DST,[_Hash+DST+H0COLR*HASHSIZE].CList.List
 ELSE
			mov		DST,[_Hash1+DST+H0COLR*HASHSIZE].CList.List
 ENDIF
ELSEIF (X EQ H1POSX+1) AND (Y EQ H1POSY) AND (USEHINTS EQ 1)
			movzx	DST,[SRC-OFFW].CField.ColD
			shl		DST,COLBITS+HASHBITS
 IF (Y LT SCOREROW)
			mov		DST,[_Hash+DST+H1COLR*HASHSIZE].CList.List
 ELSE
			mov		DST,[_Hash1+DST+H1COLR*HASHSIZE].CList.List
 ENDIF
ELSEIF (X EQ H2POSX+1) AND (Y EQ H2POSY) AND (USEHINTS EQ 1)
			movzx	DST,[SRC-OFFW].CField.ColD
			shl		DST,COLBITS+HASHBITS
 IF (Y LT SCOREROW)
			mov		DST,[_Hash+DST+H2COLR*HASHSIZE].CList.List
 ELSE
			mov		DST,[_Hash1+DST+H2COLR*HASHSIZE].CList.List
 ENDIF
ELSEIF (X EQ H0POSX) AND (Y EQ H0POSY+1)
			shr		DST,COLRBIT
 IF (Y LT SCOREROW)
			mov		DST,[_Hash+DST*HASHSIZE+H0COLD*HASHSIZE*COLNUM].CList.List
 ELSE
			mov		DST,[_Hash1+DST*HASHSIZE+H0COLD*HASHSIZE*COLNUM].CList.List
 ENDIF
ELSEIF (X EQ H1POSX) AND (Y EQ H1POSY+1) AND (USEHINTS EQ 1)
			shr		DST,COLRBIT
 IF (Y LT SCOREROW)
			mov		DST,[_Hash+DST*HASHSIZE+H1COLD*HASHSIZE*COLNUM].CList.List
 ELSE
			mov		DST,[_Hash1+DST*HASHSIZE+H1COLD*HASHSIZE*COLNUM].CList.List
 ENDIF
ELSEIF (X EQ H2POSX) AND (Y EQ H2POSY+1) AND (USEHINTS EQ 1)
			shr		DST,COLRBIT
 IF (Y LT SCOREROW)
			mov		DST,[_Hash+DST*HASHSIZE+H2COLD*HASHSIZE*COLNUM].CList.List
 ELSE
			mov		DST,[_Hash1+DST*HASHSIZE+H2COLD*HASHSIZE*COLNUM].CList.List
 ENDIF
ELSE
			shr		DST,COLRBIT
			mov		DSTH,[SRC-OFFW].CField.ColD
 IF (Y LT SCOREROW)
			mov		DST,[_Hash+DST*HASHSIZE].CList.List
 ELSE
			mov		DST,[_Hash1+DST*HASHSIZE].CList.List
 ENDIF
ENDIF
			ENDM

; ---------------------------------------------------------------------------
;                 Check colors of atom before placing it
; ---------------------------------------------------------------------------
; X,Y = field coordinates (constant)
; ATOM = atom
; ERR = error address
; ---------------------------------------------------------------------------

CheckCol	MACRO	X,Y,ATOM,ERR

IF (X EQ H0POSX-1) AND (Y EQ H0POSY)
			cmp		[ATOM].CAtom.ColR,H0COLL
			jne		ERR
ELSEIF (X EQ H1POSX-1) AND (Y EQ H1POSY) AND (USEHINTS EQ 1)
			cmp		[ATOM].CAtom.ColR,H1COLL
			jne		ERR
ELSEIF (X EQ H2POSX-1) AND (Y EQ H2POSY) AND (USEHINTS EQ 1)
			cmp		[ATOM].CAtom.ColR,H2COLL
			jne		ERR
ELSEIF (X EQ H0POSX) AND (Y EQ H0POSY-1)
			cmp		[ATOM].CAtom.ColD,H0COLU
			jne		ERR
ELSEIF (X EQ H1POSX) AND (Y EQ H1POSY-1) AND (USEHINTS EQ 1)
			cmp		[ATOM].CAtom.ColD,H1COLU
			jne		ERR
ELSEIF (X EQ H2POSX) AND (Y EQ H2POSY-1) AND (USEHINTS EQ 1)
			cmp		[ATOM].CAtom.ColD,H2COLU
			jne		ERR
ENDIF
			ENDM

; ---------------------------------------------------------------------------
;                          Attack one position
; ---------------------------------------------------------------------------
; X,Y = field coordinates (constant)
; NXT = next start label
; PRV = previous continue label
; STRT = this start label
; CONT = this continue label
; ---------------------------------------------------------------------------
; EBP = view coutner
; ESI = pointer to field
; ECX = atom
; EDX = score
; ---------------------------------------------------------------------------

At1			MACRO	X,Y,NXT,PRV,STRT,CONT

LOCAL		@Skip,@Find,@Back,@Back2,@Max

; --------- START: Shift pointer to next field (-> ESI)
; ECX must contain colors of previous field

STRT:		

IF (X EQ 0) AND (Y EQ 0)
			mov		esi,offset _Board
ELSEIF (X EQ H0POSX+1) AND (Y EQ H0POSY)
			add		esi,2*SIZE CField
ELSEIF (X EQ H1POSX+1) AND (Y EQ H1POSY) AND (USEHINTS EQ 1)
			add		esi,2*SIZE CField
ELSEIF (X EQ H2POSX+1) AND (Y EQ H2POSY) AND (USEHINTS EQ 1)
			add		esi,2*SIZE CField
ELSE
			add		esi,SIZE CField
ENDIF
			
; --------- Get atom list from hash table (-> ECX, ECX must contain colors)

			GetHash	X,Y,esi,ecx,ch
			jmp		@Skip

; --------- CONTINUE: Shift pointer to previous field (-> ESI)

CONT:		

IF (X EQ H0POSX-1) AND (Y EQ H0POSY)
			sub		esi,2*SIZE CField
ELSEIF (X EQ H1POSX-1) AND (Y EQ H1POSY) AND (USEHINTS EQ 1)
			sub		esi,2*SIZE CField
ELSEIF (X EQ H2POSX-1) AND (Y EQ H2POSY) AND (USEHINTS EQ 1)
			sub		esi,2*SIZE CField
ELSE
			sub		esi,SIZE CField
ENDIF

IF (X EQ WDTH-1)
 IF (Y EQ HGHT-1)
ALast:
 ENDIF
ENDIF

; --------- CONT: Remove atom from old position

@Back:		xor		ecx,ecx
			xchg	ecx,[esi].CField.Hash
			mov		eax,[ecx].CHash.Atom
			movzx	eax,[eax].CAtom.Block
			btr		dword ptr [_Used],eax
			movzx	edx,[esi].CField.Score
			sub		[_Score],edx

; ---------	Get next atom

@Find:		add		ecx,SIZE CHash
@Skip:		mov		eax,[ecx].CHash.Atom
			or		eax,eax
			jz		PRV

; --------- Check colors of atom

			CheckCol X,Y,eax,@Find

; --------- Try to lock block usage

			movzx	eax,[eax].CAtom.Block
			bts		dword ptr [_Used],eax
			jc		@Find

; --------- Prepare score (-> EDX)

			movzx	edx,[ecx].CHash.Score
IF (X EQ H0POSX-1) AND (Y EQ H0POSY)
			inc		edx
ELSEIF (X EQ H1POSX-1) AND (Y EQ H1POSY) AND (USEHINTS EQ 1)
			inc		edx
ELSEIF (X EQ H2POSX-1) AND (Y EQ H2POSY) AND (USEHINTS EQ 1)
			inc		edx
ELSEIF (X EQ H0POSX) AND (Y EQ H0POSY-1)
			inc		edx
ELSEIF (X EQ H1POSX) AND (Y EQ H1POSY-1) AND (USEHINTS EQ 1)
			inc		edx
ELSEIF (X EQ H2POSX) AND (Y EQ H2POSY-1) AND (USEHINTS EQ 1)
			inc		edx
ENDIF
			
; --------- Store atom into new position (-> ECX colors)

			mov		[esi].CField.Score,dl
			mov		[esi].CField.Hash,ecx
			mov		ecx,[ecx].CHash.Atom
			mov		ecx,[ecx].CAtom.ColorDW
			mov		[esi].CField.ColorDW,ecx
			
; --------- Increase score

			add		[_Score],edx

; --------- Check score

IF (Y GE SCOREROW)
			mov		edx,[_Score]
IF (Y EQ 12)		; 12: 357..387 (193..208 blocks)
			cmp		edx,(Y-1)*WDTH+Y*(WDTH-1)+X*2+1 - (X+ 3)* 3/16 -  0  ;(X+ 6)* 2/16 -  0		; (X+ 6)* 2/16 -  0
ELSEIF (Y EQ 13)	; 13: 388..418 (209..224 blocks)
			cmp		edx,(Y-1)*WDTH+Y*(WDTH-1)+X*2+1 - (X+ 3)* 3/16 -  3  ;(X+ 2)* 4/16 -  3		; (X+ 3)* 3/16 -  2
ELSEIF (Y EQ 14)	; 14: 419..449 (225..240 blocks)
			cmp		edx,(Y-1)*WDTH+Y*(WDTH-1)+X*2+1 - (X+ 0)* 9/16 -  6  ;(X+ 0)*10/16 -  7		; (X+ 0)*10/16 -  5
ELSE				; 15: 450..480 (241..256 blocks)
			cmp		edx,(Y-1)*WDTH+Y*(WDTH-1)+X*2+1 - (X+ 0)*10/16 - 15  ;(X+ 0)*10/16 - 17		; (X+ 0)*12/16 - 15
ENDIF
			jb		@Back2
ENDIF

; --------- Save new best board

			cmp		edx,[_Score0]
			jbe		@Max
			mov		[_Score0],edx

			push	ecx
			push	esi
		
			mov		esi,offset _Board
			mov		edi,offset _BestBoard
			mov		ecx,BLOCKS*SIZE CField/4
			rep		movsd

			pop		esi
			pop		ecx			

; --------- Save new best score

			cmp		edx,[_MaxScore]
			jbe		@Max
			mov		[_MaxScore],edx

			push	ecx
			push	esi
			push	ebp

			call	_SaveBest

			pop		ebp
			pop		esi
			pop		ecx
			
; --------- Display current state

@Max:		dec		ebp
			jnz		NXT
			
; --------- Display current state

			push	ecx
			push	esi

			call	_DispState

			pop		esi
			pop		ecx

			mov		ebp,PERIOD
			
			cmp		[_DoReset],0
			je		NXT
			jmp		Restart

; --------- Remove atom from old position and go to previous atom

@Back2:		xor		ecx,ecx
			xchg	ecx,[esi].CField.Hash
			mov		eax,[ecx].CHash.Atom
			movzx	eax,[eax].CAtom.Block
			btr		dword ptr [_Used],eax
			movzx	edx,[esi].CField.Score
			sub		[_Score],edx

			movzx	edx,[ecx].CHash.Skip
			lea		ecx,[ecx+edx*SIZE CHash]
			jmp		@Skip
			
			ENDM

; ===========================================================================
;                               Solve puzzle
; ===========================================================================

public		_Attack

; --------- Push registers

_Attack:	pusha						; push all registers

; --------- Prepare counter of view period

			mov		ebp,PERIOD			; EBP <- view period

; --------- Attack positions

Restart:	mov		[_DoReset],0

			prvlab textequ <AStop>
			y = 0
repeat HGHT
			x = 0
repeat WDTH
			strtlab catstr <s>, %x, <_>, %y
			contlab catstr <c>, %x, <_>, %y

IF (x EQ WDTH-1)

 IF (y EQ HGHT-1)
			nxtlab textequ <ALast>
 ELSE
			nxtlab catstr <s>, %0, <_>, %y + 1
 ENDIF

ELSE

 IF (x EQ H0POSX-1) AND (y EQ H0POSY)
			nxtlab catstr <s>, %x + 2, <_>, %y
 ELSEIF (x EQ H1POSX-1) AND (y EQ H1POSY) AND (USEHINTS EQ 1)
			nxtlab catstr <s>, %x + 2, <_>, %y
 ELSEIF (x EQ H2POSX-1) AND (y EQ H2POSY) AND (USEHINTS EQ 1)
			nxtlab catstr <s>, %x + 2, <_>, %y
 ELSE
			nxtlab catstr <s>, %x + 1, <_>, %y
 ENDIF
			
ENDIF


IF (x EQ H0POSX) AND (y EQ H0POSY)
ELSEIF (x EQ H1POSX) AND (y EQ H1POSY) AND (USEHINTS EQ 1)
ELSEIF (x EQ H2POSX) AND (y EQ H2POSY) AND (USEHINTS EQ 1)
ELSE
			At1 x, y, nxtlab, prvlab, strtlab, contlab
			prvlab textequ contlab
ENDIF
			x = x + 1
endm
			y = y + 1
endm

; --------- Pop registers

AStop:		popa						; pop all registers
			ret

			END
