
//////////////////////////////////////////////////////////////////////////////
//                      Eternity II Score Solver
//////////////////////////////////////////////////////////////////////////////

#include "Etern2Score.h"

//////////////////////////////////////////////////////////////////////////////
// Variables

struct CField		Board[BLOCKS];			// puzzle board
struct CField		BestBoard[BLOCKS];		// best board
BYTE				Used[BLOCKS/8];			// block used flag bitmap
int					Score;					// current score
int					Score0;					// display score
int					MaxScore = 0;			// maximum score
int					Tests = 0;				// test counter
DWORD				RandSeed;				// seed of random generator
int					Resets = 0;				// reset counter
#define RESETS 100
int					NextReset = RESETS;		// next reset counter
BOOL				DoReset = FALSE;		// do-reset flag

struct	CAtom		Atom[ATOMNUM];			// atoms

struct	CHash		HashBuff[MAXHASH];		// hash buffer

struct	CList		HashH0[COLNUM0];		// Hint 0
#ifdef USEHINTS
struct	CList		HashH1[COLNUM0];		// Hint 1
struct	CList		HashH2[COLNUM0];		// Hint 2
#endif // USEHINTS

struct	CList		Hash[COLNUM2];			// hash list without edge
struct	CList		HashR[COLNUM2];			// hash list with RIGHT edge
struct	CList		HashD[COLNUM2];			// hash list with DOWN edge
struct	CList		HashRD[COLNUM2];		// hash list with RIGHT+DOWN edge

struct	CList		Hash1[COLNUM2];			// hash list without edge + 1 color
struct	CList		Hash1R[COLNUM2];		// hash list with RIGHT edge + 1 color
struct	CList		Hash1D[COLNUM2];		// hash list with DOWN edge + 1 color
struct	CList		Hash1RD[COLNUM2];		// hash list with RIGHT+DOWN edge + 1 color

#define HashInx(l,u) (l + (u)*COLNUM)		// index in hash table

struct CList* HashLists[HASHNUM] = {
	HashH0,
#ifdef USEHINTS
	HashH1,	HashH2,
#endif // USEHINTS
	Hash,	HashR,	HashD,	HashRD,
	Hash1,	Hash1R,	Hash1D,	Hash1RD,
};

int HashSizes[HASHNUM] = {
	COLNUM0,
#ifdef USEHINTS
	COLNUM0, COLNUM0,
#endif // USEHINTS
	COLNUM2, COLNUM2, COLNUM2, COLNUM2,
	COLNUM2, COLNUM2, COLNUM2, COLNUM2,
};

//////////////////////////////////////////////////////////////////////////////
// Initialize random generator

void RandInit()
{
	// Get system time
	SYSTEMTIME systime;
	::GetLocalTime(&systime);

	// Prepare temporary time value
	DWORD time = systime.wMilliseconds + 137*systime.wSecond + 
						2567*systime.wMinute + 358946*systime.wDay;

	// Initialize random generator
	RandSeed = (time + 486874*time);
}

//////////////////////////////////////////////////////////////////////////////
// Generate WORD random number

WORD GetRand()
{
	RandSeed = RandSeed*214013 + 2531011;
	return (WORD)(RandSeed >> 16);
}

//////////////////////////////////////////////////////////////////////////////
// Generate WORD random number in range 0 to MAX (including)
// Notes: It is more precise random generator than simple modulo operation.

WORD _fastcall GetRandom(WORD max)
{
	WORD res;
	WORD mask;

	// Maximum value = 0
	if (max == 0) return 0;
	
	// Prepare starting value of the mask
	if (max >= (WORD)0x100) res = 0xFFFF; else res = 0xFF;

	// Calculate mask
	do {
		mask = res;
		res >>= 1;
	} while (res >= max);

	// Generate random number
	do {
		res = GetRand() & mask;
	} while (res > max);

	return res;
}

//////////////////////////////////////////////////////////////////////////////
// Add atom into one hash list

void AddHash(BOOL save, struct CList* list, struct CAtom* atom, BYTE score, BYTE type)
{
	if (save)
	{
		int i,pos;
		do { pos = GetRandom(list->Num); }
			while (((pos < list->Num) && (list->List[pos].AtomType < type)) ||
				((pos > 0) && (list->List[pos-1].AtomType > type)));
		for (i = list->Num; i > pos; i--) list->List[i] = list->List[i-1];
		list->List[i].Atom = atom;
		list->List[i].Score = score;
		list->List[i].AtomType = type;
	}
	list->Num++;
}

//////////////////////////////////////////////////////////////////////////////
// Add atom into all hash lists

void AddHashAll(BOOL save, struct CAtom* atom)
{
	// Hints
	if (atom->Block == H0BLOCK)
	{
		if (atom->Spin == H0SPIN) AddHash(save, HashH0, atom, 0, ATOMTYPE_UL);
		return;
	}

#ifdef USEHINTS
	if (atom->Block == H1BLOCK)
	{
		if (atom->Spin == H1SPIN) AddHash(save, HashH1, atom, 0, ATOMTYPE_UL);
		return;
	}

	if (atom->Block == H2BLOCK)
	{
		if (atom->Spin == H2SPIN) AddHash(save, HashH2, atom, 0, ATOMTYPE_UL);
		return;
	}
#endif // USEHINTS

	// Prepare colors
	BYTE cD = atom->ColD;
	BYTE cL = atom->ColL;
	BYTE cU = atom->ColU;
	BYTE cR = atom->ColR;

	// Prepare pointer to hash table
	struct CList* list;
	struct CList* list1;
	if (cD != EDGECOL)
	{
		if (cR != EDGECOL)
		{
			list = Hash;
			list1 = Hash1;
		}
		else
		{
			list = HashR;
			list1 = Hash1R;
		}
	}
	else
	{
		if (cR != EDGECOL)
		{
			list = HashD;
			list1 = Hash1D;
		}
		else
		{
			list = HashRD;
			list1 = Hash1RD;
		}
	}

	// Add atom into lists
	AddHash(save, &list[HashInx(cL,cU)], atom,
			((cL == EDGECOL) ? 0 : 1) + ((cU == EDGECOL) ? 0 : 1), ATOMTYPE_UL);

	AddHash(save, &list1[HashInx(cL,cU)], atom,
			((cL == EDGECOL) ? 0 : 1) + ((cU == EDGECOL) ? 0 : 1), ATOMTYPE_UL);

	for (int i = 1; i < COLORS; i++)
	{
		if ((cL != EDGECOL) && (cL != i))
			AddHash(save, &list1[HashInx(i,cU)], atom, (cU == EDGECOL) ? 0 : 1, ATOMTYPE_U);

		if ((cU != EDGECOL) && (cU != i))
			AddHash(save, &list1[HashInx(cL,i)], atom, (cL == EDGECOL) ? 0 : 1, ATOMTYPE_L);
	}	
}

//////////////////////////////////////////////////////////////////////////////
// Initilize atoms

void AtomInit()
{
	// Local variables
	int i,j,k;
	char cD,cL,cU,cR,tmp;
	BYTE s,t;
	struct CBlock* block = Block;
	struct CAtom* atom = Atom;
	struct CList* list;
	struct CHash* hash;

	// Initialize atom hash lists
	for (i = 0; i < HASHNUM; i++)
	{
		list = HashLists[i];
		for (j = HashSizes[i]; j > 0; j--)
		{
			list->Num = 0;
			list++;
		}
	}

	// Cycle through all blocks
	for (i = 0; i < BLOCKS; i++)
	{
		// Load colors of one block
		cD = block->ColD;
		cL = block->ColL;
		cU = block->ColU;
		cR = block->ColR;

		// Cycle through all spins
		for (j = 0; j < 4; j++)
		{
			// Initialize atom
			atom->Block = i;
			atom->Spin = j;

			// Set colors
			atom->ColD = cD;
			atom->ColL = cL;
			atom->ColU = cU;
			atom->ColR = cR;

			// Add atom into hash tables
			AddHashAll(FALSE, atom);

			// Rotate colors
			tmp = cD; cD = cR; cR = cU; cU = cL; cL = tmp;

			// Next atom
			atom++;
		}

		// Next block
		block++;
	}

	// Setup atom hash lists
	hash = HashBuff;
	for (i = 0; i < HASHNUM; i++)
	{
		list = HashLists[i];
		for (j = HashSizes[i]; j > 0; j--)
		{
			list->List = hash;
			hash += list->Num + 1;
			list->Num = 0;
			if (hash <= &HashBuff[MAXHASH])
			{
				hash[-1].Atom = NULL;
				hash[-1].AtomType = ATOMTYPE_STOP;
			}
			list++;
		}
	}

	// Re-add atoms in hash tables
	if (hash <= &HashBuff[MAXHASH])
	{
		atom = Atom;
		for (i = ATOMNUM; i > 0; i--)
		{
			AddHashAll(TRUE, atom);
			atom++;
		}
	}

	// Initiate "skip" entries
	for (i = 0; i < HASHNUM; i++)
	{
		list = HashLists[i];
		for (j = HashSizes[i]; j > 0; j--)
		{
			k = list->Num;
			hash = &list->List[k-1];
			t = ATOMTYPE_STOP;

			for (; k > 0; k--)
			{
				if (hash->AtomType != t)
				{
					t = hash->AtomType;
					s = 1;
				}
				hash->Skip = s;
				hash--;
				s++;
			}
			list++;
		}
	}

	// print hash buffer size
	i = (int)(hash - HashBuff);
#ifdef _DEBUG
	printf("Hash buffer size: %d\n", i);
#endif
	if (i > MAXHASH)
	{
		printf("Hash buffer overflow!!!\n");
		exit(1);
	}
}

//////////////////////////////////////////////////////////////////////////////
// Initialize one hint

void HintInit(struct CList* list, int pos)
{
	struct CField* board = &Board[pos];
	board->ColorDW = list->List[0].Atom->ColorDW;
	board->Hash = list->List;
	board->Score = 0;
	int block = list->List[0].Atom->Block;
	Used[block/8] |= 1 << (block & 7);
}

//////////////////////////////////////////////////////////////////////////////
// Initialize board

void BoardInit()
{
	// Reset score
	Score = 0;
	Score0 = 0;

	// Clear last block (flag that it is initialized)
	Board[BLOCKS-1].Hash = NULL;
	BestBoard[BLOCKS-1].Hash = NULL;

	// Clear block used flag bitmap
	int i;
	for (i = 0; i < BLOCKS; i++)
	{
		Board[i].Hash = NULL;
		BestBoard[i].Hash = NULL;
	}
	for (i = 0; i < BLOCKS/8; i++) Used[i] = 0;

	// Save hints into board
	HintInit(HashH0, H0POS);
#ifdef USEHINTS
	HintInit(HashH1, H1POS);
	HintInit(HashH2, H2POS);
#endif // USEHINTS
}

//////////////////////////////////////////////////////////////////////////////
// Save board

void SaveBoard(BOOL best)
{
	// Local variables
	FILE* file;
	char filename[50];
	int i;
	struct CAtom* atom;

	// Open file
	if (best)
	{
		if (Score < 440) return;
		sprintf(filename, "Score%03d.csv", Score);
		file = fopen(filename, "w");
	}
	else
		file = fopen("Result.csv", "w");
	if (file == NULL) return;

	// Save block set
	for (i = 0; i < BLOCKS; i++)
	{
		if (BestBoard[i].Hash == NULL)
			fprintf(file, "  0,0");
		else
		{
			atom = BestBoard[i].Hash[0].Atom;
			fprintf(file, "%3d,%1d", (int)(atom->Block+1), (int)atom->Spin);
		}

		if ((i & 0xf) != 15)
			fprintf(file, ";");
		else
			fprintf(file, "\n");
	}

	// Close file
	fclose(file);
}

//////////////////////////////////////////////////////////////////////////////
// Save best board

void SaveBest()
{
	SaveBoard(TRUE);
}

//////////////////////////////////////////////////////////////////////////////
// Display current state

void DispState()
{
	// Display current state
	Tests += 10;
	printf("\rScore: %3d, Best score: %3d, Tests: %d M, Resets: %d ",
				Score0, MaxScore, Tests, Resets);

	int oldscore0 = Score0;
	Score0 = 0;

	// Save current best board
	static int save = 5;
	save--;
	if (save <= 0)
	{
		save = 5;
		SaveBoard(FALSE);
		BestBoard[BLOCKS-1].Hash = NULL;
	}

	// Reset
	NextReset--;
	if (NextReset <= 0)
	{
		if (oldscore0 >= 455)
		{
			NextReset++;
		}
		else
		{
			NextReset = RESETS;
			DoReset = TRUE;
			Resets++;

			AtomInit();
			BoardInit();
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// Main function

int main(int argc, char* argv[])
{
#ifndef _DEBUG
	// Set low priority class (only on Release build)
	::SetPriorityClass(::GetCurrentProcess(), IDLE_PRIORITY_CLASS);
#endif

	// Message
	printf("Eternity II Score Solver\n");

	// Initialize random generator
	RandInit();

	// Initilize atoms
	AtomInit();

	// Initialize board
	BoardInit();

	// Solve puzzle
	Attack();

	return 0;
}
