
; ===========================================================================
;                      Eternity II Block Set Solver
; ===========================================================================

			.586
			.model	flat
			.code

; ---------------------------------------------------------------------------
;                               Constants
; ---------------------------------------------------------------------------

; Compilation flags (defined in ML.EXE command line with switch /Dxxx)
; _DEBUG	EQU								; Flag, debug version
; MINI		EQU								; Flag, use HINT 1 ("mini") puzzle
; MIDI		EQU								; Flag, use HINT 2 ("midi") puzzle
; MAXI		EQU								; Flag, use full ("maxi") puzzle
; USEHINTS	EQU								; flag, use hints 1 and 2 in "maxi" puzzle

IFDEF MINI
WDTH		EQU		6						; width of board
HGHT		EQU		6						; height of board
COLORS		EQU		8						; number of colors (0=edge)
OCOLMIN		EQU		1						; index of beginning outer color
OCOLMAX		EQU		4						; index of ending outer color
ICOLMIN		EQU		5						; index of beginning inner color
ICOLMAX		EQU		7						; index of ending inner color
ENDIF ; MINI

IFDEF MIDI
WDTH		EQU		12						; width of board
HGHT		EQU		6						; height of board
COLORS		EQU		9						; number of colors (0=edge)
OCOLMIN		EQU		1						; index of beginning outer color
OCOLMAX		EQU		4						; index of ending outer color
ICOLMIN		EQU		5						; index of beginning inner color
ICOLMAX		EQU		8						; index of ending inner color
ENDIF ; MIDI

IFDEF MAXI
WDTH		EQU		16						; width of board
HGHT		EQU		16						; height of board
COLORS		EQU		23						; number of colors (0=edge)
OCOLMIN		EQU		1						; index of beginning outer color
OCOLMAX		EQU		5						; index of ending outer color
ICOLMIN		EQU		6						; index of beginning inner color
ICOLMAX		EQU		22						; index of ending inner color
ENDIF ; MAXI

BLOCKS		EQU		WDTH*HGHT				; number of blocks (=36, 72, 256)
EDGECOL		EQU		0						; edge color
CBLOCKS		EQU		4						; corner blocks
EBLOCKSH	EQU		(WDTH-2)				; edge blocks on horizontal sides (=4, 10, 14)
EBLOCKSV	EQU		(HGHT-2)				; edge blocks on vertical sides (=4, 4, 14)
EBLOCKS		EQU		(2*EBLOCKSH+2*EBLOCKSV)	; edge blocks (=16, 28, 56)
CEBLOCK		EQU		(CBLOCKS+EBLOCKS)		; corner+edge blocks
COLBITS		EQU		8						; number of bits per color
COLNUM		EQU		(1 SHL COLBITS)			; number of colors in hash table (=256)
COLNUM2		EQU		(COLORS*COLNUM)			; number of colors in double hash table (=2048, 2048, 5888)

CORNINI		EQU		24						; number of corner initial spins

OCOLNUM		EQU		(OCOLMAX-OCOLMIN+1)		; number of outer colors (=4, 5)
ICOLNUM		EQU		(ICOLMAX-ICOLMIN+1)		; number of inner colors (=3, 17)

; --------- Hints

H0BLOCK		EQU		138						; hint 0 block (=block 139)
H0SPIN		EQU		2						; hint 0 spin (clockwise)
H0POSX		EQU		7						; hint 0 position X
H0POSY		EQU		8						; hint 0 position Y
H0POS		EQU		(H0POSX+H0POSY*WDTH)	; hint 0 position (I,8)

H0COLD		EQU		14						; hint 0 color down (rotated)
H0COLL		EQU		8						; hint 0 color left (rotated)
H0COLU		EQU		9						; hint 0 color up (rotated)
H0COLR		EQU		14						; hint 0 color right (rotated)

H1BLOCK		EQU		180						; hint 1 block (=block 181)
H1SPIN		EQU		3						; hint 1 spin (clockwise)
H1POSX		EQU		2						; hint 1 position X
H1POSY		EQU		13						; hint 1 position Y
H1POS		EQU		(H1POSX+H1POSY*WDTH)	; hint 1 position (N,3)

H1COLD		EQU		19						; hint 1 color down (rotated)
H1COLL		EQU		8						; hint 1 color left (rotated)
H1COLU		EQU		16						; hint 1 color up (rotated)
H1COLR		EQU		14						; hint 1 color right (rotated)

H2BLOCK		EQU		254						; hint 2 block (=block 255)
H2SPIN		EQU		3						; hint 2 spin (clockwise)
H2POSX		EQU		13						; hint 2 position X
H2POSY		EQU		2						; hint 2 position Y
H2POS		EQU		(H2POSX+H2POSY*WDTH)	; hint 2 position (C,14)

H2COLD		EQU		13						; hint 2 color down (rotated)
H2COLL		EQU		16						; hint 2 color left (rotated)
H2COLU		EQU		13						; hint 2 color up (rotated)
H2COLR		EQU		20						; hint 2 color right (rotated)

IFDEF MAXI
IFDEF USEHINTS
HINTNUM		EQU		3						; number of hints
ELSE ; USEHINTS
HINTNUM		EQU		1						; number of hints
ENDIF ; USEHINTS
ELSE ; MAXI
HINTNUM		EQU		0						; number of hints
ENDIF ; MAXI

IFDEF MAXI
USEHINT0	EQU		1						; flag: 1=use hint 0
IFDEF USEHINTS
USEHINT1	EQU		1						; flag: 1=use hint 1
USEHINT2	EQU		1						; flag: 1=use hint 2
ELSE ; USEHINTS
USEHINT1	EQU		0						; flag: 0=don't use hint 1
USEHINT2	EQU		0						; flag: 0=don't use hint 2
ENDIF ; USEHINTS
ELSE ; MAXI
USEHINT0	EQU		0						; flag: 1=don't use hint 0
USEHINT1	EQU		0						; flag: 0=don't use hint 1
USEHINT2	EQU		0						; flag: 0=don't use hint 2
ENDIF ; MAXI

IFDEF _DEBUG
HISTSIZE	EQU		1000000					; history size (comment it to not use history check)
HISTHASH	EQU		10000h					; size of history hash table (64 KB)
ENDIF ; _DEBUG

; --------- Directions and turns - from down side to the right (clockwise)

DIR_DOWN	EQU		0						; DOWN direction
DIR_LEFT	EQU		1						; LEFT direction
DIR_UP		EQU		2						; UP direction
DIR_RIGHT	EQU		3						; RIGHT direction

; ---------------------------------------------------------------------------
;                              Structures
; ---------------------------------------------------------------------------

; --------- Block (size 16 bytes, colors are in inverse direction)

CBlock		STRUCT

			UNION
Color		db		?,?,?,?,?,?,?,?			; colors in all directions
			STRUCT
ColD		db		?						; color in DOWN direction
ColR		db		?						; color in RIGHT direction
ColU		db		?						; color in UP direction
ColL		db		?						; color in LEFT direction

ColD2		db		?						; color in DOWN direction
ColR2		db		?						; color in RIGHT direction
ColU2		db		?						; color in UP direction
ColL2		db		?						; color in LEFT direction
			ENDS
ColorDW		dd		?,?						; colors as DWORD
			ENDS

Flags		db		?						; block flags (see below)
Block		db		?						; block index
Spin0		db		?						; original spin (clockwise)
Spin		db		?						; current spin (clockwise)
Next		dd		?						; next block in hash table (NULL=none)

CBlock		ENDS

BLOCKBITS	EQU		4						; size of block in bits

; --------- Block flags

HINT		EQU		1						; block is hint
LOCKED		EQU		2						; block is locked
USED		EQU		4						; block is used

; --------- Statistic (size 4 bytes)

CStat		STRUCT

			UNION
Num			db		?,?,?,?					; statistic in all directions
			STRUCT
NumD		db		?						; statistic in DOWN direction
NumL		db		?						; statistic in LEFT direction
NumU		db		?						; statistic in UP direction
NumR		db		?						; statistic in RIGHT direction
			ENDS
NumDW		dd		?						; statistic as DWORD
			ENDS
			
CStat		ENDS

; --------- Field (size 8 bytes, colors are in inverse direction)

CField		STRUCT

			UNION
Color		db		?,?,?,?					; colors in all directions
			STRUCT
ColD		db		?						; color in DOWN direction
ColR		db		?						; color in RIGHT direction
ColU		db		?						; color in UP direction
ColL		db		?						; color in LEFT direction
			ENDS
ColorDW		dd		?						; colors as DWORD
			ENDS

Block		dd		?						; pointer to block

CField		ENDS

OFFW		EQU		(WDTH*SIZE CField)		; offset of next row
OFF1		EQU		(SIZE CField)			; offset of next field

; ---------------------------------------------------------------------------
;                              Externals
; ---------------------------------------------------------------------------

extern		_DispState:NEAR					; display state
extern		_SaveBest:NEAR					; save best board
IFDEF _DEBUG
extern		_CheckStat:NEAR					; debug check color statistic
ENDIF ; _DEBUG

extern		_RandSeed:QWORD					; seed of random generator
extern		_Variants:DWORD					; number of variants
extern		_Level:DWORD					; current level
extern		_MaxLevel:DWORD					; maximal level

extern		_Block:CBlock					; blocks
extern		_Stat:CStat						; color statistic
extern		_StatSave:CStat					; saved color statistic
extern		_Repeat:DWORD					; repeat counter
extern		_CheckBlock:DWORD				; blocks for edge check

IFDEF MAXI
extern		_Hint0:DWORD					; hint 0
IFDEF USEHINTS
extern		_Hint1:DWORD					; hint 1
extern		_Hint2:DWORD					; hint 2
ENDIF ; USEHINTS
ENDIF ; MAXI

extern		_CornIni:BYTE					; corner initial spins

extern		_EdgeL:DWORD					; start pointer of left-color edges
extern		_EdgeR:DWORD					; start pointer of right-color edges
extern		_EdgeD:DWORD					; start pointer of double-color edges

extern		_EdgeLNum:BYTE					; number of left-color edges
extern		_EdgeRNum:BYTE					; number of right-color edges
extern		_EdgeDNum:BYTE					; number of double-color edges

extern		_BlockD:DWORD					; start pointer of down-color block
extern		_BlockDL:DWORD					; start pointer of down-left-color block

extern		_BlockDNum:BYTE					; number of down-color blocks
extern		_BlockDLNum:BYTE				; number of down-left-color blocks

IFNDEF MAXI
extern		_BlockC:DWORD					; start pointer of cross blocks
extern		_BlockCNum:BYTE					; number of cross blocks
ENDIF ; MAXI

extern		_EdgeFree:BYTE					; free positions in all directions
extern		_EdgeLMin:CStat					; minimal number of left-color edges
extern		_EdgeRMin:CStat					; minimal number of right-color edges
extern		_EdgeLRMax:CStat				; maximal number of edge pairs
extern		_EdgeLRIni:BYTE					; initial number of edges
extern		_EdgeLRNum:CStat				; current number of edges
extern		_EdgeLFree:CStat				; number of free left-color edge positions
extern		_EdgeRFree:CStat				; number of free right-color edge positions

extern		_BlockDDFree:CStat				; number of free down-color block positions

extern		_Board:CField					; board
extern		_Hash:DWORD						; block hash table without edges
extern		_HashR:DWORD					; block hash table with right edge
extern		_HashD:DWORD					; block hash table with down edge
extern		_HashDR:DWORD					; block hash table with down and right edge
; Hash index: (up + left*COLNUM)

IFDEF _DEBUG
extern		_Hist:BYTE						; solution history (size 64 MB)
extern		_Hist0:BYTE						; current solution
extern		_HistFirst:DWORD				; history hash - first entry (-1 = none, size 256 KB)
extern		_HistNext:DWORD					; history hash - next entry (-1 = none, size 4 MB)
extern		_Similar:DWORD					; number of similar solutions
extern		_HistNum:DWORD					; number of entries in history
ENDIF ; _DEBUG

; ---------------------------------------------------------------------------
;                                  Macros
; ---------------------------------------------------------------------------

; --------- Shift random generator (uses EDX:EAX)

Rand		MACRO

			mov		eax,214013				; EAX <- constant 1
			mul		dword ptr [_RandSeed]	; EAX <- multiple seed LOW
			add		eax,2531011				; add constant 2
			adc		edx,dword ptr [_RandSeed+4] ; add seed HIGH
			mov		dword ptr [_RandSeed],edx ; store new seed LOW
			mov		dword ptr [_RandSeed+4],eax ; store new seed HIGH

			ENDM

; ---------------------------------------------------------------------------
;                      Macro - Get block from hash table
; ---------------------------------------------------------------------------
; X,Y = field coordinates (constant)
; SRC = pointer to field (register)
; DST = destination register, on input it contains colors of previous field
; DSTL = destination register LOW (8 bits)
; ---------------------------------------------------------------------------

GetHash		MACRO	X,Y,SRC,DST,DSTL

IF (X EQ 0) AND (Y EQ 0)
			mov		DST,_Hash[0]
ELSEIF (X EQ 0)
			movzx	DST,[SRC-OFFW].CField.ColD
 IF (Y EQ HGHT-1)
			mov		DST,_HashD[DST*4]
 ELSE
			mov		DST,_Hash[DST*4]
 ENDIF
ELSEIF (Y EQ 0)
			and		DST,0ff00h
 IF (X EQ WDTH-1)
			mov		DST,_HashR[DST*4]
 ELSE
			mov		DST,_Hash[DST*4]
 ENDIF
ELSEIF (X EQ WDTH-1) AND (Y EQ HGHT-1)
			and		DST,0ff00h
			mov		DSTL,[SRC-OFFW].CField.ColD
			mov		DST,_HashDR[DST*4]
ELSEIF (X EQ WDTH-1)
			and		DST,0ff00h
			mov		DSTL,[SRC-OFFW].CField.ColD
			mov		DST,_HashR[DST*4]
ELSEIF (Y EQ HGHT-1)
			and		DST,0ff00h
			mov		DSTL,[SRC-OFFW].CField.ColD
			mov		DST,_HashD[DST*4]
ELSEIF (X EQ H0POSX+1) AND (Y EQ H0POSY) AND (USEHINT0 EQ 1)
			movzx	DST,[SRC-OFFW].CField.ColD
			mov		DST,_Hash[DST*4+H0COLR*4*COLNUM]
ELSEIF (X EQ H1POSX+1) AND (Y EQ H1POSY) AND (USEHINT1 EQ 1)
			movzx	DST,[SRC-OFFW].CField.ColD
			mov		DST,_Hash[DST*4+H1COLR*4*COLNUM]
ELSEIF (X EQ H2POSX+1) AND (Y EQ H2POSY) AND (USEHINT2 EQ 1)
			movzx	DST,[SRC-OFFW].CField.ColD
			mov		DST,_Hash[DST*4+H2COLR*4*COLNUM]
ELSEIF (X EQ H0POSX) AND (Y EQ H0POSY+1) AND (USEHINT0 EQ 1)
			and		DST,0ff00h
			mov		DST,_Hash[DST*4+H0COLD*4]
ELSEIF (X EQ H1POSX) AND (Y EQ H1POSY+1) AND (USEHINT1 EQ 1)
			and		DST,0ff00h
			mov		DST,_Hash[DST*4+H1COLD*4]
ELSEIF (X EQ H2POSX) AND (Y EQ H2POSY+1) AND (USEHINT2 EQ 1)
			and		DST,0ff00h
			mov		DST,_Hash[DST*4+H2COLD*4]
ELSE
			and		DST,0ff00h
			mov		DSTL,[SRC-OFFW].CField.ColD
			mov		DST,_Hash[DST*4]
ENDIF
			ENDM

; ---------------------------------------------------------------------------
;               Macro - Check colors of block before placing it
; ---------------------------------------------------------------------------
; X,Y = field coordinates (constants)
; BLOCK = block (register)
; SPIN = spin (register)
; ERR = error address
; ---------------------------------------------------------------------------

CheckCol	MACRO	X,Y,BLOCK,SPIN,ERR

IF (X EQ H0POSX-1) AND (Y EQ H0POSY) AND (USEHINT0 EQ 1)
			cmp		[BLOCK].CBlock.ColR[SPIN],H0COLL
			jne		ERR
ELSEIF (X EQ H1POSX-1) AND (Y EQ H1POSY) AND (USEHINT1 EQ 1)
			cmp		[BLOCK].CBlock.ColR[SPIN],H1COLL
			jne		ERR
ELSEIF (X EQ H2POSX-1) AND (Y EQ H2POSY) AND (USEHINT2 EQ 1)
			cmp		[BLOCK].CBlock.ColR[SPIN],H2COLL
			jne		ERR
ELSEIF (X EQ H0POSX) AND (Y EQ H0POSY-1) AND (USEHINT0 EQ 1)
			cmp		[BLOCK].CBlock.ColD[SPIN],H0COLU
			jne		ERR
ELSEIF (X EQ H1POSX) AND (Y EQ H1POSY-1) AND (USEHINT1 EQ 1)
			cmp		[BLOCK].CBlock.ColD[SPIN],H1COLU
			jne		ERR
ELSEIF (X EQ H2POSX) AND (Y EQ H2POSY-1) AND (USEHINT2 EQ 1)
			cmp		[BLOCK].CBlock.ColD[SPIN],H2COLU
			jne		ERR
ENDIF
			ENDM

; ---------------------------------------------------------------------------
;                          Macro - solve one position
; ---------------------------------------------------------------------------
; X,Y = field coordinates (constants)
; NXT = next start label
; PRV = previous continue label
; STRT = this start label
; CONT = this continue label
; ---------------------------------------------------------------------------
; ESI = pointer to field
; ---------------------------------------------------------------------------

Solve1		MACRO	X,Y,NXT,PRV,STRT,CONT

LOCAL	@Skip,@Find

; --------- Shift pointer to next field (-> ESI)
; ECX must contain colors of previous field

STRT:		

IF (X EQ 0) AND (Y EQ 0)
			mov		esi,offset _Board
ELSEIF (X EQ H0POSX+1) AND (Y EQ H0POSY) AND (USEHINT0 EQ 1)
			add		esi,2*SIZE CField
ELSEIF (X EQ H1POSX+1) AND (Y EQ H1POSY) AND (USEHINT1 EQ 1)
			add		esi,2*SIZE CField
ELSEIF (X EQ H2POSX+1) AND (Y EQ H2POSY) AND (USEHINT2 EQ 1)
			add		esi,2*SIZE CField
ELSE
			add		esi,SIZE CField
ENDIF
			
; --------- Get block from hash table (-> ECX, ECX must contain colors)

			GetHash	X,Y,esi,ecx,cl
			jmp		@Skip

; --------- Shift pointer to previous field (-> ESI)

CONT:		

IF (X EQ H0POSX-1) AND (Y EQ H0POSY) AND (USEHINT0 EQ 1)
			sub		esi,2*SIZE CField
ELSEIF (X EQ H1POSX-1) AND (Y EQ H1POSY) AND (USEHINT1 EQ 1)
			sub		esi,2*SIZE CField
ELSEIF (X EQ H2POSX-1) AND (Y EQ H2POSY) AND (USEHINT2 EQ 1)
			sub		esi,2*SIZE CField
ELSE
			sub		esi,SIZE CField
ENDIF

; --------- Remove block from old position			

			mov		ecx,[esi].CField.Block
			and		[ecx].CBlock.Flags,NOT USED
			
; ---------	Get next block

@Find:		mov		ecx,[ecx].CBlock.Next
@Skip:		or		ecx,ecx
			jz		PRV

; --------- Check colors of block

			movzx	ebx,[ecx].CBlock.Spin
			CheckCol X,Y,ecx,ebx,@Find

; --------- Set used flag

			test	[ecx].CBlock.Flags,USED
			jnz		@Find
			or		[ecx].CBlock.Flags,USED
			
; --------- Store block into new position (-> ECX colors)

			mov		[esi].CField.Block,ecx
			mov		ecx,dword ptr [ecx].CBlock.Color[ebx]
			mov		[esi].CField.ColorDW,ecx
			
; ---------	Increase current level

			mov		[_Level],X+Y*WDTH+1
			cmp		[_MaxLevel],X+Y*WDTH+1
			jae		NXT
			mov		[_MaxLevel],X+Y*WDTH+1
			pusha
			call	_SaveBest
			popa
			jmp		NXT
			
			ENDM

; ===========================================================================
;                               Solve puzzle
; ===========================================================================
; Solve speed (MAXI version, on AMD 2000+):
;   Corners: 10 millions per second
;   Corners + edges: 100 thousands per second
;   Corners + edges + edge check: 100 per second (it will not be used further)
;   Corners + edges + inner blocks: 1000 per second
; ---------------------------------------------------------------------------

public		_Attack

; --------- Push registers

_Attack:	pusha							; push all registers

; ---------------------------------------------------------------------------
;                             Turn corners
; ---------------------------------------------------------------------------

; --------- Clear color statistic

Corner:		mov		edi,offset _Stat		; EDI <- color statistic
			mov		ecx,(COLORS*SIZE CStat)/4 ; ECX <- size of color statistic
			xor		eax,eax					; EAX <- 0
			rep		stosd					; clear color statistic			
			mov		dword ptr [_EdgeFree],EBLOCKSH+(EBLOCKSV+(EBLOCKSH+EBLOCKSV*256)*256)*256

; --------- Add hints to color statistic

IFDEF MAXI
;			mov		esi,[_Hint0]			; ESI <- hint 0 block
;			movzx	ebx,[esi].CBlock.Spin	; EBX <- spin of block
;			movzx	edx,[esi].CBlock.ColD[ebx] ; EDX <- color in DOWN direction
;			inc		_Stat[edx*SIZE CStat].CStat.NumD ; attach down color
;			movzx	edx,[esi].CBlock.ColL[ebx] ; EDX <- color in LEFT direction
;			inc		_Stat[edx*SIZE CStat].CStat.NumL ; attach left color
;			movzx	edx,[esi].CBlock.ColU[ebx] ; EDX <- color in UP direction
;			inc		_Stat[edx*SIZE CStat].CStat.NumU ; attach upper color
;			movzx	edx,[esi].CBlock.ColR[ebx] ; EDX <- color in RIGHT direction
;			inc		_Stat[edx*SIZE CStat].CStat.NumR ; attach right color

			inc		_Stat[H0COLD*SIZE CStat].CStat.NumD
			inc		_Stat[H0COLL*SIZE CStat].CStat.NumL
			inc		_Stat[H0COLU*SIZE CStat].CStat.NumU
			inc		_Stat[H0COLR*SIZE CStat].CStat.NumR

IFDEF USEHINTS

;			mov		esi,[_Hint1]			; ESI <- hint 1 block
;			movzx	ebx,[esi].CBlock.Spin	; EBX <- spin of block
;			movzx	edx,[esi].CBlock.ColD[ebx] ; EDX <- color in DOWN direction
;			inc		_Stat[edx*SIZE CStat].CStat.NumD ; attach down color
;			movzx	edx,[esi].CBlock.ColL[ebx] ; EDX <- color in LEFT direction
;			inc		_Stat[edx*SIZE CStat].CStat.NumL ; attach left color
;			movzx	edx,[esi].CBlock.ColU[ebx] ; EDX <- color in UP direction
;			inc		_Stat[edx*SIZE CStat].CStat.NumU ; attach upper color
;			movzx	edx,[esi].CBlock.ColR[ebx] ; EDX <- color in RIGHT direction
;			inc		_Stat[edx*SIZE CStat].CStat.NumR ; attach right color

			inc		_Stat[H1COLD*SIZE CStat].CStat.NumD
			inc		_Stat[H1COLL*SIZE CStat].CStat.NumL
			inc		_Stat[H1COLU*SIZE CStat].CStat.NumU
			inc		_Stat[H1COLR*SIZE CStat].CStat.NumR

;			mov		esi,[_Hint2]			; ESI <- hint 2 block
;			movzx	ebx,[esi].CBlock.Spin	; EBX <- spin of block
;			movzx	edx,[esi].CBlock.ColD[ebx] ; EDX <- color in DOWN direction
;			inc		_Stat[edx*SIZE CStat].CStat.NumD ; attach down color
;			movzx	edx,[esi].CBlock.ColL[ebx] ; EDX <- color in LEFT direction
;			inc		_Stat[edx*SIZE CStat].CStat.NumL ; attach left color
;			movzx	edx,[esi].CBlock.ColU[ebx] ; EDX <- color in UP direction
;			inc		_Stat[edx*SIZE CStat].CStat.NumU ; attach upper color
;			movzx	edx,[esi].CBlock.ColR[ebx] ; EDX <- color in RIGHT direction
;			inc		_Stat[edx*SIZE CStat].CStat.NumR ; attach right color

			inc		_Stat[H2COLD*SIZE CStat].CStat.NumD
			inc		_Stat[H2COLL*SIZE CStat].CStat.NumL
			inc		_Stat[H2COLU*SIZE CStat].CStat.NumU
			inc		_Stat[H2COLR*SIZE CStat].CStat.NumR

ENDIF ; USEHINTS
ENDIF ; MAXI

; --------- Prepare corner variant

			Rand							; get random number
			xor		edx,edx					; EDX <- 0
			mov		ecx,CORNINI				; ECX <- number of variants
			div		ecx						; EDX <- corner variant index
			lea		edi,_CornIni[edx*4]		; EDI <- corner variant
			mov		esi,offset _Block		; ESI <- first block

; --------- Prepare new spin of this block (-> ECX)

Corner2:	movzx	ecx,byte ptr [edi]		; ECX <- new spin of the block
			mov		[esi].CBlock.Spin,cl	; set new spin

; --------- Add corner block to color statistic

			movzx	eax,[esi].CBlock.ColR	; EAX <- color in RIGHT direction
			dec		_Stat[eax*SIZE CStat].CStat.Num[ecx] ; attach right color
			inc		ecx						; ECX <- next spin
			movzx	eax,[esi].CBlock.ColU	; EDX <- color in UP direction
			and		cl,3					; ECX <- next spin
			inc		_Stat[eax*SIZE CStat].CStat.Num[ecx] ; attach upper color

; --------- Next corner block

Corner4:	inc		edi						; EDI <- spin of next block
			add		esi,SIZE CBlock			; ESI <- next block
			cmp		esi,offset _Block + CBLOCKS*SIZE CBlock ; is it valid corner?
			jne		Corner2					; serve next corner block

; ---------------------------------------------------------------------------
;                           Prepare number of edges
; ---------------------------------------------------------------------------

; --------- Prepare minimal number of edges

EdgeNum:	xor		edi,edi					; EDI <- 0, current color index
EdgeNum2:	xor		ebx,ebx					; EBX <- 0, direction DOWN
			xor		esi,esi					; ESI <- 0, accumulator of maximal edges
EdgeNum3:	movsx	eax,_Stat[edi*SIZE CStat+OCOLMIN*SIZE CStat].CStat.Num[ebx] ; EAX <- color statistic
			cdq								; EDX:EAX <- color statistic
			and		edx,eax					; EDX <- negative
			not		edx						; EDX <- NOT negative
			and		eax,edx					; EAX <- edge colors in positive direction
			inc		edx						; EDX <- edge colors in negative direction
			mov		_EdgeLMin[edi*SIZE CStat].CStat.Num[ebx],dl ; minimal number of left-color edges
			mov		_EdgeRMin[edi*SIZE CStat].CStat.Num[ebx],al ; minimal number of right-color edges

; --------- Prepare maximal number of edge pairs

			movzx	ecx,_EdgeFree[ebx]		; ECX <- free positions
			sub		ecx,eax					; subtract right-color edges
			sub		ecx,edx					; subtract left-color edges
			js		Corner					; error, not enough free positions, do reset
			shr		cl,1					; ECX <- number of block pairs
			mov		_EdgeLRMax[edi*SIZE CStat].CStat.Num[ebx],cl ; set maximal number of edge pairs
			add		esi,ecx					; add maximal number to accumulator

; --------- Next direction

			inc		ebx						; EBX <- next direction
			cmp		bl,4					; was it last direction?
			jb		EdgeNum3				; prepare next direction

; --------- Prepare initial number of edge pairs

			movzx	eax,_EdgeLNum[edi]		; EAX <- number of left-color edges
			mov		ecx,_EdgeLMin[edi*SIZE CStat].CStat.NumDW ; ECX <- minimal left-color edges
			sub		al,cl					; subtract minimal blocks in DOWN direction
			sub		al,ch					; subtract minimal blocks in LEFT direction
			shr		ecx,16					; shift minimal blocks
			sub		al,cl					; subtract minimal blocks in UP direction
			sub		al,ch					; subtract minimal blocks in RIGHT direction
			js		Corner					; not enough free blocks, do reset

			movzx	ebx,_EdgeRNum[edi]		; EBX <- number of right-color edges
			mov		ecx,_EdgeRMin[edi*SIZE CStat].CStat.NumDW ; ECX <- minimal right-color edges
			sub		bl,cl					; subtract minimal blocks in DOWN direction
			sub		bl,ch					; subtract minimal blocks in LEFT direction
			shr		ecx,16					; shift minimal blocks
			sub		bl,cl					; subtract minimal blocks in UP direction
			sub		bl,ch					; subtract minimal blocks in RIGHT direction

			cmp		eax,ebx					; are pairs balanced?
			jne		Corner					; pairs are not balanced, do reset
			cmp		eax,esi					; check free positions
			ja		Corner					; no free position, do reset
			mov		_EdgeLRIni[edi],al		; set initial number of edge pairs

; --------- Generate number of edge pairs

EdgeNum7:	movzx	ebx,_EdgeLRIni[edi]		; EBX <- number of edges to generate
			Rand							; get random number

			movzx	ecx,_EdgeLRMax[edi*SIZE CStat].CStat.NumD ; ECX <- maximal number of edge pairs
			cmp		ecx,ebx					; check number of edges
			jb		EdgeNum72				; number of edges is OK
			mov		ecx,ebx					; ECX <- limit number of edges
EdgeNum72:	xor		edx,edx					; EDX <- 0
			inc		ecx						; number of edge pairs + 1
			div		ecx						; EDX <- get number of edge pairs
			mov		_EdgeLRNum[edi*SIZE CStat].CStat.NumD,dl ; set new number of edge pairs
			sub		ebx,edx					; sub number of edges

			movzx	ecx,_EdgeLRMax[edi*SIZE CStat].CStat.NumL ; ECX <- maximal number of edge pairs
			cmp		ecx,ebx					; check number of edges
			jb		EdgeNum74				; number of edges is OK
			mov		ecx,ebx					; ECX <- limit number of edges
EdgeNum74:	xor		edx,edx					; EDX <- 0
			inc		ecx						; number of edge pairs + 1
			div		ecx						; EDX <- get number of edge pairs
			mov		_EdgeLRNum[edi*SIZE CStat].CStat.NumL,dl ; set new number of edge pairs
			sub		ebx,edx					; sub number of edges

			movzx	ecx,_EdgeLRMax[edi*SIZE CStat].CStat.NumU ; ECX <- maximal number of edge pairs
			cmp		ecx,ebx					; check number of edges
			jb		EdgeNum76				; number of edges is OK
			mov		ecx,ebx					; ECX <- limit number of edges
EdgeNum76:	xor		edx,edx					; EDX <- 0
			inc		ecx						; number of edge pairs + 1
			div		ecx						; EDX <- get number of edge pairs
			mov		_EdgeLRNum[edi*SIZE CStat].CStat.NumU,dl ; set new number of edge pairs
			sub		ebx,edx					; sub number of edges

			movzx	ecx,_EdgeLRMax[edi*SIZE CStat].CStat.NumR ; ECX <- maximal number of edge pairs
			cmp		ebx,ecx					; check number of edges
			ja		EdgeNum7				; there is not enough free edges
			mov		_EdgeLRNum[edi*SIZE CStat].CStat.NumR,bl ; set new number of edge pairs

; ---------------------------------------------------------------------------
;                                   Turn edges
; ---------------------------------------------------------------------------

; --------- Remove blocks from free positions and prepare number of free edge positions

EdgeTurn:	mov		eax,_EdgeLRNum[edi*SIZE CStat].CStat.NumDW ; EAX <- number of edge pairs
			mov		ecx,_EdgeLMin[edi*SIZE CStat].CStat.NumDW ; ECX <- minimal left-color edges
			mov		edx,_EdgeRMin[edi*SIZE CStat].CStat.NumDW ; ECX <- minimal right-color edges
			add		ecx,eax					; ECX <- number of left-color blocks
			add		edx,eax					; EDX <- number of right-color blocks
			mov		_EdgeLFree[edi*SIZE CStat].CStat.NumDW,ecx ; set number of free left-color blocks
			mov		_EdgeRFree[edi*SIZE CStat].CStat.NumDW,edx ; set number of free right-color blocks
			add		ecx,edx					; ECX <- total number of blocks			
			sub		dword ptr [_EdgeFree],ecx ; decrease free positions

; --------- Prepare to generate left-color blocks

EdgeTurn1:	mov		esi,_EdgeL[edi*4]		; ESI <- first block
			movzx	ecx,_EdgeLNum[edi]		; ECX <- number of blocks
			jecxz	EdgeTurn5				; no left-color block

; --------- Generate new spin of block

EdgeTurn2:	Rand							; get random number
EdgeTurn3:	movzx	ebx,al					; EBX <- random number
			shr		eax,2					; EAX <- shift random number
			jz		EdgeTurn2				; get new random number
			dec		ebx						; preset random spin - 1
			
; --------- Find free spin
			
EdgeTurn4:	inc		ebx						; EBX <- next spin
			and		bl,3					; EBX = next spin
			movzx	edx,_EdgeLFree[edi*SIZE CStat].CStat.Num[ebx] ; EDX <- number of free blocks
			or		edx,edx					; is this spin free?
			jz		EdgeTurn4				; get next spin
			dec		edx						; decrease number of free edges
			mov		_EdgeLFree[edi*SIZE CStat].CStat.Num[ebx],dl ; set new number of free edges
			mov		[esi].CBlock.Spin,bl	; set new spin
			
; --------- Add block to color statistic

			movzx	edx,[esi].CBlock.ColR	; EDX <- color in RIGHT direction
			dec		_Stat[edx*SIZE CStat].CStat.Num[ebx] ; attach right color
			inc		_Stat[edi*SIZE CStat+OCOLMIN*SIZE CStat].CStat.Num[ebx] ; attach left color
			movzx	edx,[esi].CBlock.ColU	; EDX <- color in UPPER direction
			xor		bl,2					; change new direction to upper
			inc		_Stat[edx*SIZE CStat].CStat.Num[ebx] ; attach upper color

; --------- Next block

			add		esi,SIZE CBlock			; ESI <- next block
			loop	EdgeTurn3				; next left-color block

; --------- Prepare to generate right-color blocks

EdgeTurn5:	mov		esi,_EdgeR[edi*4]		; ESI <- first block
			movzx	ecx,_EdgeRNum[edi]		; ECX <- number of blocks
			jecxz	EdgeTurn9				; no right-color block
			jmp		EdgeTurn7

; --------- Generate new spin of block

EdgeTurn6:	Rand							; get random number
EdgeTurn7:	movzx	ebx,al					; EBX <- random number
			shr		eax,2					; EAX <- shift random number
			jz		EdgeTurn6				; get new random number
			dec		ebx						; preset random spin - 1
			
; --------- Find free spin
			
EdgeTurn8:	inc		ebx						; EBX <- next spin
			and		bl,3					; EBX = next spin
			movzx	edx,_EdgeRFree[edi*SIZE CStat].CStat.Num[ebx] ; EDX <- number of free blocks
			or		edx,edx					; is this spin free?
			jz		EdgeTurn8				; get next spin
			dec		edx						; decrease number of free edges
			mov		_EdgeRFree[edi*SIZE CStat].CStat.Num[ebx],dl ; set new number of free edges
			mov		[esi].CBlock.Spin,bl	; set new spin
			
; --------- Add block to color statistic

			movzx	edx,[esi].CBlock.ColL	; EDX <- color in LEFT direction
			dec		_Stat[edi*SIZE CStat+OCOLMIN*SIZE CStat].CStat.Num[ebx] ; attach right color
			inc		_Stat[edx*SIZE CStat].CStat.Num[ebx] ; attach left color
			movzx	edx,[esi].CBlock.ColU	; EDX <- color in UPPER direction
			xor		bl,2					; change new direction to upper
			inc		_Stat[edx*SIZE CStat].CStat.Num[ebx] ; attach upper color

; --------- Next block

			add		esi,SIZE CBlock			; ESI <- next block
			loop	EdgeTurn7				; next left-color block

; --------- Prepare to generate double-color blocks

EdgeTurn9:	mov		esi,_EdgeD[edi*4]		; ESI <- first block
			movzx	ecx,_EdgeDNum[edi]		; ECX <- number of blocks
			jecxz	EdgeTurn13				; no double-color block
			jmp		EdgeTurn11

; --------- Generate new spin of block

EdgeTurn10:	Rand							; get random number
EdgeTurn11:	movzx	ebx,al					; EBX <- random number
			shr		eax,2					; EAX <- shift random number
			jz		EdgeTurn10				; get new random number
			dec		ebx						; preset random spin - 1
			
; --------- Find free spin
			
EdgeTurn12:	inc		ebx						; EBX <- next spin
			and		bl,3					; EBX = next spin
			movzx	edx,_EdgeFree[ebx]		; EDX <- number of free blocks
			or		edx,edx					; is this spin free?
			jz		EdgeTurn12				; get next spin
			dec		edx						; decrease number of free edges
			mov		_EdgeFree[ebx],dl		; set new number of free edges
			mov		[esi].CBlock.Spin,bl	; set new spin
			
; --------- Add block to color statistic

			movzx	edx,[esi].CBlock.ColU	; EDX <- color in UPPER direction
			xor		bl,2					; change new direction to upper
			inc		_Stat[edx*SIZE CStat].CStat.Num[ebx] ; attach upper color

; --------- Next block

			add		esi,SIZE CBlock			; ESI <- next block
			loop	EdgeTurn11				; next double-color block

; --------- Check unbalanced state of edges

EdgeTurn13:	cmp		edi,OCOLNUM-1			; is it last color?
			je		EdgeTurn20				; it is last color
			xor		esi,esi					; ESI <- 0, direction

EdgeTurn14:	xor		ebx,ebx					; EBX <- 0, accumulator for left color
			xor		ecx,ecx					; ECX <- 0, accumulator for right color

			movsx	eax,_Stat[1*SIZE CStat].CStat.Num[esi] ; EAX <- state for color 1
			cdq								; EDX:EAX <- extend EAX
			and		edx,eax					; EDX <- state in right direction
			sub		ecx,edx					; add state to right accumulator
			not		edx						; inver state mask
			and		eax,edx					; EAX <- state in left direction
			add		ebx,eax					; add state to left accumulator

			movsx	eax,_Stat[2*SIZE CStat].CStat.Num[esi] ; EAX <- state for color 2
			cdq								; EDX:EAX <- extend EAX
			and		edx,eax					; EDX <- state in right direction
			sub		ecx,edx					; add state to right accumulator
			not		edx						; inver state mask
			and		eax,edx					; EAX <- state in left direction
			add		ebx,eax					; add state to left accumulator

			movsx	eax,_Stat[3*SIZE CStat].CStat.Num[esi] ; EAX <- state for color 3
			cdq								; EDX:EAX <- extend EAX
			and		edx,eax					; EDX <- state in right direction
			sub		ecx,edx					; add state to right accumulator
			not		edx						; inver state mask
			and		eax,edx					; EAX <- state in left direction
			add		ebx,eax					; add state to left accumulator

			movsx	eax,_Stat[4*SIZE CStat].CStat.Num[esi] ; EAX <- state for color 4
			cdq								; EDX:EAX <- extend EAX
			and		edx,eax					; EDX <- state in right direction
			sub		ecx,edx					; add state to right accumulator
			not		edx						; inver state mask
			and		eax,edx					; EAX <- state in left direction
			add		ebx,eax					; add state to left accumulator
IFDEF MAXI
			movsx	eax,_Stat[5*SIZE CStat].CStat.Num[esi] ; EAX <- state for color 5
			cdq								; EDX:EAX <- extend EAX
			and		edx,eax					; EDX <- state in right direction
			sub		ecx,edx					; add state to right accumulator
			not		edx						; inver state mask
			and		eax,edx					; EAX <- state in left direction
			add		ebx,eax					; add state to left accumulator
ENDIF ; MAXI
			cmp		bl,_EdgeFree[esi]		; check free positions for left direction
			ja		Corner					; not enough free positions
			cmp		cl,_EdgeFree[esi]		; check free positions for right direction
			ja		Corner					; not enough free positions

			inc		esi						; ESI <- increase direction
			cmp		esi,3					; is it valid next direction?
			jbe		EdgeTurn14				; check next direction

; --------- Next color

			inc		edi						; EDI <- increase color index
			jmp		EdgeNum2				; next color
EdgeTurn20:

; ---------------------------------------------------------------------------
;        Check edge block set (it eliminates aprox. 50% of variants)
; ---------------------------------------------------------------------------

; --------- Clear USED flag of edges

EdgeCheck:	mov		esi,offset _Block		; ESI <- pointer to first block
			mov		ecx,CEBLOCK				; ECX <- number of blocks
EdgeCheck1:	and		[esi].CBlock.Flags,NOT USED ; clear used flag
			add		esi,SIZE CBlock			; ESI <- pointer to next block
			loop	EdgeCheck1				; next block
			mov		dword ptr [_EdgeFree],WDTH-1+(HGHT-1+(WDTH-1+(HGHT-1)*256)*256)*256

; --------- Prepare registers

			mov		ebx,3					; EBX <- RIGHT direction
EdgeCheck2:	mov		cl,0					; CL <- 0 old color 1
			xor		edx,edx					; EDX <- 0, current level
			mov		esi,offset _Block		; ESI <- pointer to first block
			
; --------- Check one block

EdgeCheck3:	test	[esi].CBlock.Flags,USED	; is block free?
			jnz		EdgeCheck5				; block is not free
			cmp		bl,[esi].CBlock.Spin	; has it good spin?
			jne		EdgeCheck5				; spin does not meet
			cmp		cl,[esi].CBlock.ColL	; has it good left color?
			jne		EdgeCheck5				; left color does not meet
			
; --------- Check upper color of corner except first corner

			cmp		cl,EDGECOL				; is it corner block?
			jne		EdgeCheck4				; it is not corner block
			cmp		bl,3					; is it first corner (spin 3) ?
			je		EdgeCheck4				; it is first corner, we check it later
			cmp		ch,[esi].CBlock.ColU	; check upper color
			jne		EdgeCheck5				; upper color does not meet
			
; --------- Mark block as used

EdgeCheck4:	or		[esi].CBlock.Flags,USED	; mark block as used
			mov		cl,[esi].CBlock.ColR	; CL <- new old color 1
			mov		_CheckBlock[edx*4],esi	; store block address
			inc		edx						; increase current level
			cmp		dl,_EdgeFree[ebx]		; is it last block in edge?
			je		EdgeCheck8				; last block, check next spin
			mov		esi,offset _Block+CBLOCKS*SIZE CBlock ; ESI <- first edge
			jmp		EdgeCheck3				; check next block

; --------- Increase block pointer

EdgeCheck5:	add		esi,SIZE CBlock			; ESI <- next block
EdgeCheck6:	cmp		esi,offset _Block+CEBLOCK*SIZE CBlock ; valid block?
			jb		EdgeCheck3				; check next block
			
; --------- Block not found, undo levels			
			
			dec		edx						; decrease current level
			js		Corner					; error, reset all
			mov		esi,_CheckBlock[edx*4]	; ESI <- return old pointer
			and		[esi].CBlock.Flags,NOT USED ; clear used flag
			mov		cl,[esi].CBlock.ColL	; CL <- restore old left color
			jmp		EdgeCheck5				; next block

; --------- Next spin

EdgeCheck8:	mov		ch,cl					; CH <- old color 2
			dec		ebx						; decrease spin
			jns		EdgeCheck2				; next spin

; --------- Check corner with spin 3

			cmp		_Block[0*SIZE CBlock].CBlock.Spin,3 ; spin 3?
			jne		EdgeCheck9				; it is not spin 3
			cmp		_Block[0*SIZE CBlock].CBlock.ColU,ch ; check upper color
			jne		Corner					; error, reset all
			jmp		EdgeCheck12				; check is OK
EdgeCheck9:	cmp		_Block[1*SIZE CBlock].CBlock.Spin,3 ; spin 3?
			jne		EdgeCheck10				; it is not spin 3
			cmp		_Block[1*SIZE CBlock].CBlock.ColU,ch ; check upper color
			jne		Corner					; error, reset all
			jmp		EdgeCheck12				; check is OK
EdgeCheck10:cmp		_Block[2*SIZE CBlock].CBlock.Spin,3 ; spin 3?
			jne		EdgeCheck11				; it is not spin 3
			cmp		_Block[2*SIZE CBlock].CBlock.ColU,ch ; check upper color
			jne		Corner					; error, reset all
			jmp		EdgeCheck12				; check is OK
EdgeCheck11:cmp		_Block[3*SIZE CBlock].CBlock.ColU,ch ; check upper color
			jne		Corner					; error, reset all
EdgeCheck12:

; ---------------------------------------------------------------------------
;                            Push current state
; ---------------------------------------------------------------------------

; --------- Push current state

PushState:	mov		esi,offset _Stat		; ESI <- color statistic
			mov		edi,offset _StatSave	; EDI <- saved color statistic
			mov		ecx,COLORS*SIZE CStat/4	; ECX <- size of color statistic
			rep		movsd					; push current state
IFDEF MAXI
			mov		[_Repeat],10000			; initialize repeat counter
ELSE
			mov		[_Repeat],100			; initialize repeat counter
ENDIF

; ---------------------------------------------------------------------------
;                      Turn down-left color blocks
; ---------------------------------------------------------------------------

; --------- Initialize color index

DLTurn:		xor		edi,edi					; EDI <- 0, current color index

; --------- Prepare to generate down-left-color blocks

DLTurn1:	mov		esi,_BlockDL[edi*4]		; ESI <- first block
			movzx	ecx,_BlockDLNum[edi]	; ECX <- number of blocks
			jecxz	DLTurn5					; no down-left-color block

; --------- Prepare new spin of the block

			Rand							; get random number
DLTurn3:	movzx	ebx,al					; EBX <- random number
			shr		eax,2					; EAX <- shift random number
			and		bl,3					; EBX = next spin
			mov		[esi].CBlock.Spin,bl	; set new spin
			
; --------- Add block to color statistic

			movzx	edx,[esi].CBlock.ColD[ebx] ; EDX <- color in DOWN direction
			inc		_Stat[edx*SIZE CStat].CStat.NumD ; attach down color
			movzx	edx,[esi].CBlock.ColL[ebx] ; EDX <- color in LEFT direction
			inc		_Stat[edx*SIZE CStat].CStat.NumL ; attach left color
			movzx	edx,[esi].CBlock.ColU[ebx] ; EDX <- color in UP direction
			inc		_Stat[edx*SIZE CStat].CStat.NumU ; attach upper color
			movzx	edx,[esi].CBlock.ColR[ebx] ; EDX <- color in RIGHT direction
			inc		_Stat[edx*SIZE CStat].CStat.NumR ; attach right color

; --------- Next block

			add		esi,SIZE CBlock			; ESI <- next block
			loop	DLTurn3					; next down-left-color block
DLTurn5:

; ---------------------------------------------------------------------------
;                       Prepare number of inner blocks
; ---------------------------------------------------------------------------

; --------- Prepare minimal number of down-color blocks in DOWN/UP direction

BlockNum:	movzx	ecx,_BlockDNum[edi]		; ECX <- number of down-color blocks
			movsx	eax,_Stat[edi*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumD ; EAX <- DOWN color statistic
			movsx	edx,_Stat[edi*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumU ; EDX <- UP color statistic
			sub		eax,edx					; EAX <- difference DOWN - UP
			cdq								; EDX:EAX <- difference DOWN - UP
			and		edx,eax					; EDX <- negative difference if DOWN < UP
			not		edx						; EDX <- complement
			and		eax,edx					; EAX <- positive difference DOWN - UP
			inc		edx						; EDX <- negative difference UP - DOWN
			mov		_BlockDDFree[edi*SIZE CStat].CStat.NumD,dl ; number of blocks in DOWN direction
			mov		_BlockDDFree[edi*SIZE CStat].CStat.NumU,al ; number of blocks in UP direction
			sub		ecx,edx					; subtract minimal blocks in DOWN direction
			sub		ecx,eax					; subtract minimal blocks in UP direction

; --------- Prepare minimal number of down-color blocks in LEFT/RIGHT direction

			movsx	eax,_Stat[edi*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumL ; EAX <- LEFT color statistic
			movsx	edx,_Stat[edi*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumR ; EDX <- RIGHT color statistic
			sub		eax,edx					; EAX <- difference LEFT - RIGHT
			cdq								; EDX:EAX <- difference LEFT - RIGHT
			and		edx,eax					; EDX <- negative difference if LEFT < RIGHT
			not		edx						; EDX <- complement
			and		eax,edx					; EAX <- positive difference LEFT - RIGHT
			inc		edx						; EDX <- negative difference RIGHT - LEFT
			mov		_BlockDDFree[edi*SIZE CStat].CStat.NumL,dl ; number of blocks in LEFT direction
			mov		_BlockDDFree[edi*SIZE CStat].CStat.NumR,al ; number of blocks in RIGHT direction
			sub		ecx,edx					; subtract minimal blocks in LEFT direction
			sub		ecx,eax					; subtract minimal blocks in RIGHT direction

; --------- Prepare initial number of block pairs

			js		PopState				; error, not enough blocks
			shr		ecx,1					; ECX <- number of block pairs
			jc		PopState				; some block is odd
			
; --------- Generate number of block pairs

			Rand							; get random number
			xor		edx,edx					; EDX <- 0
			inc		ecx						; ECX <- number of block pairs + 1
			div		ecx						; ECX <- get number of block pairs
			add		_BlockDDFree[edi*SIZE CStat].CStat.NumL,dl ; number of blocks in LEFT direction
			add		_BlockDDFree[edi*SIZE CStat].CStat.NumR,dl ; number of blocks in RIGHT direction
			dec		ecx						; ECX <- number of block pairs
			sub		ecx,edx					; ECX <- number of block pairs in UP/DOWN direction
			add		_BlockDDFree[edi*SIZE CStat].CStat.NumD,cl ; number of blocks in DOWN direction
			add		_BlockDDFree[edi*SIZE CStat].CStat.NumU,cl ; number of blocks in UP direction

; ---------------------------------------------------------------------------
;                             Turn inner blocks
; ---------------------------------------------------------------------------

; --------- Prepare to generate down-color blocks

BlockTurn:	mov		esi,_BlockD[edi*4]		; ESI <- first block
			movzx	ecx,_BlockDNum[edi]		; ECX <- number of blocks
			or		ecx,ecx					; any blocks?
			jz		BlockTurn6				; no blocks

; --------- Prepare random number (-> EDX:EAX)

BlockTurn3:	Rand							; get random number

; --------- Get random spin (-> EBX)

BlockTurn4:	movzx	ebx,al					; EBX <- random number
			shr		eax,2					; shift random number
			jz		BlockTurn3				; generate next random number
			and		bl,3					; EBX <- random spin
			
; --------- Check if color in this direction is greater than current one
			
			movzx	edx,[esi].CBlock.ColD[ebx] ; EDX <- color in DOWN direction
			sub		edx,ICOLMIN				; EDX <- color relative to inner colors
			cmp		edx,edi					; is this color greater than current one?
			jle		BlockTurn42				; color is not greater
			mov		[esi].CBlock.Spin,bl	; set new spin (color is in DOWN direction)
			
; --------- Test statistic in DOWN/UP direction, if we can compensate it

			mov		bl,_Stat[edx*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumD ; BL <- DOWN statistic
			sub		bl,_Stat[edx*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumU ; BL <- DOWN-UP difference
			jb		BlockTurn46				; DOWN < UP, we can compensate it with this direction
			xor		[esi].CBlock.Spin,2		; turn color to UP direction
			or		bl,bl					; is DOWN > UP ?
			jnz		BlockTurn46				; DOWN > UP, we can compensate it with this direction

; --------- Change direction of color to LEFT
			
			dec		[esi].CBlock.Spin		; color is in LEFT direction
			and		[esi].CBlock.Spin,3		; mask direction
			
; --------- Test statistic in LEFT/RIGHT direction, if we can compensate it
			
			mov		bl,_Stat[edx*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumL ; BL <- LEFT statistic
			sub		bl,_Stat[edx*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumR ; BL <- LEFT-RIGHT difference
			jb		BlockTurn46				; LEFT < RIGHT, we can compensate it with this direction
			xor		[esi].CBlock.Spin,2		; turn color to RIGHT direction
			or		bl,bl					; is LEFT > RIGHT ?
			jnz		BlockTurn46				; LEFT > RIGHT, we can compensate it with this direction

; --------- Get another random spin (-> EBX)

BlockTurn42:movzx	ebx,al					; EBX <- random number
			shr		eax,2					; shift random number
			jz		BlockTurn3				; generate next random number
			and		bl,3					; EBX <- random spin

; --------- Check if color in this direction is greater than current one
			
			movzx	edx,[esi].CBlock.ColD[ebx] ; EDX <- color in DOWN direction
			sub		edx,ICOLMIN				; EDX <- color relative to inner colors
			cmp		edx,edi					; is this color greater than current one?
			jle		BlockTurn48				; color is not greater
			mov		[esi].CBlock.Spin,bl	; set new spin (color is in DOWN direction)

; --------- Test statistic in DOWN/UP direction, if we can compensate it

			mov		bl,_Stat[edx*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumD ; BL <- DOWN statistic
			sub		bl,_Stat[edx*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumU ; BL <- DOWN-UP difference
			jb		BlockTurn46				; DOWN < UP, we can compensate it with this direction
			xor		[esi].CBlock.Spin,2		; turn color to UP direction
			or		bl,bl					; is DOWN > UP ?
			jnz		BlockTurn46				; DOWN > UP, we can compensate it with this direction

; --------- Change direction of color to LEFT
			
			dec		[esi].CBlock.Spin		; color is in LEFT direction
			and		[esi].CBlock.Spin,3		; mask direction
			
; --------- Test statistic in LEFT/RIGHT direction, if we can compensate it
			
			mov		bl,_Stat[edx*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumL ; BL <- LEFT statistic
			sub		bl,_Stat[edx*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumR ; BL <- LEFT-RIGHT difference
			jb		BlockTurn46				; LEFT < RIGHT, we can compensate it with this direction
			xor		[esi].CBlock.Spin,2		; turn color to RIGHT direction
			or		bl,bl					; is LEFT > RIGHT ?
			jz		BlockTurn48				; LEFT == RIGHT, we cannot compensate it

; --------- Check if this direction is available for this block			
			
BlockTurn46:movzx	ebx,[esi].CBlock.Spin	; EBX <- current spin
			cmp		_BlockDDFree[edi*SIZE CStat].CStat.Num[ebx],0 ; is this direction available?
			jne		BlockTurn52				; direction is available, we can use it

; --------- Get another random spin (-> EBX)
			
			movzx	ebx,al					; EBX <- random number
			shr		eax,2					; shift random number
			jz		BlockTurn3				; generate next random number
			and		bl,3					; EBX <- random spin

; --------- Check if color in this direction is greater than current one
			
			movzx	edx,[esi].CBlock.ColD[ebx] ; EDX <- color in DOWN direction
			sub		edx,ICOLMIN				; EDX <- color relative to inner colors
			cmp		edx,edi					; is this color greater than current one?
			jle		BlockTurn48				; color is not greater
			
; --------- Prepare spin for LEFT direction

			inc		ebx						; EBX <- increase spin
			and		bl,3					; EBX <- mask spin
			mov		[esi].CBlock.Spin,bl	; set new spin (color is in LEFT direction)

; --------- Test statistic in LEFT/RIGHT direction, if we can compensate it
			
			mov		bl,_Stat[edx*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumL ; BL <- LEFT statistic
			sub		bl,_Stat[edx*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumR ; BL <- LEFT-RIGHT difference
			jb		BlockTurn47				; LEFT < RIGHT, we can compensate it with this direction
			xor		[esi].CBlock.Spin,2		; turn color to RIGHT direction
			or		bl,bl					; is LEFT > RIGHT ?
			jnz		BlockTurn47				; LEFT > RIGHT, we can compensate it with this direction

; --------- Change direction of color to UP
			
			inc		[esi].CBlock.Spin		; color is in UP direction
			and		[esi].CBlock.Spin,3		; mask direction
			
; --------- Test statistic in DOWN/UP direction, if we can compensate it

			mov		bl,_Stat[edx*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumD ; BL <- DOWN statistic
			sub		bl,_Stat[edx*SIZE CStat+ICOLMIN*SIZE CStat].CStat.NumU ; BL <- DOWN-UP difference
			jb		BlockTurn47				; DOWN < UP, we can compensate it with this direction
			xor		[esi].CBlock.Spin,2		; turn color to UP direction
			or		bl,bl					; is DOWN > UP ?
			jz		BlockTurn48				; DOWN == UP, we cannot compensate it

; --------- Prepare new spin (-> EBX)			

BlockTurn47:movzx	ebx,[esi].CBlock.Spin	; EBX <- current (new) spin
			jmp		BlockTurn52

; --------- Generate new spin of block

BlockTurn48:movzx	ebx,al					; EBX <- random number
			shr		eax,2					; EAX <- shift random number
			jz		BlockTurn3				; get new random number
			dec		ebx						; preset random spin - 1
			
; --------- Find free spin
			
BlockTurn5:	inc		ebx						; EBX <- next spin
			and		bl,3					; EBX = next spin
BlockTurn52:movzx	edx,_BlockDDFree[edi*SIZE CStat].CStat.Num[ebx] ; EDX <- number of free blocks
			or		edx,edx					; is this spin free?
			jz		BlockTurn5				; get next spin
			dec		edx						; decrease number of free blocks
			mov		_BlockDDFree[edi*SIZE CStat].CStat.Num[ebx],dl ; set new number of free blocks
			mov		[esi].CBlock.Spin,bl	; set new spin
			
; --------- Add block to color statistic

			movzx	edx,[esi].CBlock.ColD[ebx] ; EDX <- color in DOWN direction
			inc		_Stat[edx*SIZE CStat].CStat.NumD ; attach down color
			movzx	edx,[esi].CBlock.ColL[ebx] ; EDX <- color in LEFT direction
			inc		_Stat[edx*SIZE CStat].CStat.NumL ; attach left color
			movzx	edx,[esi].CBlock.ColU[ebx] ; EDX <- color in UP direction
			inc		_Stat[edx*SIZE CStat].CStat.NumU ; attach upper color
			movzx	edx,[esi].CBlock.ColR[ebx] ; EDX <- color in RIGHT direction
			inc		_Stat[edx*SIZE CStat].CStat.NumR ; attach right color

; --------- Next block

			add		esi,SIZE CBlock			; ESI <- next block
			dec		ecx						; counter of blocks
			jnz		BlockTurn4				; generate newxt block

; --------- Next color

BlockTurn6:	inc		edi						; EDI <- increase color index
			cmp		edi,ICOLNUM
			jb		DLTurn1					; next color

; ---------------------------------------------------------------------------
;                             Turn cross blocks
; ---------------------------------------------------------------------------

IFNDEF		MAXI

			mov		esi,[_BlockC]			; ESI <- first block
			movzx	ecx,[_BlockCNum]		; ECX <- number of blocks
			Rand							; get random number
CrossTurn2:	movzx	ebx,al					; EBX <- random number
			shr		eax,2					; EAX <- shift random number
			and		bl,3					; EBX = next spin
			mov		[esi].CBlock.Spin,bl	; set new spin
			add		esi,SIZE CBlock			; ESI <- next block
			loop	CrossTurn2				; next cross block

ENDIF ; MAXI

; ---------------------------------------------------------------------------
;                        Check variant history
; ---------------------------------------------------------------------------

IFDEF _DEBUG

; --------- Check color statistic

			call	_CheckStat				; check color statistic

; --------- Generate current solution (-> EBX hash)

			mov		esi,offset _Block		; ESI <- pointer to first block
			mov		ecx,BLOCKS/4			; ECX <- number of blocks / 4
			mov		edi,offset _Hist0		; EDI <- pointer to current solution
			xor		ebx,ebx					; EBX <- 0, hash accumulator
			xor		eax,eax					; EAX <- 0

CheckHist1:	mov		al,[esi].CBlock.Spin	; AL <- spin of one block
			add		esi,SIZE CBlock			; ESI <- next block
			shl		al,2					; rotate spin

			or		al,[esi].CBlock.Spin	; add spin of next block
			add		esi,SIZE CBlock			; ESI <- next block
			shl		al,2					; rotate spin

			or		al,[esi].CBlock.Spin	; add spin of next block
			add		esi,SIZE CBlock			; ESI <- next block
			shl		al,2					; rotate spin

			or		al,[esi].CBlock.Spin	; add spin of next block
			add		esi,SIZE CBlock			; ESI <- next block
			stosb							; store one spin tetrade
			
			add		bx,ax					; add spin tetrade to hash accumulator
			ror		bx,1					; shift hash accumulator
			loop	CheckHist1				; next spin tetrade
			
; --------- Check history

			mov		edx,_HistFirst[ebx*4]	; EDX <- index of solution
			or		edx,edx					; is any solution?
			js		CheckHist4				; no solution in history
CheckHist2:
IFDEF MAXI
			shl		edx,6					; EDX <- offset of one solution
			lea		esi,_Hist[edx]			; ESI <- solution in history
			shr		edx,4					; EDX <- index of solution * 4
			mov		edx,_HistNext[edx]		; EDX <- next solution
			mov		edi,offset _Hist0		; EDI <- this solution
			mov		ecx,BLOCKS/4/4			; ECX <- solution size in DWORDS
			repe	cmpsd					; compare solution
ELSE ; MAXI
			push	edx						; push EDX (solution index)
			mov		eax,BLOCKS/4			; EAX <- size of one solution
			mul		edx						; EAX <- solution offset
			pop		edx						; pop EDX
			mov		edx,_HistNext[edx*4]	; EDX <- next solution
			lea		esi,_Hist[eax]			; ESI <- solution in history
			mov		edi,offset _Hist0		; EDI <- this solution
			mov		ecx,BLOCKS/4			; ECX <- size of one solution
			repe	cmpsb					; compare solution
ENDIF ; MAXI			
			jne		CheckHist3				; solution is not equal
			inc		[_Similar]				; counter of similar solutions
CheckHist3:	or		edx,edx					; it is valid solution?
			jns		CheckHist2				; check next solution

; --------- Save solution

CheckHist4:	mov		eax,[_HistNum]			; EDX <- number of solutions in history
			cmp		eax,HISTSIZE			; is history full?
			jae		CheckHist5				; history is full
			inc		eax						; increase number of solutions
			mov		[_HistNum],eax			; set new number of solutions
			dec		eax						; return solution index
			mov		edx,_HistFirst[ebx*4]	; EDX <- first solution
			mov		_HistNext[eax*4],edx	; store next solution
			mov		_HistFirst[ebx*4],eax	; set this solution as first
IFDEF MAXI
			shl		eax,6					; EAX <- offset of this solution
			lea		edi,_Hist[eax]			; EDI <- solution history
			mov		esi,offset _Hist0		; ESI <- this solution
			mov		ecx,BLOCKS/4/4			; ECX <- solution size
			rep		movsd					; save this solution
ELSE ; MAXI
			mov		edx,BLOCKS/4			; EDX <- size of one solution
			mul		edx						; EAX <- offset of this solution
			lea		edi,_Hist[eax]			; EDI <- solution history
			mov		esi,offset _Hist0		; ESI <- this solution
			mov		ecx,BLOCKS/4			; ECX <- solultion size
			rep		movsb					; save this solution
ENDIF ; MAXI
CheckHist5:

ENDIF ; _DEBUG

; --------- Count number of variants and display current state

			inc		[_Variants]				; count variants
			call	_DispState				; display current state

; ---------------------------------------------------------------------------
;                         Generate hash tables
; ---------------------------------------------------------------------------

; --------- Clear hash tables

HashInit:	xor		eax,eax					; EAX <- 0
			mov		edi,offset _Hash		; EDI <- hash table without edges
			mov		ecx,COLNUM2				; ECX <- size of hash table
			rep		stosd					; clear hash table
			mov		edi,offset _HashR		; EDI <- hash table with right edge
			mov		ecx,COLNUM2				; ECX <- size of hash table
			rep		stosd					; clear hash table
			mov		edi,offset _HashD		; EDI <- hash table with down edge
			mov		ecx,COLNUM2				; ECX <- size of hash table
			rep		stosd					; clear hash table
			mov		edi,offset _HashDR		; EDI <- hash table with edges
			mov		ecx,COLNUM2				; ECX <- size of hash table
			rep		stosd					; clear hash table
			
; --------- Add blocks into hash tables (except hints)

			mov		esi,offset _Block		; ESI <- blocks
			mov		ecx,BLOCKS				; ECX  <- number of blocks
HashInit2:	and		[esi].CBlock.Flags,NOT USED ; clear used flag
			test	[esi].CBlock.Flags,HINT ; is it hint?
			jnz		HashInit5				; block is hint
			
			movzx	ebx,[esi].CBlock.Spin	; EBX <- current spin
			cmp		[esi].CBlock.ColR[ebx],0 ; has it edge in right direction?
			je		HashInit3				; it has edge in right direction
			mov		edi,offset _Hash		; EDI <- hash table without edges
			cmp		[esi].CBlock.ColD[ebx],0 ; has it edge in down direction?
			jne		HashInit4				; no, it has no edge
			mov		edi,offset _HashD		; EDI <- hash table with down edge
			jmp		short HashInit4
HashInit3:	mov		edi,offset _HashR		; EDI <- hash table with right edge
			cmp		[esi].CBlock.ColD[ebx],0 ; has it edge in down direction?
			jne		HashInit4				; no, it has only right edge
			mov		edi,offset _HashDR		; EDI <- hash table with edges
	
HashInit4:	movzx	edx,[esi].CBlock.ColU[ebx] ; EDX <- color in UP direction
			mov		dh,[esi].CBlock.ColL[ebx] ; add left color
			lea		edi,[edi+edx*4]			; EDI <- pointer to hash table

			mov		eax,[edi]				; EAX <- old entry in hash table
			mov		[esi].CBlock.Next,eax	; link to old entry
			mov		[edi],esi				; store this block into hash table

HashInit5:	add		esi,SIZE CBlock			; ESI <- next block
			loop	HashInit2				; next block

; ---------------------------------------------------------------------------
;                             Initialize board
; ---------------------------------------------------------------------------

IFDEF MAXI

; --------- Save hint 0 into board

BoardInit:	mov		esi,[_Hint0]			; ESI <- hint 0 block
			mov		edi,offset _Board + H0POS*SIZE CField ; EDI <- hint 0 position
			mov		[edi].CField.Block,esi ; store pointer to block
			mov		eax,dword ptr [esi].CBlock.ColorDW ; EAX <- block colors
			mov		[edi].CField.ColorDW,eax ; store colors into field
			or		[esi].CBlock.Flags,USED	; set used flag

IFDEF USEHINTS

; --------- Save hint 1 into board

			mov		esi,[_Hint1]			; ESI <- hint 1 block
			mov		edi,offset _Board + H1POS*SIZE CField ; EDI <- hint 1 position
			mov		[edi].CField.Block,esi ; store pointer to block
			mov		eax,dword ptr [esi].CBlock.ColorDW ; EAX <- block colors
			mov		[edi].CField.ColorDW,eax ; store colors into field
			or		[esi].CBlock.Flags,USED	; set used flag
			
; --------- Save hint 2 into board

			mov		esi,[_Hint2]			; ESI <- hint 2 block
			mov		edi,offset _Board + H2POS*SIZE CField ; EDI <- hint 2 position
			mov		[edi].CField.Block,esi ; store pointer to block
			mov		eax,dword ptr [esi].CBlock.ColorDW ; EAX <- block colors
			mov		[edi].CField.ColorDW,eax ; store colors into field
			or		[esi].CBlock.Flags,USED	; set used flag

ENDIF ; USEHINTS

ENDIF ; MAXI

; ---------------------------------------------------------------------------
;                          Try to solve puzzle
; ---------------------------------------------------------------------------

			prvlab textequ <PopState>
			y = 0
REPEAT HGHT
			x = 0
REPEAT WDTH
			strtlab catstr <s>, %x, <_>, %y
			contlab catstr <c>, %x, <_>, %y

IF (x EQ WDTH-1)

 IF (y EQ HGHT-1)
			nxtlab textequ <AttackStop>
 ELSE
			nxtlab catstr <s>, %0, <_>, %y + 1
 ENDIF

ELSE

 IF (x EQ H0POSX-1) AND (y EQ H0POSY) AND (USEHINT0 EQ 1)
			nxtlab catstr <s>, %x + 2, <_>, %y
 ELSEIF (x EQ H1POSX-1) AND (y EQ H1POSY) AND (USEHINT1 EQ 1)
			nxtlab catstr <s>, %x + 2, <_>, %y
 ELSEIF (x EQ H2POSX-1) AND (y EQ H2POSY) AND (USEHINT2 EQ 1)
			nxtlab catstr <s>, %x + 2, <_>, %y
 ELSE
			nxtlab catstr <s>, %x + 1, <_>, %y
 ENDIF
			
ENDIF

IF (x EQ H0POSX) AND (y EQ H0POSY) AND (USEHINT0 EQ 1)
ELSEIF (x EQ H1POSX) AND (y EQ H1POSY) AND (USEHINT1 EQ 1)
ELSEIF (x EQ H2POSX) AND (y EQ H2POSY) AND (USEHINT2 EQ 1)
ELSE
			Solve1 x, y, nxtlab, prvlab, strtlab, contlab
			prvlab textequ contlab
ENDIF
			x = x + 1
ENDM
			y = y + 1
ENDM

; ---------------------------------------------------------------------------
;                   Pop current state and repeat solution
; ---------------------------------------------------------------------------

; --------- Pop current state

PopState:	mov		edi,offset _Stat		; EDI <- color statistic
			mov		esi,offset _StatSave	; ESI <- saved color statistic
			mov		ecx,COLORS*SIZE CStat/4	; ECX <- size of color statistic
			rep		movsd					; pop current state
			dec		[_Repeat]				; repeat counter
			jnz		DLTurn					; repeat
			jmp		Corner					; reset

; --------- Pop registers

AttackStop:	popa							; pop all registers
			ret

			END
