
//////////////////////////////////////////////////////////////////////////////
//                    Eternity II Block Set Solver
//////////////////////////////////////////////////////////////////////////////

#include "Etern2Set.h"

//////////////////////////////////////////////////////////////////////////////
// Variables

unsigned __int64	RandSeed;				// seed of random generator
int					Variants = 0;			// number of variants
int					Level = 0;				// current level
int					MaxLevel = 0;			// maximal level
int					AllOK = 0;				// solved counter

// Current state
struct CBlock		Block[BLOCKS];			// blocks
struct CStat		Stat[COLORS];			// color statistic
struct CStat		StatSave[COLORS];		// saved color statistic
int					Repeat;					// repeat counter
struct CBlock*		CheckBlock[WIDTH];		// blocks for edge check

#ifdef MAXI
struct CBlock*		Hint0;					// hint 0
#ifdef USEHINTS
struct CBlock*		Hint1;					// hint 1
struct CBlock*		Hint2;					// hint 2
#endif // USEHINTS
#endif // MAXI

// Corner initial spins
char CornIni[CORNINI*4];					// corner initial spins

// Sorted edges
struct CBlock*		EdgeL[OCOLNUM];			// start pointer of left-color edges
struct CBlock*		EdgeR[OCOLNUM];			// start pointer of right-color edges
struct CBlock*		EdgeD[OCOLNUM];			// start pointer of double-color edges

char				EdgeLNum[OCOLNUM];		// number of left-color edges
char				EdgeRNum[OCOLNUM];		// number of right-color edges
char				EdgeDNum[OCOLNUM];		// number of double-color edges

// Sorted inner blocks
struct CBlock*		BlockD[ICOLNUM];		// start pointer of down-color block
struct CBlock*		BlockDL[ICOLNUM];		// start pointer of down-left-color block

char				BlockDNum[ICOLNUM];		// number of down-color blocks
char				BlockDLNum[ICOLNUM];	// number of down-left-color blocks

// Cross blocks (they do not affect color statistic)
#ifndef MAXI
struct CBlock*		BlockC = 0;				// start pointer of cross blocks
char				BlockCNum = 0;			// number of cross blocks
#endif // MAXI

// Edge statistic
char				EdgeFree[4];			// free positions in all directions
struct CStat		EdgeLMin[OCOLNUM];		// minimal number of left-color edges
struct CStat		EdgeRMin[OCOLNUM];		// minimal number of right-color edges
struct CStat		EdgeLRMax[OCOLNUM];		// maximal number of edge pairs
char				EdgeLRIni[OCOLNUM];		// initial number of edge pairs
struct CStat		EdgeLRNum[OCOLNUM];		// current number of edge pairs
struct CStat		EdgeLFree[OCOLNUM];		// number of free left-color edge positions
struct CStat		EdgeRFree[OCOLNUM];		// number of free right-color edge positions

// Inner block statistic
struct CStat		BlockDDFree[ICOLNUM];	// number of free down-color block positions

// Solve
struct CField		Board[BLOCKS];			// board
struct CBlock*		Hash[COLNUM2];			// block hash table without edges
struct CBlock*		HashR[COLNUM2];			// block hash table with right edge
struct CBlock*		HashD[COLNUM2];			// block hash table with down edge
struct CBlock*		HashDR[COLNUM2];		// block hash table with down and right edge
// Hash index: (up + left*COLNUM)

// History check
#ifdef _DEBUG
BYTE Hist[HISTSIZE*BLOCKS/4];				// solution history (size 64 MB)
BYTE Hist0[BLOCKS/4];						// current solution
int HistFirst[HISTHASH];					// history hash - first entry (-1 = none, size 256 KB)
int HistNext[HISTSIZE];						// history hash - next entry (-1 = none, size 4 MB)
int Similar = 0;							// number of similar solutions
int HistNum = 0;							// number of entries in history
#endif // _DEBUG

// Color map
#ifdef MAXI
char ColInvMap[COLORS] = {
	0, 1, 2, 3, 4, 5,
	11,16,12,8,13,14,6,22,9,19,20,10,15,21,18,7,17
};
char ColMap[COLORS];
#endif // MAXI

//////////////////////////////////////////////////////////////////////////////
// Initialize random generator

void RandInit()
{
	// Get system time
	SYSTEMTIME systime;
	::GetLocalTime(&systime);

	// Prepare temporary time value
	unsigned __int64 time = systime.wMilliseconds + 137*systime.wSecond + 
						2567*systime.wMinute + 358946*systime.wDay;

	// Initialize random generator
	RandSeed = (time + 486874*time);
}

//////////////////////////////////////////////////////////////////////////////
// Generate WORD random number

WORD GetRand()
{
	RandSeed = RandSeed*214013 + 2531011;
	RandSeed = (RandSeed >> 8) | (RandSeed << (64-8));
	return (WORD)(RandSeed >> 16);
}

//////////////////////////////////////////////////////////////////////////////
// Generate blocks

void GenBlocks()
{
	// Local variables
	char tmp;
	int i,j;
	struct CColor* color = Color;
	struct CBlock* block = Block;

#ifdef MAXI
	for (i = 0; i < COLORS; i++) ColMap[ColInvMap[i]] = i;
#endif // MAXI

	// Cycle through all blocks
	for (i = 0; i < BLOCKS; i++)
	{
		// Set colors (colors are in inverse direction)
#ifdef MAXI
		block->ColD = ColMap[color->ColD];
		block->ColR = ColMap[color->ColR];
		block->ColU = ColMap[color->ColU];
		block->ColL = ColMap[color->ColL];
#else // MAXI
		block->ColD = color->ColD;
		block->ColR = color->ColR;
		block->ColU = color->ColU;
		block->ColL = color->ColL;
#endif // MAXI

		// Initialize block
		block->Flags = 0;
		block->Block = i;
		block->Spin0 = 0;
		block->Spin = 0;

#ifndef MAXI
		// Normalize corners
		if ((block->ColR == EDGECOL) && (block->ColD == EDGECOL)) block->Spin0 = 1;
		if ((block->ColU == EDGECOL) && (block->ColR == EDGECOL)) block->Spin0 = 2;
		if ((block->ColL == EDGECOL) && (block->ColU == EDGECOL)) block->Spin0 = 3;

		// Normalize edges
		if ((block->ColU != EDGECOL) && (block->ColR == EDGECOL) && (block->ColD != EDGECOL)) block->Spin0 = 1;
		if ((block->ColL != EDGECOL) && (block->ColU == EDGECOL) && (block->ColR != EDGECOL)) block->Spin0 = 2;
		if ((block->ColD != EDGECOL) && (block->ColL == EDGECOL) && (block->ColU != EDGECOL)) block->Spin0 = 3;
#endif // MAXI

		// Hints
#ifdef MAXI
		if (i == H0BLOCK) { block->Spin0 = H0SPIN; block->Flags = HINT; }
#ifdef USEHINTS
		if (i == H1BLOCK) { block->Spin0 = H1SPIN; block->Flags = HINT; }
		if (i == H2BLOCK) { block->Spin0 = H2SPIN; block->Flags = HINT; }
#endif // USEHINTS
#endif // MAXI

		// Rotate block
		for (j = block->Spin0; j > 0; j--)
		{
			tmp = block->ColD;
			block->ColD = block->ColR;
			block->ColR = block->ColU;
			block->ColU = block->ColL;
			block->ColL = tmp;
		}
		block->ColorDW[1] = block->ColorDW[0];

		// Internal check - hint colors
#ifdef _DEBUG
#ifdef MAXI
		if (i == H0BLOCK)
			ASSERT((block->ColD == H0COLD) && (block->ColL == H0COLL) &&
					(block->ColU == H0COLU) && (block->ColR == H0COLR));
#ifdef USEHINTS
		if (i == H1BLOCK)
			ASSERT((block->ColD == H1COLD) && (block->ColL == H1COLL) &&
					(block->ColU == H1COLU) && (block->ColR == H1COLR));
		if (i == H2BLOCK)
			ASSERT((block->ColD == H2COLD) && (block->ColL == H2COLL) &&
					(block->ColU == H2COLU) && (block->ColR == H2COLR));
#endif // USEHINTS
#endif // MAXI
#endif // _DEBUG

		// Shift pointers
		block++;
		color++;
	}
}

//////////////////////////////////////////////////////////////////////////////
// Sort blocks

void SortBlocks()
{
	// Local variables
	char col, cD, cL, cU, cR, cT;
	int spin;
	struct CBlock tmp;
	struct CBlock* block;
	struct CBlock* dst;

	// Sort corners
#ifdef MAXI
	dst = &Block[4];
#else // MAXI
	dst = &Block[0];
	for (block = dst; block != &Block[BLOCKS]; block++)
	{
		if (block->ColL == EDGECOL)
		{
			tmp = *dst;
			*dst = *block;
			*block = tmp;
			dst++;
		}
	}
	ASSERT(dst == &Block[4]);
#endif // MAXI

	// Cycle through outer colors
	for (col = OCOLMIN; col <= OCOLMAX; col++)
	{
		// Initialize start pointer of left-color edges
		EdgeL[col-OCOLMIN] = dst;
		EdgeLNum[col-OCOLMIN] = 0;

		// Cycle through edge blocks
		for (block = dst; block != &Block[BLOCKS]; block++)
		{
			// Check if it is left-color edge
			if ((block->ColD == EDGECOL) && (block->ColL == col) && (block->ColR != col))
			{
				// Store this edge
				tmp = *dst;
				*dst = *block;
				*block = tmp;
				dst++;
				EdgeLNum[col-OCOLMIN]++;
			}
		}

		// Initialize start pointer of right-color edges
		EdgeR[col-OCOLMIN] = dst;
		EdgeRNum[col-OCOLMIN] = 0;

		// Cycle through edge blocks
		for (block = dst; block != &Block[BLOCKS]; block++)
		{
			// Check if it is right-color edge
			if ((block->ColD == EDGECOL) && (block->ColL != col) && (block->ColR == col))
			{
				// Store this edge
				tmp = *dst;
				*dst = *block;
				*block = tmp;
				dst++;
				EdgeRNum[col-OCOLMIN]++;
			}
		}

		// Initialize start pointer of double-color edges
		EdgeD[col-OCOLMIN] = dst;
		EdgeDNum[col-OCOLMIN] = 0;

		// Cycle through edge blocks
		for (block = dst; block != &Block[BLOCKS]; block++)
		{
			// Check if it is double-color edge
			if ((block->ColD == EDGECOL) && (block->ColL == col) && (block->ColR == col))
			{
				// Store this edge
				tmp = *dst;
				*dst = *block;
				*block = tmp;
				dst++;
				EdgeDNum[col-OCOLMIN]++;
			}
		}
	}

	// Debug check
	ASSERT(dst == &Block[CEBLOCK]);

	// Cycle through inner colors
	for (col = ICOLMIN; col <= ICOLMAX; col++)
	{
		// Initialize start pointer of down-color blocks
		BlockD[col-ICOLMIN] = dst;
		BlockDNum[col-ICOLMIN] = 0;

		// Cycle through inner blocks
		for (block = dst; block != &Block[BLOCKS]; block++)
		{
			// Check if it is not locked block
			if ((block->Flags & HINT) != 0) continue;

			// Get colors of block
			cD = block->ColD;
			cL = block->ColL;
			cU = block->ColU;
			cR = block->ColR;

			// One-color block
			spin = -1;
			if ((cD == col) && (cL != col) && (cU != col) && (cR != col)) spin = 0;
			if ((cD != col) && (cL == col) && (cU != col) && (cR != col)) spin = 1;
			if ((cD != col) && (cL != col) && (cU == col) && (cR != col)) spin = 2;
			if ((cD != col) && (cL != col) && (cU != col) && (cR == col)) spin = 3;

			// Three-color block
			if ((cD == col) && (cL == col) && (cU != col) && (cR == col)) spin = 0;
			if ((cD == col) && (cL == col) && (cU == col) && (cR != col)) spin = 1;
			if ((cD != col) && (cL == col) && (cU == col) && (cR == col)) spin = 2;
			if ((cD == col) && (cL != col) && (cU == col) && (cR == col)) spin = 3;

			// Block is valid
			if (spin >= 0)
			{
				// Anti-rotate block
				block->Spin0 = (4 - spin) & 3;

				for (; spin > 0; spin--)
				{
					cT = cD; cD = cL; cL = cU; cU = cR; cR = cT;
				}

				// Store rotated colors
				block->ColD = cD;
				block->ColL = cL;
				block->ColU = cU;
				block->ColR = cR;
				block->ColorDW[1] = block->ColorDW[0];

				// Store this block
				tmp = *dst;
				*dst = *block;
				*block = tmp;
				dst++;
				BlockDNum[col-ICOLMIN]++;
			}
		}

		// Initialize start pointer of down-left-color blocks
		BlockDL[col-ICOLMIN] = dst;
		BlockDLNum[col-ICOLMIN] = 0;

		// Cycle through inner blocks
		for (block = dst; block != &Block[BLOCKS]; block++)
		{
			// Check if it is not locked block
			if ((block->Flags & HINT) != 0) continue;

			// Get colors of block
			cD = block->ColD;
			cL = block->ColL;
			cU = block->ColU;
			cR = block->ColR;

			// Two-color block
			spin = -1;
			if ((cD == col) && (cL == col) && (cU != col) && (cR != col)) spin = 0;
			if ((cD != col) && (cL == col) && (cU == col) && (cR != col)) spin = 1;
			if ((cD != col) && (cL != col) && (cU == col) && (cR == col)) spin = 2;
			if ((cD == col) && (cL != col) && (cU != col) && (cR == col)) spin = 3;

			// Block is valid
			if (spin >= 0)
			{
				// Anti-rotate block
				block->Spin0 = (4 - spin) & 3;

				for (; spin > 0; spin--)
				{
					cT = cD; cD = cL; cL = cU; cU = cR; cR = cT;
				}

				// Store rotated colors
				block->ColD = cD;
				block->ColL = cL;
				block->ColU = cU;
				block->ColR = cR;
				block->ColorDW[1] = block->ColorDW[0];

				// Store this block
				tmp = *dst;
				*dst = *block;
				*block = tmp;
				dst++;
				BlockDLNum[col-ICOLMIN]++;
			}
		}
	}

#ifndef MAXI
	// Initialize start pointer of cross blocks
	BlockC = dst;
	BlockCNum = 0;

	// Cycle through cross blocks
	for (block = dst; block != &Block[BLOCKS]; block++)
	{
		// Check if it is cross block
		if ((block->ColD == block->ColU) && (block->ColL == block->ColR))
		{
			// Store this block
			tmp = *dst;
			*dst = *block;
			*block = tmp;
			dst++;
			BlockCNum++;
		}	
	}
#endif // MAXI

#ifdef MAXI
	// Save hints
	Hint0 = &Block[255];
	ASSERT(Hint0->Block == H0BLOCK);
#ifdef USEHINTS
	Hint1 = &Block[253];
	ASSERT(Hint1->Block == H1BLOCK);
	Hint2 = &Block[254];
	ASSERT(Hint2->Block == H2BLOCK);
#endif // USEHINTS
#endif // MAXI

	// Debug check
	ASSERT(dst == &Block[BLOCKS-HINTNUM]);
}

//////////////////////////////////////////////////////////////////////////////
// Debug check color statistic

#ifdef _DEBUG
void CheckStat()
{
	// Local variables
	int i;
	struct CStat stat[COLORS];
	struct CBlock* block = Block;

	// Clear color statistic
	for (i = 0; i < COLORS; i++) stat[i].NumDW = 0;

	// Cycle through all blocks
	for (i = 0; i < BLOCKS; i++)
	{
		// Add block into color statistic
		if (block->ColD == EDGECOL)
		{
			if (block->ColL == EDGECOL)
				stat[block->ColU].Num[(block->Spin+1)&3]++;
			else
			{
				stat[block->ColL].Num[block->Spin]++;
				stat[block->ColU].Num[block->Spin^2]++;
			}
			stat[block->ColR].Num[block->Spin]--;
		}
		else
		{
			stat[*(&block->ColD+block->Spin)].NumD++;
			stat[*(&block->ColR+block->Spin)].NumR++;
			stat[*(&block->ColU+block->Spin)].NumU++;
			stat[*(&block->ColL+block->Spin)].NumL++;
		}
		block++;
	}

	// Check corner spins
	ASSERT(Block[1].Spin != Block[0].Spin);
	ASSERT((Block[2].Spin != Block[1].Spin)
			&& (Block[2].Spin != Block[0].Spin));
	ASSERT((Block[3].Spin != Block[2].Spin)
			&& (Block[3].Spin != Block[1].Spin)
			&& (Block[3].Spin != Block[0].Spin));

	for (i = OCOLMIN; i <= OCOLMAX; i++)
	{
		ASSERT(Stat[i].NumD == 0);
		ASSERT(Stat[i].NumL == 0);
		ASSERT(Stat[i].NumU == 0);
		ASSERT(Stat[i].NumR == 0);

		ASSERT(stat[i].NumD == 0);
		ASSERT(stat[i].NumL == 0);
		ASSERT(stat[i].NumU == 0);
		ASSERT(stat[i].NumR == 0);
	}

	for (i = ICOLMIN; i <= ICOLMAX; i++)
	{
		ASSERT(Stat[i].NumD == Stat[i].NumU);
		ASSERT(Stat[i].NumL == Stat[i].NumR);

		ASSERT(stat[i].NumD == stat[i].NumU);
		ASSERT(stat[i].NumL == stat[i].NumR);
	}
}
#endif // _DEBUG

//////////////////////////////////////////////////////////////////////////////
// Generate corner variants

void CornInit()
{
	// Local variables
	int s0,s1,s2,s3;
	int i = 0;

	// Cycle through all variants of corners
	for (s0 = 0; s0 < 4; s0++)
	{
		for (s1 = 0; s1 < 4; s1++)
		{
			for (s2 = 0; s2 < 4; s2++)
			{
				for (s3 = 0; s3 < 4; s3++)
				{
					if ((s1 != s0) &&
						(s2 != s1) && (s2 != s0) &&
						(s3 != s2) && (s3 != s1) && (s3 != s0))
					{
						CornIni[i++] = s0;
						CornIni[i++] = s1;
						CornIni[i++] = s2;
						CornIni[i++] = s3;
					}
				}
			}
		}
	}
	ASSERT(i == 4*CORNINI);
}

//////////////////////////////////////////////////////////////////////////////
// Save board

void SaveBoard(BOOL best)
{
	// Local variables
	FILE* file;
	char filename[50];
	int i;
	struct CBlock* block;

	// Open file
	if (best)
	{
		sprintf(filename, "Best%03d.csv", MaxLevel);
		file = fopen(filename, "w");
	}
	else
		file = fopen("Result.csv", "w");

	if ((file == NULL) && best)
		file = fopen("Best.csv", "w");

	if (file == NULL) return;

	// Save block set
	for (i = 0; i < BLOCKS; i++)
	{
		block = Board[i].Block;
		if ((block == NULL)	|| ((i >= Level)
#ifdef MAXI
			&& (i != H0BLOCK)
#ifdef USEHINTS
			&& (i != H1BLOCK) && (i != H2BLOCK)
#endif
#endif
			))
			fprintf(file, "  0,0");
		else
			fprintf(file, "%3d,%1d", (int)(block->Block+1), (int)((block->Spin + block->Spin0) & 3));

		if ((i % WIDTH) != WIDTH-1)
			fprintf(file, ";");
		else
			fprintf(file, "\n");
	}

	// Close file
	fclose(file);
}

//////////////////////////////////////////////////////////////////////////////
// Save block set

void SaveSet(BOOL best)
{
	// Local variables
	FILE* file;
	char filename[50];

	// Open file
	if (best)
	{
		sprintf(filename, "Set%03d.csv", MaxLevel);
		file = fopen(filename, "w");
	}
	else
		file = fopen("Blocks.csv", "w");

	if (file == NULL) return;

	// Save block set
	for (int i = 0; i < BLOCKS; i++)
	{
		fprintf(file, "%3d,%1d", (int)(Block[i].Block+1), (int)((Block[i].Spin + Block[i].Spin0) & 3));

		if ((i % WIDTH) != WIDTH-1)
			fprintf(file, ";");
		else
			fprintf(file, "\n");
	}

	// Close file
	fclose(file);
}

//////////////////////////////////////////////////////////////////////////////
// Display current state

void DispState()
{
#ifdef MAXI
	static int disp = 500;
	disp--;
	if (disp > 0) return;
	disp = 500;
#elif MIDI
	static int disp = 100;
	disp--;
	if (disp > 0) return;
	disp = 100;
#else
	static int disp = 5000;
	disp--;
	if (disp > 0) return;
	disp = 5000;
#endif

#ifdef MINI

#ifdef _DEBUG
	printf("\rRepeat %d of %d variants, Max. level %d, Solutions: %d ", Similar, Variants, MaxLevel, AllOK);
#else // HISTSIZE
	printf("\rVariants %d, Max. level %d, Solutions: %d  ", Variants, MaxLevel, AllOK);
#endif // HISTSIZE

#else // MINI

#ifdef _DEBUG
	printf("\rRepeat %d of %d variants, Max. level %d ", Similar, Variants, MaxLevel);
#else // HISTSIZE
	printf("\rVariants %d, Max. level %d  ", Variants, MaxLevel);
#endif // HISTSIZE

#endif // MINI

	//static int save = 5;
	//save--;
	//if (save <= 0)
	//{
	//	save = 5;
	//	SaveBoard(FALSE);
//		SaveSet(FALSE);
//	}
}

//////////////////////////////////////////////////////////////////////////////
// Save best board

void SaveBest()
{
	SaveBoard(TRUE);
//	SaveSet(TRUE);
	SaveBoard(FALSE);
	SaveSet(FALSE);
}

//////////////////////////////////////////////////////////////////////////////
// Main function

int main(int argc, char* argv[])
{
#ifndef _DEBUG
	// Set low priority class (only with Release build)
	::SetPriorityClass(::GetCurrentProcess(), IDLE_PRIORITY_CLASS);
#endif

	// Message
#ifdef MAXI

#ifdef USEHINTS
	printf("Eternity II Block Set Solver (with hints)\n");
#else // USEHINTS
	printf("Eternity II Block Set Solver\n");
#endif // USEHINTS

#elif MIDI
	printf("Eternity II Block Set Solver (6x12 Puzzle)\n");
#elif MINI
	printf("Eternity II Block Set Solver (6x6 Puzzle)\n");
#endif

	// Initialize random generator
	RandInit();

	// Clear history
#ifdef _DEBUG
	memset(HistFirst, -1, HISTHASH*sizeof(int));
	memset(HistNext, -1, HISTSIZE*sizeof(int));
#endif // _DEBUG

#ifdef MINI
	for (;;)
#endif
	{
		// Generate blocks
		GenBlocks();

		// Sort blocks
		SortBlocks();

		// Generate corner variants
		CornInit();

		// Solve puzzle
		Attack();

		// Save puzzle
		AllOK++;
		SaveBest();
		SaveSet(TRUE);
		DispState();
	};
	printf("\rFinished: Variants %d, Max. level %d  ", Variants, MaxLevel);

	return 0;
}
